(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_mid lang_tgt.
Require Import model_common model_untyped model_properties_untyped.
Require Import gw_common gw_untyped codegen.
Require Import ssr util.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


Lemma tr_mod_aux_labels: forall a b c m,
  map fst m = map fst (g_cptrs (tr_mod_aux a b c m)).
Proof.
  i; ginduction m; vauto; i; s; destruct a; s; erewrite IHm; eauto.
Qed.



(** * The local world used for the proof. *)

Program Definition mylwsts: transys :=
  
  {| state := stack * heap
   ; ext    s' s := <<HP: forall l v (IN: pmread (snd s) l v), pmread (snd s') l v>>
   ; extpub s' s := <<HP: forall l v (IN: pmread (snd s) l v), pmread (snd s') l v>>
                 /\ <<ST: fst s' = fst s>>
  |}.
Next Obligation. esplits; ss; [auto|congruence]. Qed.

(** Conceptually, the local stack is actually a stack of stacks: one
    per function that hasn't returned yet.
*)

Program Definition mylwconf: method_conf tgt_lsi mid_lsi (tsprod (gwsts tm_gw) mylwsts) :=
  {| confrel g s ca cb :=
       <<Ca: ca = {| cpc := Nothing; crf := Nothing
                   ; cst := fst (snd s); chp := snd (snd s) |}>> /\
       <<Cb: cb = (pmemp, Nothing)>>
  |}.

Program Definition mylw: world_loca tm_gw :=
  {| lwwfo := gwfo
   ; lwsts := mylwsts; lwconf := mylwconf |}.
Next Obligation.
  unfold tgt_extra, mid_extra; unnw; subst; ss; eauto.
Qed.


Notation tm_rf := (fun s => fst (fst (fst (fst s)))).
Notation tm_db := (fun s => snd (fst (fst (fst s)))).
Notation tm_lhp := (fun s: state (wlift mylw) => snd (snd s)).
Notation tm_lst := (fun s: state (wlift mylw) => fst (snd s)).


Lemma mylw_stable: stable mylw.
Proof.
  ii; rrdes CR; rr; esplits; s; eauto.
Qed.


Lemma mylw_confrel_shape: forall s r ca cb
    (CR: confrel mylw s r ca cb),
  exists lsta lhpa,
    ca = {| cpc := Nothing; crf := Nothing; cst := lsta; chp := lhpa |} /\
    cb = (pmemp, Nothing).
Proof.
  i; destruct ca, cb; rrdes CR; ss; subst; eauto.
Qed.




(** * Continuation and program relations. *)

Fixpoint ksplit (envb: env) : option (contvar * env) :=
  match envb with
    | nil => None
    | (inr k, _) :: envb' => Some (k, envb')
    | _ :: envb' => ksplit envb'
  end.

(** It's important that [tm_ksim] is defined inductively.  Otherwise
    the definition wouldn't say anything about continuations that have
    been shadowed and hence one would need to rely on the uniqueness
    assumption, which is unnecessarily complicated.
*)

Inductive tm_ksim g (si s: state (wlift mylw)) kia kib i (envb: env) n : list contvar -> Prop :=
| tm_ksim_nil
    (SP: 2 * n <= tm_rf s r_sp)
  : tm_ksim g si s kia kib i envb n nil
| tm_ksim_cons envb' k ks ka kaenv kb
    (SPL: ksplit envb = Some (k, envb'))
    (KSIM: tm_ksim g si s kia kib i envb' (1 + n) ks)
    (RET: pmread (tm_lst s) (tm_rf s r_sp - 2 * (1 + n)) ka)
    (ENV: pmread (tm_lst s) (tm_rf s r_sp - 2 * (1 + n) + 1) kaenv)
    (Kb: lookup (inr k) envb = Some kb)
    (ESIM: forall g' (LE: g <g= g')
                  s' (EXT: (wlift mylw).(ext) s' s)
                  
                  
                  (SP: tm_rf s' r_sp = tm_rf s r_sp - 2 * (1 + n))
                  popped
                  (LST: tm_lst s = tm_lst s' ## {tm_rf s' r_sp |=> popped})
                  (LEN: length popped = 2 * (1 + n))

                  
                  (ENV: tm_rf s' r_env = kaenv)

                  va vb (RV: vclos (wlift mylw) s' (g' s') va vb)
                  ea (RETa: cquerya (wlift mylw) s' (ret va ka) ea)
                  eb (RETb: cqueryb (wlift mylw) s' (ret vb kb) eb),

             esim (wlift mylw) i kia kib g' si s' true ea eb)
  : tm_ksim g si s kia kib i envb n (k :: ks)
.
Hint Constructors tm_ksim.


(** At this level, a "program" on the target side is a function from
    imports (labelled addresses) to (encoded) instructions.  It does
    not take a load address as [tr_exp] does not depend on it (only
    the module level translation does).
 *)
Definition tm_psim (G: list var) ea eb : Prop := forall i, exists i',
  forall g si s kia kib senva denva envb
    (GK: gknow (wlift mylw) g)
    (EXT: (wlift mylw).(ext) s si)

    
    (RSV: forall l (IN: In (inl (inl l)) G), exists va vb,
           <<INa: lookup l senva = Some va>> /\
           <<INb: lookup (inl (inl l)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (SORD: fvars G = map fst senva)

    
    (RDV: forall x (IN: In (inl (inr x)) G), exists va vb,
           <<INa: lookup x denva = Some va>> /\
           <<INb: lookup (inl (inr x)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (ENV: reprlist (tm_lhp s) (tm_rf s r_env) (map snd denva))
    (DORD: tvars G = map fst denva)

    
    cur oldst
    (LST: tm_lst s = {tm_rf si r_sp |=> cur} ## oldst)
    (LEN: length cur = 2 * length (kvars G))
    (SP: tm_rf s r_sp = 2 * length (kvars G) + tm_rf si r_sp)
    (KSIM: tm_ksim g si s kia kib i envb 0 (kvars G))

    
    laddr n
    (NZ: laddr <> 0) (Eladdr: exists h, tm_lhp s = {laddr |=> ea n} ## {laddr - n |=> map snd senva} ## h),

  esim (wlift mylw) (wfo_add i' i) kia kib g si s true
       {| cpc := Just laddr; crf := Nothing; cst := pmemp; chp := pmemp |}
       (pmemp, Just (envb, eb)).


(** Notes on basic/atomic expressions:
    - The code "returns" to whatever comes after.
    - Its result is stored in the heap location pointed to by [r_arg].
    - Hence we bake in a custom notion of continuation relation here.
    - Registers [r_arg], [r_env] and [r_sp] are preserved.
    - The code does not touch the stack, but grows the heap and of course modifies
      the result location pointed to by [r_arg].
    - This location is *not yet* owned by the local world because its final
      value has yet to be written (and the heap owned by the local world is
      immutable).
*)

Definition tm_bsim (G: list var) ea eb : Prop :=
  forall g si s kia kib i senva denva envb
    (GK: gknow (wlift mylw) g)
    (EXT: (wlift mylw).(ext) s si)

    
    (RSV: forall l (IN: In (inl (inl l)) G), exists va vb,
           <<INa: lookup l senva = Some va>> /\
           <<INb: lookup (inl (inl l)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (SORD: fvars G = map fst senva)

    
    (RDV: forall x (IN: In (inl (inr x)) G), exists va vb,
           <<INa: lookup x denva = Some va>> /\
           <<INb: lookup (inl (inr x)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (ENV: reprlist (tm_lhp s) (tm_rf s r_env) (map snd denva))
    (DORD: tvars G = map fst denva)

    
    laddr n
    (NZ: laddr <> 0) (Eladdr: exists h, tm_lhp s = {laddr |=> ea n} ## {laddr - n |=> map snd senva} ## h)

    
    v

    
    ka (Ka: ka = length (ea n) + laddr)
    x kb j
    (KSIM: forall s' (EXT: extpub (wlift mylw) s' s)
                  
                  (ARG: tm_rf s' r_arg = tm_rf s r_arg)
                  va vb (Va: pmread (tm_lhp s') (tm_rf s' r_arg) va)
                  
                  (VR: vclos (wlift mylw) s' (g s') va vb),
             esim (wlift mylw) j kia kib g si s' true
                  {| cpc := Just ka; crf := Nothing; cst := pmemp; chp := pmemp |}
                  (pmemp, Just ((inl (inr x), vb) :: envb, kb)))

    csa csb (CR: confrel (wlift mylw) g s csa csb),

  esim (wlift mylw) i kia kib g si s false
       (@cplus tgt_lsi {| cpc := Just laddr; crf := Nothing; cst := pmemp; chp := {tm_rf s r_arg |-> v} |} csa)
       (@cplus mid_lsi (pmemp, Just (envb, IElet eb (inl x) kb)) csb). 



(** * Proof utilities. *)

Lemma ksplit_app: forall envb envb'
    (KV: kvars (map fst envb') = nil),
  ksplit (envb' ++ envb) = ksplit envb.
Proof.
  i; ginduction envb'; i; auto.
  destruct a; destruct v; ss; eauto.
Qed.

Lemma ksplit_lookup: forall k k' v envb envb'
    (SPL: ksplit envb = Some (k', envb'))
    (NE: k <> k')
    (IN: lookup (inr k) envb' = Some v),
  lookup (inr k) envb = Some v.
Proof.
  i; ginduction envb; vauto.
  i; destructs a v0.
  - rewrite list_assoc_cons_neq; vauto.
    s in SPL; eauto.
  - s; des_if; clarify; s in SPL; clarify.
Qed.


Lemma tm_ksimD_sp: forall g si s kia kib i envb n ks
    (KSIM: tm_ksim g si s kia kib i envb n ks),
  2 * (n + length ks) <= tm_rf s r_sp.
Proof.
  i; ginduction KSIM; ss; omega.
Qed.

Lemma tm_ksimD: forall k ks g si s kia kib i envb n
    (IN: In k ks)
    (KSIM: tm_ksim g si s kia kib i envb n ks),
  exists ka kaenv kb,
    <<RET: pmread (tm_lst s) (tm_rf s r_sp - 2 * (1 + n + index k ks)) ka>> /\
    <<ENV: pmread (tm_lst s) (tm_rf s r_sp - 2 * (1 + n + index k ks) + 1) kaenv>> /\
    <<Kb: lookup (inr k) envb = Some kb>> /\
    <<ESIM: forall g' (LE: g <g= g')
                s' (EXT: (wlift mylw).(ext) s' s)         
                (SP: tm_rf s' r_sp = tm_rf s r_sp - 2 * (1 + n + index k ks))
                popped
                (LST: tm_lst s = tm_lst s' ## {tm_rf s' r_sp |=> popped})
                (LEN: length popped = 2 * (1 + n + index k ks))
                (ENV: tm_rf s' r_env = kaenv)
                va vb (RV: vclos (wlift mylw) s' (g' s') va vb)
                ea (RETa: cquerya (wlift mylw) s' (ret va ka) ea)
                eb (RETb: cqueryb (wlift mylw) s' (ret vb kb) eb),
              esim (wlift mylw) i kia kib g' si s' true ea eb>>.
Proof.                          
  i; ginduction KSIM; [by auto|]; i.
  apply in_consD in IN; des; subst.
  - simpl list_index; des_if; [|by vauto].
    simpl withd; rewrite !plus_0_r.
    esplits; eauto.
  - simpl list_index; des_if; [by vauto|].
    dup IN; eapply list_index_some in IN; des.
    hexploit IHKSIM; eauto; i; des.
    rewrite ->IN in *; simpl withd in *.
    rewrite plus_comm; simpl plus in *; rewrite plus_comm.
    esplits; eauto using ksplit_lookup.
Qed.


Lemma tm_ksim_mon: forall g g' si (s s': state (wlift mylw)) kia kib i i' envb envb' n ks
    (SIM: tm_ksim g si s kia kib i envb n ks)
    (EXT: ext _ s' s)
    (SP: tm_rf s' r_sp = tm_rf s r_sp)
    (ST: tm_lst s' = tm_lst s)
    (KV: kvars (map fst envb') = nil)
    (GLE: wfo_le i i')
    (LE: g <g= g'),
  tm_ksim g' si s' kia kib i' (envb' ++ envb) n ks.
Proof.
  i; ginduction SIM; i; first (by econs 1; rewrite SP0; eauto); econs 2.
  - rewrite ksplit_app; eauto.
  - hexploit (IHSIM g' s' i' nil); eauto.
  - rrdes EXT EXT0; rewrite SP ST; eauto. 
  - rrdes EXT EXT0; rewrite SP ST; eauto. 
  - rewrite list_assoc_app Kb; clear -KV; induction envb'0; [by s; eauto|].
    destructs a v; rewrite list_assoc_cons_neq; eauto; s in KV; vauto.
  - i; hexploit ESIM; first 2 last; eauto using ext_trans.
    + rewrite SP0 SP; eauto.
    + rewrite -ST; eauto.
    + i; eapply pesim_wfo_le; eauto.
    + eauto.
Qed.

Lemma tm_ksim_push: forall g si (s s': state (wlift mylw)) kia kib i envb n ks pushed p
    (SIM: tm_ksim g si s kia kib i envb n ks)
    (EXT: ext _ s' s)
    (SP: tm_rf s' r_sp = 2 * p + tm_rf s r_sp)
    (ST: tm_lst s' = {tm_rf s r_sp |=> pushed} ## tm_lst s)
    (LEN: length pushed = 2 * p)
    (VLD: pmvalid (tm_lst s')),
  tm_ksim g si s' kia kib i envb (p + n) ks.
Proof.
  i; ginduction SIM; i; first (by econs 1; rewrite SP0; omega); econs 2; eauto.
  - rewrite plus_assoc (plus_comm 1) -plus_assoc; eapply IHSIM; eauto.
  - rewrite SP ST; replace (_-_) with (fst (fst (fst (fst s))) r_sp - 2*(1+n))
    ; inst; [|s; omega].
    rewrite ->ST in *; eapply pmread_plus; eauto; rewrite ->pmplus_comm; auto.
  - rewrite SP ST; replace (_-_) with (fst (fst (fst (fst s))) r_sp - 2*(1+n))
    ; inst; [|s; omega].
    rewrite ->ST in *; eapply pmread_plus; eauto; rewrite ->pmplus_comm; auto.
  - i; eapply ESIM; eauto using ext_trans.
    + rewrite SP0 SP; omega.
    + rewrite ST in LST.
      assert (SPL' := firstn_skipn (2 * (1 + n)) popped).
      rewrite <-SPL', pmcells_app in LST.
      rewrite (pmplus_comm _ {_|=>_}) (pmplus_swap (fst (snd s'0))) in LST.
      assert (fst (fst (fst (fst s'0))) r_sp + length (firstn (2 * (1 + n)) popped) = fst (fst (fst (fst s))) r_sp).
      { rewrite SP0 SP firstn_length LEN0 min_l; last omega.
        hexploit tm_ksimD_sp; eauto. i.
        match goal with |- ?lhs = _ =>
           replace lhs with (2 * p + fst (fst (fst (fst s))) r_sp - 2*p - 2 * (1 + n) + 2 * (1 + n))
        end; inst; last (clear; s; omega).
        rewrite plus_minus_eq; [|omega]; rewrite minus_diag.
        rewrite plus_0_l -plus_minus_eq; [omega|].
        eapply le_trans; last first; eauto; omega. 
      }
      rewrite H in LST; rewrite ->ST in *.
      hexploit pmcells_uniq; eauto; first (rewrite LEN skipn_length LEN0; omega).
      i; des; eauto.
    + rewrite firstn_length LEN0 min_l; auto; omega.
Qed.













































(** ** Tactics. *)












Ltac tm_prepare' :=
  intro HR;                     
  match type of HR with confrel _ _ _ ?csa _ =>
    eapply pesim_real; intros ma ? RLa ?;
    destruct csa as [cpc ? ? ?], ma;
    destruct_tgt_real RLa;
    destruct cpc; [by vauto|clarify|by vauto];
    rrdes HR WHX WHX0; subst;
    hexploit tgt_confpred_shape; eauto;
    hexploit mid_confpred_shape; eauto;
    hexploit tm_refconfrel_shape; eauto;
    hexploit mylw_confrel_shape; eauto;
    i; des; clarify; noemp;
    simpl cplus in *; clarify; noemp; clean;
    rewrite <-!pmplus_assoc in *;
    rrdes WHY; clarify; clean
  end.



Ltac tm_prepare :=
  let HR := fresh in
  apply pesim_acquire; intros ? ? HR;
  revert HR; tm_prepare'.


(** ** Helper lemmas about variable lookup. *)

Lemma values_for_var: forall {G: list var} {senva denva}
    {envb: list ((label + termvar + contvar) * lang_spec.val mid_lsi)}
    {g s lx}
    (RSV: forall l (IN: In (inl (inl l)) G), exists va vb,
           <<INa: lookup l senva = Some va>> /\
           <<INb: lookup (inl (inl l)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (RDV: forall x (IN: In (inl (inr x)) G), exists va vb,
           <<INa: lookup x denva = Some va>> /\
           <<INb: lookup (inl (inr x)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (IN: In (inl lx) G),
  exists va vb,
    <<INa: match lx with
             | inl l => lookup l senva = Some va
             | inr x => lookup x denva = Some va
           end>> /\
    <<INb: lookup (inl lx) envb = Some vb>> /\
    <<VCL: vclos (wlift mylw) s (g s) va vb>>.
Proof.
  i; destruct lx; ss; eauto.
Qed.

Lemma values_for_varnat: forall {G: list var} {senva denva}
    {envb: list ((label + termvar + contvar) * lang_spec.val mid_lsi)}
    {g s lxn}
    (RSV: forall l (IN: In (inl (inl l)) G), exists va vb,
           <<INa: lookup l senva = Some va>> /\
           <<INb: lookup (inl (inl l)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (RDV: forall x (IN: In (inl (inr x)) G), exists va vb,
           <<INa: lookup x denva = Some va>> /\
           <<INb: lookup (inl (inr x)) envb = Some vb>> /\
           <<SIM: vclos (wlift mylw) s (g s) va vb>>)
    (IN: Ini lxn G),
  exists va vb,
    <<INa: match lxn with
             | inl (inl l) => lookup l senva = Some va
             | inl (inr x) => lookup x denva = Some va
             | inr n => n = va
           end>> /\
    <<INb: ilookup lxn envb = Some vb>> /\
    <<VCL: vclos (wlift mylw) s (g s) va vb>>.
Proof.
  i; destruct lxn; try destruct i; ss; eauto.
  esplits; eauto; rewrite plus_0_r.
  econs; eapply vclgen_nat; last first; s; eauto.
Qed.

Lemma getVal_length: forall xs,
   exists n, forall lx m senva reg, length (getVal lx m senva xs reg) <= n.
Proof.
  i; exists (3 + length xs); i; destruct lx; s.
  - omega.
  - rewrite app_length repeat_length plus_comm; s.
    destruct (list_index _ _) eqn:E; s.
    + eapply list_index_length in E; omega.
    + destruct (length _); omega.
Qed.

Lemma pesim_getVal_termvar: forall (x: termvar) senva o denva hp st pc va G reg eb
    (r: esimtyp (wlift mylw)) i i' kia kib g si s rf
    (IN: lookup x denva = Some va)
    (COR: tvars G = map fst denva)
    (ENV: reprlist hp (rf r_env) (map snd denva))
    (NZ: pc <> 0)
    (CODE: forall w v (RD: pmread { pc |=> map enc (getVal (inr x) o senva (tvars G) reg) } w v),
             pmread hp w v)
    (LE: wfo_le_n (length (getVal (inr x) o senva (tvars G) reg)) i' i)
    (SIM: pesim r i' kia kib g si s false
                {| cpc := Just (pc + length (getVal (inr x) o senva (tvars G) reg));
                   crf := Just [va//reg]>>rf; cst := st; chp := hp |} eb),
  pesim r i kia kib g si s false
        {| cpc := Just pc; crf := Just rf; cst := st; chp := hp |} eb.
Proof.
  i; clean; des; subst.

  
  eapply pesim_real; intros ma ? RLa ?.
  destruct ma; destruct_tgt_real RLa; clarify.

  eapply pesim_dance; first (by eauto); clear RLa0.
  { clear SV HV FIN.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld]; [|by eauto].
      eapply pmread_plus; first (apply CODE); eauto.
      eapply pmread_plus; eauto.
      s; (rhs rewrite pmcells_cons); eauto.
    }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := depmach0); eauto.
  }

  
  revert ENV; generalize (mrf r_env) as myenv.
  pose (m := 0); change denva with (skipn m denva).
  replace (S mpc) with (S mpc + m); [|unfold m; omega].
  revert LE; simpl getVal; simpl length; rewrite app_length repeat_length; simpl length.
  replace (S (index x (tvars G) + 1)) with (S (index x (tvars G) + 1 - m))
  ; [|unfold m; omega].
  assert (MLE: myle m (index x (tvars G))) by (apply myle_le; auto using le_0_n).
  revert MLE i; generalize m; clear m.
  remember (index _ _) as N; intros ? MLE.
  induction MLE.

  { 
    intros ? ?.
    replace (S _) with 2 in LE; [|omega].
    apply wfo_le_n_S' in LE; des.
    i; rrdes XTRb RLb0; clarify.
    destruct eb; s in Ec; noemp; clarify.
    esplits; first 1 last.
    - right; split; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
    - i; frames_nodep RLa cfas; simpl plus in RLa.
      destruct ea'.
      esplits; eauto.
      eapply pesim_dance; first (by eauto); clear RLa.
      + 
        clear SV HV FIN FIN0.
        rewrite plus_comm.
        dance_init.
        { set (n := index _ _) in *.
          eapply Sok; [i; ss; omega|eapply Sld]; fold n.
          - eapply pmread_plus; first (apply CODE); eauto.
            eapply pmread_plus; eauto.
            s; fold n; rewrite map_app app_comm_cons pmcells_app.
            simpl length; rewrite repeat_map repeat_length.
            rhs rewrite ->plus_comm, ->?(pmcells_cons mpc), <-?pmplus_assoc; s.
            rewrite ->?pmcells_cons; pmplus_eq'.
            
            
          - dup IN; apply list_assoc_someD in IN.
            apply (in_map fst) in IN; clean; unfold n in ENV.
            eapply list_index_some in IN; des; rewrite COR IN in ENV.
            eapply skipn_list_assoc in IN; [|by eauto]; des.
            rewrite IN in ENV; s in ENV; des.
            eapply ev_hind; tmext_simpl; s.
            eapply pmread_plus; eauto.
        }
        { depmcplus0. }
        eapply dance_O; instantiate (1 := fun _ => _); s; eauto.
      + 
        i; rrdes XTRb RLb0; clarify.
        s in Ec; noemp; clarify.
        rewrite plus_comm in SIM.
        esplits; first 2 last; last first.
        * right; split; eauto.
        * i; frames_nodep RLa cfas.
          destruct ea'.
          clear SV HV FIN.
          destruct_tgt_real RLa.
          esplits; eauto.
          s in SIM; rewrite app_length repeat_length in SIM.
          s in SIM; rewrite -(plus_comm 1) in SIM; s in SIM.
          rewrite tmext_shadow; eauto.
        * rr; esplits; s; eauto.
  }
  { 
    i; rrdes XTRb RLb0; clarify.
    destruct eb, ee; s in Ec; noemp; clarify.
    inv LE.                     
    esplits; last first.
    - inst_pairs; s; noemp; rr; esplits; eauto.
    - right; split; eauto.
    - i; frames_nodep RLa cfas.
      destruct ea'.
      esplits; eauto.
      dup MLE; rewrite COR in MLE.
      eapply myle_le, skipn_cons in MLE; eauto; []; des.
      des; rewrite MLE in ENV; s in ENV; des.
      eapply pesim_dance; first (by eauto); clear RLa.
      + 
        clear SV HV FIN FIN0.
        dance_init.
        { eapply Sok; [by vauto|eapply Sld].
          - eapply pmread_plus; first (apply CODE); eauto.
            eapply pmread_plus; eauto.
            s; set (m := index _ _).
            rewrite map_app app_comm_cons pmcells_app.
            rewrite (@repeat_split _ _ (S n)); [|by apply myle_le; eauto].
            rewrite map_app repeat_map repeat_S.
            rewrite pmcells_cons 2!pmcells_app app_length repeat_length; s.
            rewrite ->?pmcells_cons; pmplus_eq'.
          - eapply ev_hind; tmext_simpl; s.
            eapply pmread_plus; eauto.
        }
        { depmcplus0. }
        eapply dance_O; instantiate (1 := fun _ => _); s; eauto.
      + i; hexploit IHMLE; eauto.
        * revert GLE; rewrite plus_comm; simpl plus; i; simpl minus.
          rewrite minus_Sn_m; eauto.
          apply myle_le; eauto.
        * i; nbdes; esplits; eauto.
          i; eapply H; eauto.
          s; s in RLa; rewrite plus_comm; s; rewrite plus_comm.
          rewrite tmext_shadow in RLa; eauto.
  }
Qed.


Lemma list_assoc_index: forall X {M: EqDec X} Y {N: EqDec Y} eqdec x y (l: list (X * Y))
    (IN: list_assoc eqdec x l = Some y),
  exists i,
    <<FST: list_index x (map fst l) = Some i>> /\
    <<ALL: list_index (x, y) l = Some i>>.
Proof.
  i; ginduction l; vauto.
  i; destruct a; ss; des_ifs; eauto.
  hexploit IHl; eauto; []; i; des.
  rewrite FST ALL; s; eauto.
Qed.























Lemma pesim_getVal: forall (lx: itermvar) o senva denva hp st pc va G reg eb
    (r: esimtyp (wlift mylw)) i i' kia kib g si s rf
    (IN: match lx with
           | inl l => lookup l senva = Some va
           | inr x => lookup x denva = Some va
         end)
    (COR: tvars G = map fst denva)
    (SENV: forall w v (RD: pmread {pc - o |=> map snd senva} w v),
             pmread hp w v)
    (DENV: reprlist hp (rf r_env) (map snd denva))
    (NZ: pc <> 0)
    (CODE: forall w v (RD: pmread {pc |=> map enc (getVal lx o (map fst senva) (tvars G) reg)} w v),
             pmread hp w v)
    (LE: wfo_le_n (length (getVal lx o (map fst senva) (tvars G) reg)) i' i)
    (SIM: pesim r i' kia kib g si s false
                {| cpc := Just (pc + length (getVal lx o (map fst senva) (tvars G) reg));
                   crf := Just [va//reg]>>rf; cst := st; chp := hp |} eb),
  pesim r i kia kib g si s false
        {| cpc := Just pc; crf := Just rf; cst := st; chp := hp |} eb.
Proof.
  i; destruct lx; eauto using pesim_getVal_termvar.
  unfold getVal in CODE, LE, SIM; s in LE; s in SIM.

  eapply pesim_real; intros ma ? RLa ?.
  destruct ma; destruct_tgt_real RLa; clarify.

  eapply pesim_dance; first (by eauto); clear RLa0.
  { clear SV HV FIN.

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc].
      eapply pmread_plus; first (apply CODE); eauto.
      eapply pmread_plus; eauto.
      unfold mpc; rewrite pmcells_cons; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; last by eauto.
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        rewrite !pmcells_cons; pmplus_eq'.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        rewrite !pmcells_cons; pmplus_eq'.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; first (apply SENV); eauto.
        dup IN; eapply list_assoc_index in IN; des.
        rewrite FST; s.
        apply list_index_split in ALL; des.
        eapply pmread_plus; eauto.
        (rhs rewrite E map_app); s.
        rhs rewrite pmcells_app pmcells_cons map_length LEN plus_comm.
        rhs rewrite pmplus_swap.
        pmplus_eq'.
    }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ _ => depmach0); eauto.
  }

  
  i; rrdes XTRb RLb0; clarify.
  destruct eb; s in Ec; noemp; clarify.
  esplits; first 2 last; last first; first by eauto.
  - i; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    rewrite tmext_shadow; rewrite plus_comm in SIM; esplits; eauto.
  - rr; esplits; s; eauto.
Qed.

Ltac do_getVal y Eladdr LEN SORD :=
  eapply (pesim_getVal y)
  ; [ by eauto
    | by eauto
    |
    | by tmext_simpl; eapply reprlist_mon; eauto
      ; i; eapply pmread_plus; eauto; pmplus_eq'
    | auto; omega
    | by i; eapply pmread_plus; eauto
      ; rewrite Eladdr !map_app !pmcells_app SORD ?map_length; pmplus_eq'
    | by eapply wfo_le_n_anti; [|by eapply LEN]
      ; eapply wfo_le_n_wfo_le; [by eapply wfo_le_n_add|by eapply wfo_add_le1]
    |].

Lemma pesim_getValNat: forall (lxn: itermvar + nat) o senva denva hp st pc va G reg eb
    (r: esimtyp (wlift mylw)) i i' kia kib g si s rf
    (IN: match lxn with
           | inl (inl l) => lookup l senva = Some va
           | inl (inr x) => lookup x denva = Some va
           | inr n => n = va
         end)
    (COR: tvars G = map fst denva)
    (SENV: forall w v (RD: pmread {pc - o |=> map snd senva} w v),
             pmread hp w v)
    (DENV: reprlist hp (rf r_env) (map snd denva))
    (NZ: pc <> 0)
    (CODE: forall w v (RD: pmread { pc |=> map enc (getValNat lxn o (map fst senva) (tvars G) reg) } w v),
             pmread hp w v)
    (LE: wfo_le_n (length (getValNat lxn o (map fst senva) (tvars G) reg)) i' i)
    (SIM: pesim r i' kia kib g si s false
                {| cpc := Just (pc + length (getValNat lxn o (map fst senva) (tvars G) reg));
                   crf := Just [va//reg]>>rf; cst := st; chp := hp |} eb),
  pesim r i kia kib g si s false
        {| cpc := Just pc; crf := Just rf; cst := st; chp := hp |} eb.
Proof.
  i; destruct lxn; [by eauto using pesim_getVal|].

  simpl getVal in CODE, LE, SIM; rewrite IN in CODE, LE, SIM.
  s in CODE; s in LE; s in SIM.

  eapply pesim_real; intros ma ? RLa ?.
  destruct ma; destruct_tgt_real RLa; clarify.

  eapply pesim_dance; first (by eauto); clear RLa0.
  { clear SV HV FIN.
    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        unfold mpc; rewrite pmcells_cons; eauto.
      - eauto.
    }
    { depmcplus0. }
    eapply dance_O; instantiate (1 := depmach0); s; eauto.
  }

  
  i; rrdes XTRb RLb0; clarify.
  destruct eb; s in Ec; noemp; clarify.
  esplits; first 2 last; last first; first by eauto.
  - i; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    rewrite plus_comm in SIM; esplits; eauto. 
  - rr; esplits; s; eauto.
Qed.

Lemma pesim_getCont: forall (k: contvar) hp st pc G eb kpc kenv sp
    (r: esimtyp (wlift mylw)) i i' kia kib g si s rf
    (NZ: pc <> 0)
    (CODE: forall w v (RD: pmread {pc |=> map enc (getCont k (kvars G))} w v),
             pmread hp w v)
    (LE: wfo_le_n (length (getCont k (kvars G))) i' i)
    (RET: pmread st (rf r_sp - 2 * (1 + index k (kvars G))) kpc)
    (ENV: pmread st (1 + (rf r_sp - 2 * (1 + index k (kvars G)))) kenv)
    (SP: rf r_sp = 2 * (1 + index k (kvars G)) + sp)
    (SIM: pesim r i' kia kib g si s false
                {| cpc := Just (pc + length (getCont k (kvars G)));
                   crf := Just [kenv//r_env]>>[kpc//r_ret]>>[sp//r_sp]>>rf; cst := st; chp := hp |} eb),
  pesim r i kia kib g si s false
        {| cpc := Just pc; crf := Just rf; cst := st; chp := hp |} eb.
Proof.
  i; eapply pesim_real; intros ma ? RLa ?.
  destruct ma; destruct_tgt_real RLa; clarify; s in CODE.
  eapply pesim_dance; first (by eauto); clear RLa0.

  - clear SV HV FIN; dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; last by eauto.
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite pmcells_cons); eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite !pmcells_cons); eauto.
        (rhs rewrite pmplus_swap); eauto.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; [eapply RET| |]; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite !pmcells_cons); eauto.
        (rhs rewrite -!(pmplus_swap {S (S mpc0) |-> _})); eauto.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; [eapply ENV| |]; eauto.
    }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ _ _ => _); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    destruct eb; s in Ec; noemp; clarify.
    esplits; first 2 last; last first; first by eauto.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      rewrite plus_comm in SIM; esplits; eauto.
      replace (_ - _) with sp; eauto; omega.
    + rr; esplits; s; eauto.
Qed.

Lemma tm_pesim_return: forall (lw: world_loca tm_gw) (r: esimtyp (wlift lw))
   i ka kb vb g si s env xk xv
   (K: lookup (inr xk) env = Some kb)
   (V: lookup (inl xv) env = Some vb)
   (RV: vclos (wlift lw) s (g s) (tm_rf s r_arg) vb)
   (EXT: extpub _ s si),
 pesim r i ka kb g si s true
   {| cpc := Just ka; crf := Nothing; cst := pmemp; chp := pmemp |}
   (pmemp, Just (env, IEret xk xv)).
Proof.
  i; pfold; econs; i; esplits; eauto.
  eapply esim_ret; eauto; rr; eauto.
Qed.


(** Inversion for value closure.

    Note that we _cannot_ have a generic version of this, because the
    language spec does not require the various value queries to be
    disjoint (for a good reason, namely assembly).
*)
Lemma tm_vclosDb: forall (w: world_loca tm_gw) s r va vb
    (VCL: vclos (wlift w) s r va vb),
  match vb with
    | IVunit =>
        vquerya tm_gw (fst s) valid_unit va
    | IVnat n =>
        vquerya tm_gw (fst s) (valid_nat n) va
    | IVpair vb' vb'' => exists va' va'',
        vquerya tm_gw (fst s) (valid_pair va' va'') va /\
        <<VCL': vclos (wlift w) s r va' vb'>> /\
        <<VCL'': vclos (wlift w) s r va'' vb''>>
    | IVinl vb' => exists va',
        vquerya tm_gw (fst s) (valid_inl va') va /\
        <<VCL': vclos (wlift w) s r va' vb'>>
    | IVinr vb' => exists va',
        vquerya tm_gw (fst s) (valid_inr va') va /\
        <<VCL': vclos (wlift w) s r va' vb'>>
    | IVloc _ =>
        rquery tm_gw (fst s) va vb
    | IVclo envb (IEfix xfb xvb xkb eb) =>
        <<VQ: vquerya tm_gw (fst s) valid_fun va>> /\
        <<RV: r va vb>>
    | IVclo envb (IEgen xkb eb) =>
        <<VQ: vquerya tm_gw (fst s) valid_gen va>> /\
        <<RV: r va vb>>
    | _ => True
  end.
Proof.
  i; destruct vb eqn:E; rewrite <-E in VCL
  ; move VCL at top; revert_until VCL; pattern va, vb
  ; apply (vclos_induction_strong VCL); i
  ; try (by s; eauto)
  ; invc IH
  ; try (by invc CB)
  ; try (by invc CA CB; des; eauto)
  ; try (by invc VR; des; vauto)
  ; try (by invc VR CB)
  ; try (by invc CB; des; esplits; ss; eauto)
  ; try (by des; destruct e; auto; invc CB; specialize (VR _ _ eq_refl)
         ; ss; subst; des; eauto).
Qed.


(** Convenience lemmas for reasoning about reference allocation. *)

Lemma tm_refconfrel_newa: forall {g mrf P0 s s3 refha refhb la va}
    (RCR: tm_refconfrel g (mrf, P0, s, s3)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := refha |}
            (refhb, Nothing))
    (VLD: pmvalid ({la |-> va} ## refha)),
  forall lb', ~ s3 la lb'.
Proof.
  i; rrdes RCR; clarify.
  apply RELV in FH; des; rrdes RDa; subst.
  rewrite pmplus_assoc in VLD; apply pmvalid_plus_l in VLD.
  apply pmvalid_cell_cell in VLD; vauto.
Qed.

Lemma tm_refconfrel_newb: forall {g mrf P0 s s3 refha refhb lb vb}
    (RCR: tm_refconfrel g (mrf, P0, s, s3)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := refha |}
            (refhb, Nothing))
    (VLD: pmvalid ({lb |-> vb} ## refhb)),
  forall la', ~ s3 la' lb.
Proof.
  i; rrdes RCR; clarify.
  apply RELV in FH; des; rrdes RDb; subst.
  rewrite pmplus_assoc in VLD; apply pmvalid_plus_l in VLD.
  apply pmvalid_cell_cell in VLD; vauto.
Qed.

Lemma tm_refconfrel_alloc: forall g mrf P0 s s3 s4 ha hb va vb la lb mrf' P0' s' s3' s4'
    (RCR: tm_refconfrel (fun sg => g (sg, s4)) (mrf, P0, s, s3)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := ha |}
            (hb, Nothing))
    (VCL: vclos (wlift mylw) (mrf, P0, s, s3, s4) (g (mrf, P0, s, s3, s4)) va vb)
    (EXT: ext (wlift mylw) (mrf', P0', s', s3', s4') (mrf, P0, s, s3, s4))
    (GK: gknow (wlift mylw) g)
    (ALC: refst_alloc s3 la lb = s3')
    (VLDa: pmvalid ({la |-> va} ## ha))
    (VLDb: pmvalid ({lb |-> vb} ## hb)),
  tm_refconfrel (fun sg => g (sg, s4')) (mrf', P0', s', s3')
    {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va} ## ha |}
    ({lb |-> vb} ## hb, Nothing).
Proof.
  unfold refst_alloc; i.
  rrdes RCR; clarify; clean.
  rr; esplits; eauto; last first; clean.
  - rewrite -ALC; i; des.
    + hexploit RELV; eauto; i; des; esplits
      ; try by (eapply pmread_plus; eauto; pmplus_eq').
      dup EXT; rrdes EXT; eapply vclos_mon; eauto using (gk_mon GK).
    + subst; esplits
      ; try by (eapply pmread_plus; eauto; pmplus_eq').
      eapply vclos_unlift in VCL.
      dup EXT; rrdes EXT; eapply vclos_mon; eauto using (gk_mon GK).
  - rewrite -ALC; i; apply pmread_plusD in ALC0; des.
    + apply pmread_cellD in ALC0; des; s; eauto.
    + hexploit DOMb; eauto; i; des; s; eauto.
  - rewrite -ALC; i; apply pmread_plusD in ALC0; des.
    + apply pmread_cellD in ALC0; des; s; eauto.
    + hexploit DOMa; eauto; i; des; s; eauto.
Qed.


(** Convenvience lemmas for reasoning about reference writing. *)

Lemma tm_refconfrel_write: forall (g : state (wlift mylw) -> vrelf) s la lb va vb ha hb
    (RCR: tm_refconfrel (fun sg => (g (sg, snd s))) (fst s)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va} ## ha |}
            ({lb |-> vb} ## hb, Nothing))
    (RL: snd (fst s) la lb)
    va' vb' (VCL: vclos tm_gw (fst s) (g s) va' vb'),
  tm_refconfrel (fun sg => (g (sg, snd s))) (fst s)
    {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va'} ## ha |}
    ({lb |-> vb'} ## hb, Nothing).
Proof.
  i; rrdes RCR; clarify; rr; esplits; eauto using pmvalid_plus_cell.
  - i; dup ALC; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; subst; eauto.
    + hexploit DOMa; [|by eauto].
      rrdes ALC0; eapply pmvalid_plus_cell in VLD.
      eapply pmread_plus; eauto; pmplus_eq'.
  - i; dup ALC; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; subst; eauto.
    + hexploit DOMb; [|by eauto].
      rrdes ALC0; eapply pmvalid_plus_cell in VLD.
      eapply pmread_plus; eauto; pmplus_eq'.
  - i; dup REG; apply RELV in REG; des.
    dup RDa; rrdes RDa; dup RDb; rrdes RDb.
    apply pmread_plusD in RDa0; apply pmread_plusD in RDb0; des.
    + apply pmread_cellD in RDa0; apply pmread_cellD in RDb0; des; subst.
      rewrite -prod_eta; esplits; last (exact VCL)
      ; eauto using pmread_plus, pmread_cell, pmvalid_plus_cell.
    + exfalso.
      apply pmread_cellD in RDb0; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X0 eq_refl). subst.
      rrdes RDa0; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLD.
      apply pmvalid_plus_l in VLD.
      apply pmvalid_cell_cell in VLD; vauto.
    + exfalso.
      apply pmread_cellD in RDa0; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X eq_refl). subst.
      rrdes RDb0; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLD0.
      apply pmvalid_plus_l in VLD0.
      apply pmvalid_cell_cell in VLD0; vauto.
    + eauto 10 using pmread_plus, pmplus_comm, pmvalid_plus_cell.
Qed.
