(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import vgtac basis.
Require Import lang_src_lsi lang_mid lang_tgt lang_spec.
Require Import model_common.

Set Implicit Arguments. 



(** * Definitions for global worlds involving the src language. *)

Program Definition src_gw_sts : transys :=
  {| state := lang_src.env
   ; ext s' s := s' = s
   ; extpub s' s := s' = s
  |}.

Inductive src_vquery' : @vquery src_lsi -> lang_src.exp -> Prop :=
| src_vquery_unit
  : src_vquery' valid_unit Eunit
| src_vquery_nat n
  : src_vquery' (valid_nat n) (Enat n)
| src_vquery_pair (va vb: src_lsi.(val))
  : src_vquery' (valid_pair va vb) (Epair va vb)
| src_vquery_inl (v: src_lsi.(val))
  : src_vquery' (valid_inl v) (Einl v)
| src_vquery_inr (v: src_lsi.(val))
  : src_vquery' (valid_inr v) (Einr v)
| src_vquery_roll (v: src_lsi.(val))
  : src_vquery' (valid_roll v) (Eroll v)
| src_vquery_pack (v: src_lsi.(val))
  : src_vquery' (valid_pack v) (Epack v)
| src_vquery_fun xf xa e
  : src_vquery' valid_fun (Efix xf xa e)
| src_vquery_gen e
  : src_vquery' valid_gen (Egen e)
| src_vquery_name (v: src_lsi.(val))
  : src_vquery' valid_name v
| src_vquery_good_fun (v: src_lsi.(val))
  : src_vquery' good_fun v
| src_vquery_good_gen (v: src_lsi.(val))
  : src_vquery' good_gen v
.
Hint Constructors src_vquery'.

Definition src_vquery (q: @vquery src_lsi) (v: src_lsi.(val)) : Prop :=
  <<SVQ: src_vquery' q v>> /\ <<CLO: forall a, ~ fv_exp v a>>.


Inductive src_cquery : @cquery src_lsi -> src_lsi.(conf) -> Prop :=
| src_cquery_app v f x e e' a k
    (VAL: e_of v = Efix f x e)
    (EXP: e' = app_cont k (subst_exp [e_of a//x]>>[e_of v//f]>>vid e))
  : src_cquery (@app src_lsi v a k) (pmemp, Nothing, Just e')
| src_cquery_inst v e e' k
    (VAL: e_of v = Egen e)
    (EXP: e' = app_cont k e)
  : src_cquery (@inst src_lsi v k) (pmemp, Nothing, Just e')
| src_cquery_ret e' (v: src_lsi.(val)) k
    (EXP: e' = app_cont k v)
  : src_cquery (@ret src_lsi v k) (pmemp, Nothing, Just e')
.
Hint Constructors src_cquery.


Definition src_confpred (s: src_gw_sts.(state)) (c: src_lsi.(conf)) :=
  c = (pmemp, Just s, Nothing).

Lemma src_confpred_shape: forall s c
    (CP: src_confpred s c),
  c = (pmemp, Just s, Nothing).
Proof.
  auto.
Qed.



(** * Definitions for global worlds involving the mid language. *)

Program Definition mid_gw_sts : transys :=
  {| state := unit
   ; ext s' s := True
   ; extpub s' s := True
  |}.

Inductive mid_vquery (s: mid_gw_sts.(state)) : @vquery mid_lsi -> mid_lsi.(val) -> Prop :=
| mid_vquery_unit
  : mid_vquery s valid_unit IVunit
| mid_vquery_nat n
  : mid_vquery s (valid_nat n) (IVnat n)
| mid_vquery_pair va vb
  : mid_vquery s (valid_pair va vb) (IVpair va vb)
| mid_vquery_inl v
  : mid_vquery s (valid_inl v) (IVinl v)
| mid_vquery_inr v
  : mid_vquery s (valid_inr v) (IVinr v)
| mid_vquery_roll v
  : mid_vquery s (valid_roll v) v
| mid_vquery_pack v
  : mid_vquery s (valid_pack v) v
| mid_vquery_fun env bf bx bk e
  : mid_vquery s valid_fun (IVclo env (IEfix bf bx bk e))
| mid_vquery_gen env bk e
  : mid_vquery s valid_gen (IVclo env (IEgen bk e))
| mid_vquery_name n
  : mid_vquery s valid_name (IVclo nil (IEnat n))
  (** Something that is disjoint from other valid values and
      that leads to an error when used.
   *)
| mid_vquery_good_fun env bf bx bk e (NB: ~ bad_fun (IVclo env (IEfix bf bx bk e)))
  : mid_vquery s good_fun (IVclo env (IEfix bf bx bk e))
| mid_vquery_good_gen env bk e (NB: ~ bad_gen (IVclo env (IEgen bk e)))
  : mid_vquery s good_gen (IVclo env (IEgen bk e))
.
Hint Constructors mid_vquery.

Inductive mid_cquery (s: mid_gw_sts.(state)) : @cquery mid_lsi -> mid_lsi.(conf) -> Prop :=
| mid_cquery_app env bf bx bk e vx vk
  : mid_cquery s (@app mid_lsi (IVclo env (IEfix bf bx bk e)) vx vk)
               (pmemp, Just ((inr bk, vk) ::
                             (inl (inr bx), vx) ::
                             (inl (inr bf), IVclo env (IEfix bf bx bk e)) :: env, e))
| mid_cquery_inst env bk e vk
  : mid_cquery s (@inst mid_lsi (IVclo env (IEgen bk e)) vk)
               (pmemp, Just ((inr bk, vk) :: env, e))
| mid_cquery_ret v k env xk xv
    (GETk: lookup (inr xk) env = Some k)
    (GETv: lookup (inl xv) env = Some v)
  : mid_cquery s (@ret mid_lsi v k) (pmemp, Just (env, IEret xk xv))
.
Hint Constructors mid_cquery.


Definition mid_confpred (s: mid_gw_sts.(state)) (c: mid_lsi.(conf)) :=
  c = cemp.

Lemma mid_confpred_shape: forall s c
    (CP: mid_confpred s c),
  c = cemp.
Proof.
  auto.
Qed.



(** * Definitions for global worlds involving the tgt language. *)

Inductive tquery_state :=
| ts_pair (va: tgt_lsi.(val)) (vb: tgt_lsi.(val))
| ts_inl  (v: tgt_lsi.(val))
| ts_inr  (v: tgt_lsi.(val))
| ts_fun  (v: tgt_lsi.(val))    
.

Definition tgt_gw_sts : transys.
  refine {| state := regfile * (tquery_state -> tgt_lsi.(val) -> Prop)
          ; ext s' s := forall q v (IN: snd s q v), snd s' q v
          ; extpub s' s := 
            <<RF: fst s r_sp = fst s' r_sp /\ fst s r_env = fst s' r_env>> /\
            <<DB: forall q v (IN: snd s q v), snd s' q v>>
         |}; i; des; eauto.
  i; destruct x, y, z; des; ss; splits; try rewrite RF; try rewrite RF0; eauto.
Defined.

Definition tgt_vquery (s: tgt_gw_sts.(state)) (q: @vquery tgt_lsi) v : Prop :=
  match q with
    | valid_nat v'
    | valid_roll v'
    | valid_pack v' => v = v'
    | valid_unit
    | valid_name
    | good_fun
    | good_gen => True
    | valid_pair va vb => snd s (ts_pair va vb) v
    | valid_inl v' => snd s (ts_inl v') v
    | valid_inr v' => snd s (ts_inr v') v
    | valid_fun => exists v', snd s (ts_fun v') v
    | valid_gen => exists v', snd s (ts_fun v') v
  end.

Lemma tgt_vquery_mon:
  forall s q v (IN: tgt_vquery s q v) s' (EXT: ext _ s' s),
    tgt_vquery s' q v.
Proof.
  i; destruct q; ss; des; try eapply EXT; eauto.
Qed.

Inductive tgt_cquery (s: tgt_gw_sts.(state)) : @cquery tgt_lsi -> tgt_lsi.(conf) -> Prop :=
| tgt_cquery_app f v k pc
    (Ef: f = fst s r_clo)
    (Ev: v = fst s r_arg)
    (Ek: k = fst s r_ret)
    (PC: snd s (ts_fun pc) f)
  : tgt_cquery s (@app tgt_lsi f v k)
               {| cpc := Just pc; crf := Nothing; cst := pmemp; chp := pmemp |}
| tgt_cquery_inst g k pc
    (Eg: g = fst s r_clo)
    (Ek: k = fst s r_ret)
    (PC: snd s (ts_fun pc) (fst s r_clo))
  : tgt_cquery s (@inst tgt_lsi g k)
               {| cpc := Just pc; crf := Nothing; cst := pmemp; chp := pmemp |}
| tgt_cquery_ret v pc
    (Ev: v = fst s r_arg)
  : tgt_cquery s (@ret tgt_lsi v pc)
               {| cpc := Just pc; crf := Nothing; cst := pmemp; chp := pmemp |}
.
Hint Constructors tgt_cquery.


Definition repr (q: tquery_state) v h : Prop :=
  
  match q with
    | ts_pair v1 v2 => pmread h v v1 /\ pmread h (1+v) v2
    | ts_inl v1 => pmread h v 0 /\ pmread h (1+v) v1
    | ts_inr v1 => exists n, n <> 0 /\ pmread h v n /\ pmread h (1+v) v1
    | ts_fun v1 => pmread h v v1
  end.

Lemma repr_mon: forall hp hp' q
    (SUB: forall l v (IN: pmread hp l v), pmread hp' l v)
    v (REP: repr q v hp),
  repr q v hp'.
Proof.
  i; destruct q; ss; des; eauto 10.
Qed.

Definition tgt_confpred (s: tgt_gw_sts.(state)) (c: tgt_lsi.(conf)) :=
    <<PC: cpc c = Nothing>> /\
    <<RF: crf c = Just (fst s)>> /\
    <<ST: forall a, a >= fst s r_sp <-> exists v, pmread (cst c) a v>> /\
    <<HP: forall q v (IN: snd s q v), repr q v (chp c)>>.

Lemma tgt_confpred_shape: forall s c
    (CP: tgt_confpred s c),
  exists freest hp,
    c = {| cpc := Nothing; crf := Just (fst s); cst := freest; chp := hp |}.
Proof.
  i; destruct c; rrdes CP; ss; subst; eauto.
Qed.
 
Lemma tgt_vquery_fun_uniq: forall rf db h v wa wb
    (CP: tgt_confpred (rf, db) h)
    (FQy: db (ts_fun wa) v)
    (FQz: db (ts_fun wb) v),
  wa = wb.
Proof.
  i; rrdes CP; ss; assert (HP' := HP _ _ FQy); specialize (HP _ _ FQz).
  rrdes HP; rrdes HP'; rewrite RD0 in *; eapply (@pmcell_uniq nat); eauto.
Qed.                          
