(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Simple contification. *)

Require Import Arith Bool List Program Relations.
Require Import cps cps_correct countable vgtac basis lang_common lang_spec.
Require Import lang_mid lang_mid_renaming.
Require Import model_common model_untyped model_properties_untyped gw_untyped.
Require Import mm_test ssr mm opt_freshen pilsner_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".

Require opt_transform_new.
Module otf := opt_transform_new.

Definition teq X {A:EqDec X} x y :=
  match eqdec x y with
  | left _ => true
  | _ => false
  end.




Inductive is_app: itermvar -> exp -> Prop :=
  is_app_ : forall f v k, is_app f (IEapp f v k).


Fixpoint gather_all_apps (e : exp) (f : itermvar) : list (sig (is_app f)) :=
  match e return list (sig (is_app f)) with
  | IEcont bx ea => gather_all_apps ea f
  | IEnew x bx ea => gather_all_apps ea f
  | IEread x bx ea => gather_all_apps ea f

  | IEwrite x y ea => gather_all_apps ea f

  | IEinput bx ea => gather_all_apps ea f

  | IEoutput x ea => gather_all_apps ea f

  | IEifnz x ea eb
    => let apps1 := gather_all_apps ea f in
       let apps2 := gather_all_apps eb f in
       apps1 ++ apps2

  | IEcase x bx ea eb
    => let apps1 := gather_all_apps ea f in
       let apps2 := gather_all_apps eb f in
       apps1 ++ apps2

  | IEapp xf v k
    => if teq xf f then [exist _ (IEapp f v k) (is_app_ _ _ _)] else []

  | IElet ae bx ea
    => let apps1 := gather_all_apps ae f in
       let apps2 := gather_all_apps ea f in
       apps1 ++ apps2
  | _ => []
  end.


Ltac crush_obs :=
  program_simpl
  ; (match goal with
    | [pf: is_app ?f ?e |- _] =>
    destruct e; inversion pf; []
    ; (match goal with
       | [H: context[exist (is_app f) (IEapp ?xf ?x ?xk) ?pf :: ?apps']
         |- _ ] => 
       clarify; vauto; specialize (H xf x xk pf apps'); apply H; auto
       |  |- _ => idtac
       end)
     | |- _ => idtac
     end).

Local Obligation Tactic := crush_obs.


Program Fixpoint check_app_ret (f : itermvar)(apps : list (sig (is_app f)))
                               (k : contvar) : option contvar :=
  match apps with
  | nil => Some k
  | exist (IEapp _ _ k') _ :: apps' => if teq k' k then
                                 check_app_ret apps' k
                               else
                                 None
  | exist e pf :: apps' => False_rec _ _
  end.
Arguments check_app_ret : clear implicits.


Inductive all_apps_return_to (f : itermvar) :
  list (sig (is_app f)) -> contvar -> Prop :=
 | empty : forall k, all_apps_return_to nil k
 | ok : forall v k apps pf, all_apps_return_to apps k ->
                           all_apps_return_to
                            ((exist _ (IEapp f v k) pf)::apps) 
                            k.

Hint Constructors all_apps_return_to.


Lemma returns_same: forall f (apps : list (sig (is_app f))) k
    (CH: check_app_ret f apps k = Some k),
  all_apps_return_to apps k.
Proof.
  i; induction apps as [| t apps']; auto.
  destruct t as [e pf]; destruct e; inversion pf.
  simpl check_app_ret in CH; unfold teq in CH; destruct (eqdec _ _)
  ; vauto; auto.
Qed.


Program Definition is_contifiable (e : exp) (f : itermvar) : option contvar :=
  let apps := gather_all_apps e f in
  match apps with
  | nil => None
  | exist (IEapp _ _ k') _ :: apps' => check_app_ret f apps' k'
  | exist eb pf :: apps' => False_rec _ _
  end.

Fixpoint replace (f : termvar) (k fc : contvar) (e : exp) :=
  match e with
  | IEcont bx ea => IEcont bx (replace f k fc ea)
  | IEnew x bx ea => IEnew x bx (replace f k fc ea)
  | IEread x bx ea => IEread x bx (replace f k fc ea)

  | IEwrite x y ea => IEwrite x y (replace f k fc ea)

  | IEinput bx ea => IEinput bx (replace f k fc ea)

  | IEoutput x ea => IEoutput x (replace f k fc ea)

  | IEifnz x ea eb
    => IEifnz x (replace f k fc ea) (replace f k fc eb)

  | IEcase x bx ea eb
    => IEcase x bx (replace f k fc ea) (replace f k fc eb)

  | IEapp (inr xf) x xk
    => if teq xf f && teq xk k
       then IEret fc x
       else e

  | IElet ae bx ea
    => IElet ae bx (replace f k fc ea)
  | _ => e
  end.

Definition add_cont (k : contvar) (G: list var) (n: nat)
                    (e : exp) : option (exp * nat) :=
  match e with
  | IElet (IEfix bf y bk eb) (inl bx) ea
    =>
    let n' := next n in
    let (eb',n'') := freshen (next n') id id eb in  
    let ed := rename (ren_id[inr bx/bf][inr n/y][c k/bk]) eb' in
      if list_mem eqdec (tkvar2var (inr k)) G then  
        Some ((IElet (IEfix bf y bk eb) (inl bx)
                     (IElet (IEcont n ed) (inr n') ea)), n'')
      else None
  | _ => None
  end.

Fixpoint alpha_eq (e e' : exp) (ren: renaming) : bool :=
  match e, e' with
  | IElet ea (inl bx) eb, IElet ea' (inl bx') eb'
    => alpha_eq ea ea' ren && alpha_eq eb eb' ren[inr bx/bx']
  | IElet ea (inr bk) eb, IElet ea' (inr bk') eb'
    => alpha_eq ea ea' ren && alpha_eq eb eb' ren[c bk/bk']
  | IEfix bf bx bk ea, IEfix bf' bx' bk' ea'
    => alpha_eq ea ea' ren[inr bf/bf'][inr bx/bx'][c bk/bk']
  | IEgen bk ea,IEgen bk' ea'
    => alpha_eq ea ea' ren[c bk/bk']
  | IEcont bx ea,IEcont bx' ea'
    => alpha_eq ea ea' ren[inr bx/bx']
  | IEnew x y ea, IEnew x' y' ea'
  | IEread x y ea, IEread x' y' ea'
    => teq x (rename_itvar ren x') && alpha_eq ea ea' ren[inr y/y']
  | IEwrite x y ea, IEwrite x' y' ea'
    => teq x (rename_itvar ren x') && teq y (rename_itvar ren y') &&
           alpha_eq ea ea' ren
  | IEinput x ea, IEinput x' ea'
    => alpha_eq ea ea' ren[inr x/x']
  | IEoutput x ea, IEoutput x' ea'
    => teq x (rename_itvar ren x') && alpha_eq ea ea' ren
  | IEifnz x ea eb, IEifnz x' ea' eb'
    => teq x (rename_itvar ren x') && alpha_eq ea ea' ren &&
           alpha_eq eb eb' ren
  | IEcase x y ea eb, IEcase x' y' ea' eb'
    => teq x (rename_itvar ren x') && alpha_eq ea ea' ren[inr y/y'] &&
           alpha_eq eb eb' ren[inr y/y']
  | IEret k x, IEret k' x'
    => teq k (rename_contvar ren k') && teq x (rename_itvar ren x')
  | IEapp f x k, IEapp f' x' k'
    => teq f (rename_itvar ren f') && teq x (rename_itvar ren x') &&
           teq k (rename_contvar ren k')
  | IEinst x k, IEinst x' k'
    => teq x (rename_itvar ren x') && teq k (rename_contvar ren k')
  | IErefeq x y, IErefeq x' y'
    => teq x (rename_itvar ren x') && teq y (rename_itvar ren y')
  | IEbop Plus xn yn, IEbop Plus xn' yn'
  | IEbop Minus xn yn, IEbop Minus xn' yn'
  | IEbop Mult xn yn, IEbop Mult xn' yn'
  | IEbop Equal xn yn, IEbop Equal xn' yn'
    => let xb := match xn,xn' with
                 | inl x,inl x'
                   => teq x (rename_itvar ren x')
                 | inr n,inr n'
                   => beq_nat n n'
                 | _,_ => false
                 end in
       let yb := match yn,yn' with
                 | inl y,inl y'
                   => teq y (rename_itvar ren y')
                 | inr n,inr n'
                   => beq_nat n n'
                 | _,_ => false
                 end in xb && yb
  | IEfst x,IEfst x' => teq x (rename_itvar ren x')
  | IEsnd x,IEsnd x' => teq x (rename_itvar ren x')
  | IEinr x, IEinr x'
    => beq_nat (enum (IEinr x)) (enum (IEinr (rename_itvar ren x')))
  | IEinl x, IEinl x'
    => beq_nat (enum (IEinl x)) (enum (IEinl (rename_itvar ren x')))
  | IEpair x y, IEpair x' y'
    => beq_nat (enum (IEpair x y))
               (enum (IEpair (rename_itvar ren x') (rename_itvar ren y')))
  | IEnat n, IEnat m => beq_nat n m
  | IEunit,IEunit => true
  | _,_ => false
  end.


Definition subst_cont (k : contvar) (G: list var) (n: nat)
                      (e : exp) : option (exp * nat) :=
  match e with
  | IElet (IEfix bf y bk eb) (inl bx) (IElet (IEcont cx ed) (inr fc) ec)
    =>
    if alpha_eq ed eb ren_id[inr bx/bf][inr cx/y][c k/bk] then
      if list_mem eqdec (tkvar2var (inr k)) G then
        
        Some ((IElet (IEfix bf y bk eb) (inl bx)
                     (IElet (IEcont cx ed) (inr fc)
                            (replace bx k fc ec))), n)
      else None
    else None
  | _ => None
  end.



Lemma purge_wfexp: forall e G,
  wfexp (purge G) e -> wfexp G e.
Proof.
  i; eapply wfexp_sublist; [vauto | apply purge_sublist].
Qed.

Lemma purge_wfexp_x : forall e x xsa xsb,
  wfexp (xsa++x :: purge xsb) e -> wfexp (xsa++x :: xsb) e.
Proof.
  i; eapply wfexp_sublist; eauto using sublist_app.
Qed.

Lemma wfexp_in: forall e k G
    (IN: In (inr k) G)
    (WF: wfexp (inr k :: G) e),
  wfexp G e.
Proof.
  i; apply (in_split (inr k)) in IN; inversion IN as [G' H]
  ; inversion H as [G'' H'];
  rewrite <- (app_nil_l _); rewrite H'; apply wfexp_contract; s; rewrite <- H'
  ; assumption.
Qed.

Lemma replace_wf: forall e f k mx G
    (INmx: In (inr mx) G)
    (INk: In (inr k) G)
    (WF: wfexp G e),
  wfexp G (replace f k mx e).
Proof.
 induction e as [? IH] using (well_founded_ind size_lt_well_founded).
 i; destruct e; vauto; ss; des; intuition;
 try (destruct bx; des; splits; vauto);
 try (match goal with
      | [|- context[replace _ _ _ ?X]]
        => hexploit (IH X); last first; try omega; eauto using in_cons
           ; i; eapply in_inv in FH; des; eauto; vauto
      end).
 - destruct xf; vauto; destruct teq, teq; s; splits; vauto.
Qed.

Ltac des_max :=
 repeat match goal with
           | |- context[max ?a ?b] =>
             let H := fresh in
             destruct (Max.max_dec a b) as [H|H]; rewrite ->H in *
             ; apply NPeano.Nat.eq_le_incl in H
             ; rewrite ->!NPeano.Nat.max_lub_iff in H; des; clear H; try omega
           end.

Lemma within_maxbvar: forall n n' e
    (GT: 0 < n')
    (WI: within n n' (map varid (bvars e))),
  otf.maxbvar e < n'.
Proof.
  ii; unfold within in *; induction e
  ; unfold otf.nextbvar; rewrite otf.maxbvar_r_rewrite; otf.svbv; s
  ; try omega
  ; try (try (destruct bx; s; rewrite ->!map_app in *)
         ; rewrite ->fold_right_app in *; rewrite <-otf.fold_within in *
         ; try (apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm
                 || apply NPeano.Nat.max_idempotent))
  ; try (by des_max; ss; try (by apply lt_le_S; apply WI1; tauto)
         ; rewrite <-otf.maxbvar_r_rewrite; unfold otf.nextbvar in *
         ; apply IHe; splits; firstorder)
  ; try (by des_max; ss; try (by apply lt_le_S; apply WI1; tauto)
         ; rewrite <-otf.maxbvar_r_rewrite; unfold otf.nextbvar in *
         ; try (rewrite ->!map_app in *)
         ; try (by (apply IHe1 || apply IHe2); splits; firstorder)).
Qed.

Lemma subuniq_lb_within: forall n n' xs
    (UN: uniq xs)
    (WI: within n n' xs),
  subuniq xs (otf.list_between n n').
Proof.
  ii; unfold within in WI; des.
  apply uniq_appD in UN0; des.
  apply uniq_app; splits; auto.
  specialize (WI0 _ INb); specialize (WI1 _ INb).
  apply UN2 in INa; ins.
  rewrite otf.list_between_in; omega.
Qed.

Lemma contify_add_wf: forall e e' k G n n'
    (E: add_cont k G n e = Some (e',n'))
    (MX: otf.nextbvar e <= n)
    (WF: wfexp G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: otf.vars2nat (bvars e) ++ otf.vars2nat (purge_lbl G))), 
  <<MAX: max (otf.nextbvar e') n <= n' >> /\
  <<WF: wfexp G e'>> /\
  <<BV: subuniq (otf.vars2nat (bvars e'))
                (otf.list_between n (otf.nextbvar e') ++
                                  otf.vars2nat (bvars e))>> /\
  <<GT: is_good_thunk e -> is_good_thunk e'>> /\
  <<GE: good_exp e -> good_exp e'>>.
Proof.
  i.
  assert (WF0: wfexp (inr k :: G) e).
  { rewrite <- (app_nil_l _); auto using wfexp_weaken. }
 
  destruct e; vauto; s in E; des
  ; destruct bx; try (by ss); destruct WF0 as [WF1 WF2]
  ; destruct e1; vauto
  ; destruct (list_mem _ _ _) eqn:IN; try by desf.
  rewrite ->list_mem_true in IN; s in IN.
  destruct (freshen _ _ _ _) eqn:FR.
  hexploit freshen_le; eauto; i.
  splits; vauto.
  { otf.rnbv; s in MX; otf.ssbvar; splits; unfold next in *; try omega.
    hexploit freshen_uniq_; eauto; i; des
    ; rewrite <-otf.nextbvar_equiv in *; rewrite otf.nextbvar_rename
    ; eapply within_maxbvar; eauto; []; by omega. }
  {
    apply wfexp_in with (k:=k); auto.
    s; splits; vauto.
    - eapply wfexp_exchange with
             (inr k :: inl (inr n) :: inl (inr t) :: G); first last.
      + crush_perm.
      + s in WF1; hexploit (freshen_wfexp e1); eauto; []; intros WF3.
        rewrite cons_app2 in WF3.
        eapply rename_wfexp; first (eapply purge_wfexp_x); eauto; inst; s
        ; eauto.
    - eapply wfexp_exchange with (inl (inr t) :: inr k :: inr (next n) :: G);
      try by crush_perm.
      rewrite cons_app2; auto using wfexp_weaken.
  }
  {
    otf.rnbv; s; s in MX; otf.ssbvar.
    rewrite <-!otf.nextbvar_equiv in *; rewrite otf.nextbvar_rename
    ; rewrite ->!otf.nextbvar_equiv in *.
    rewrite <-!NPeano.Nat.max_assoc; rewrite rename_bvars.
    unfold next in *.
    des_max.
    clear_upto H7.
    - otf.svbv; s; rewrite !map_app.
      destruct (le_lt_dec n' (otf.nextbvar e)).
      {
        assert (subuniq (n :: S n :: map varid (bvars e))
                        (otf.list_between n (otf.nextbvar e))).
        {
          hexploit freshen_uniq_; eauto; []; i; des.
          eapply subuniq_lb_within.
          - s; splits; ii; des; try omega; auto
            ; unfold within in WI; unfold next in *; des
            ; apply WI0 in FH; omega.
          - rewrite cons_app2.
            eapply within_app; try exact WI; eauto; unfold next.
            rewrite cons_app; eapply within_app; try (by apply within_sng)
            ; try omega.
        }
        rewrite cons_app4; rhs3 rewrite cons_app4; rewrite !app_comm_cons.
        eapply subuniq_trans; [eapply subuniq_app|]
        ; [auto | apply subuniq_app|]; [auto | apply subuniq_app|]
        ; [eapply subuniq_trans| auto |]; [| exact H|].
        { otf.rnbv; eapply subuniq_perm; crush_perm. }
        { otf.rnbv; eapply subuniq_perm; crush_perm. }
      }
      {
        assert (subuniq (map varid (bvars e))
                        (otf.list_between (S (S n)) n')).
        {
          hexploit freshen_uniq_; eauto; []; i; des.
          eapply subuniq_lb_within; auto.
        }
        rewrite <-app_nil_r in H.
        hexploit (otf.subuniq_bvar_bd H); auto.
        i. rewrite app_nil_r in H8.
        rewrite cons_app4; rhs3 rewrite cons_app4; rewrite !app_comm_cons.
        rewrite <-(@otf.list_between_sub n (S (S n)) _); auto.
        eapply subuniq_trans; [eapply subuniq_app|]
        ; [auto | apply subuniq_app|]; [auto | apply subuniq_app|]
        ; [eapply subuniq_trans| auto|]; try (rewrite cons_app2)
        ; [apply subuniq_app| |]
        ; [auto | exact H8; eapply subuniq_perm; crush_perm | auto |].
        otf.rnbv; unfold otf.list_between at 2; rewrite <-!minus_Sn_m; auto
        ; s; rewrite minus_diag; s; auto; apply subuniq_perm; crush_perm.
      }
    - otf.svbv; s; rewrite !map_app.
      destruct (le_lt_dec n' (S (S n))).
      {
        rewrite ->(le_antisym _ _ X l) in *.
        hexploit freshen_uniq_; eauto; []; i; des.
        assert (forall xs n, within n n xs -> xs = nil).
        {
          ii. induction xs as [|x xs']; vauto.
          unfold within in H; des. ss.
          specialize (H36 x (or_introl (eq_refl _))).
          specialize (H37 x (or_introl (eq_refl _))).
          omega.
        }
        assert (X1: subuniq ([S n; n]) (otf.list_between n n')).
        {
          rewrite <-(@otf.list_between_sub n (S (S n)) _); auto.
          unfold otf.list_between at 1; rewrite <-!minus_Sn_m; auto
          ; s; rewrite minus_diag; s; auto.
          eapply subuniq_trans; [eapply subuniq_perm; apply perm_swap|].
          rhs3 rewrite cons_app2; eauto using subuniq_sublist.
        }
        specialize (H _ _ WI); rewrite ->H in *; s.
        rewrite cons_app4; rhs3 rewrite cons_app4.
        eapply subuniq_trans; [apply subuniq_app|]
        ; [auto | apply subuniq_app|]; try (rewrite cons_app2)
        ; [auto | apply subuniq_app|]
        ; [exact X1 | auto|].
        apply subuniq_perm; crush_perm.
      }
      {
        assert (subuniq (map varid (bvars e))
                        (otf.list_between (S (S n)) n')).
        {
          hexploit freshen_uniq_; eauto; []; i; des.
          eapply subuniq_lb_within; eauto.
        }
        assert (subuniq (map varid (bvars e))
                        (otf.list_between (S (S n)) (S (S n)))).
        {
          hexploit freshen_uniq_; eauto; []; i; des.
          rewrite <-app_nil_r; rewrite <-app_nil_r in H.
          eapply otf.subuniq_lb_weaken_r with (n':=otf.nextbvar e)
          ; otf.rnbv; auto.
          hexploit (otf.subuniq_bvar_bd H); otf.rnbv; auto; try omega.
          otf.rnbv; auto.
        }
        otf.rnbv; unfold otf.list_between in H36
        ; rewrite minus_diag in H36; s in H36; auto.
        rewrite cons_app4; rhs3 rewrite cons_app4.
        eapply subuniq_trans; [apply subuniq_app|]
        ; [auto | apply subuniq_app|]; try (rewrite cons_app2)
        ; [auto | apply subuniq_app|]
        ; [auto | apply subuniq_app|]
        ; [exact H36 | auto|].
        otf.rnbv; unfold otf.list_between; rewrite <-!minus_Sn_m; auto
        ; s; rewrite minus_diag; s; auto; apply subuniq_perm; crush_perm.
      }
  }
Qed.

Lemma uniq_le: forall xs n
    (UN: forall m (BIG: n <= m),
           uniq (m :: xs)),
  forall x (IN: In x xs),
    x < n.
Proof.
  ii; destruct (lt_dec x n); auto.
  apply not_lt in n0.
  assert (BIG: n <= x) by omega.
  specialize (UN _ BIG); s in UN; des; vauto.
Qed.

Lemma contify_add : forall lw ea eb G k n n' phia phib
    (WF: wfexp G eb)
    (E: add_cont k G n eb = Some (ea,n'))
    (MX: otf.nextbvar eb <= n)
    (UN: forall m (BIG: n <= m),
           uniq (m :: otf.vars2nat (bvars eb) ++ otf.vars2nat (purge_lbl G)))
    (SLW: stable lw),
  <<MAX: max (otf.nextbvar ea) n <= n'>> /\
  <<SIM: mm_test.mm_psim lw G ren_id phia phib ea eb>>.
Proof.
  i; unfold add_cont; destruct eb; vauto.

  
  destruct eb1; destruct bx; vauto.
  unfold add_cont in E; vauto.
  destruct (list_mem _ _ _) eqn:INk; try by desf.
  rewrite ->list_mem_true in INk; s in INk.
  destruct (freshen _ _ _ _) eqn:FR.
  hexploit freshen_le; eauto; i.
  splits; vauto.
  { otf.rnbv; s in MX; otf.ssbvar; splits; unfold next in *; try omega.
    hexploit freshen_uniq_; eauto; i; des
    ; rewrite <-otf.nextbvar_equiv in *; rewrite otf.nextbvar_rename
    ; eapply within_maxbvar; eauto; []; by omega. }

  apply mm_psim_phi_weaken.
  clear MX FR X phia phib.

  
  apply mm_compat_let; [|apply mm_compat_fix|]
  ; eauto using extendable_t_same, ren_var_id
  ; [rewrite !ren_id_self_t !ren_id_self_c; apply mm_psim_refl; simpls; des
     ; vauto|].

  
  hexploit mm_psim_refl; [simpls; des| |]; eauto.
  i; rrdes H.


  red; splits; try done.
  eexists; i; unfold esim.
  apply pesim_acquire; i.
  apply pesim_real; i.

  rrdes HR. dup WHX. apply mm_confrel_shape in WHX. rrdes WHX.
  subst; clarify; dup RLa; dup RLb.
  rrdes RLa; rrdes RLb; inversion Em; inversion Em0.
  destruct hay, hby; destruct ee, ee0; inv Ec; inv Ec0; subst.
  clean; simpl cplus in *; noemp.
  destruct e0, e1; s; clarify.
  rename e2 into enva, e4 into envb.

  
  eapply (pesim_dtsteps (n := 0)).
  - eassumption.
  - econs; first by eapply mid_wdet. inst_pairs; econs; by econs.
    by econs.
  - i; econs; esplits; [| |exact PV| exact FIN]; inst_pairs; last reflexivity
    ; eauto.
  - eauto.
  - i. rrdes RLb; subst.
    destruct cfb; destruct ee; inv Ec1; subst.
    clean; simpl cplus in *; noemp.
    destruct e0; s; clarify.
    rename e1 into envb.

    eexists ,((hb' ## h2), Just _),. dup RLb0; rrdes RLb0.
    destruct XTR as [Xh]; subst. inv Ec1; clean.
    esplits; first 2 last.
    econs; esplits; eauto; [by econs].
    rewrite mid_csplit; erewrite (mid_csplit (hb' ## h2)).
    eapply pesim_release; first last; vauto.
    rr; esplits; eauto.
    eapply H1; eauto.
    + 
      i; specialize (RV _ IN); des; rewrite ->ren_id_self_t in *; esplits
      ; [rewrite ->ren_itvar_id in *; rewrite list_assoc_cons_neq| |]; vauto.
    + 
      i; specialize (RK _ IN); des; rewrite ->ren_id_self_t in *; esplits
      ; [rewrite ->ren_contvar_id in *; rewrite list_assoc_cons_neq| |]
      ; vauto.
      unfold next; ii; apply uniq_le with (x:=(S n)) in UN; [by omega|].
      clarify; apply in_or_app; right; otf.svbv
      ; apply (in_map varid) with (x:=inr (S n))
      ; apply purge_lbl_c; s in IN; des; vauto. 
    + right; esplits; try (by vauto)
      ; []; econs; vauto; eapply wfo_nat_lt_mon
      ; instantiate (1 := i' + 1); omega.
Qed.

Lemma respects_weaken: forall env phi e x G
    (RE: respects G env phi)
    (NL: nolbl x)
    (FR: fresh G x phi)
    (NIN: ~ In x G),
  respects (x :: G) ((x, IVclo env e) :: env) ((x, e) :: phi).
Proof.
  ii; destruct (eq_sum_dec _ _ x0 x) in IN; clarify. 
  rewrite list_assoc_cons_eq in IN; clarify.
   - exists env; splits; vauto.
     + i. destruct (LEM (x = y)).
       * clarify. rewrite defbefore_des in IN; vauto.
       * rewrite list_assoc_cons_neq; vauto.
   - rewrite list_assoc_cons_neq in IN; [|ii; auto]
     ; exploit RE; eauto; []; i; des; subst
     ; exists xenv; splits; [apply in_cons | rewrite list_assoc_cons_neq |]
     ; vauto; [ii|rewrite <- defbefore_cons]; vauto
     ; i; specialize (EN _ IN1)
     ; rewrite list_assoc_cons_neq; vauto
     ; ii; apply FR in IN; des; rewrite <- H in IN1; apply IN2 in IN1
     ; assumption.
Qed.

Lemma respects_weaken_G: forall env phi x G
    (RE: respects G env phi)
    (NL: nolbl x)
    (FR: fresh G x phi),
  respects (x :: G) env phi.
Proof.
  ii; exploit RE; eauto; []; i; des; subst.
  apply FR in IN; []; des; esplits; first [rewrite <- defbefore_cons | s]
  ; vauto.
Qed.

Ltac myclar :=
 repeat 
 match goal with [H : _ = Some _ , H' : _ = Some _ |- _ ] =>
    eapply eq_trans in H'; [|symmetry; exact H]; clarify
 end.

Lemma alpha_eq_asim: forall lw G ea eb ren
    (AEQ: alpha_eq ea eb ren = true)
    (FV: forall x (IN: In x G), ~ In (rename_var ren x) (bvars ea))
    (UN: uniq (bvars ea ++ bvars eb ++ purge_lbl G))
    (WFB: wfatom G eb)
    (SLW: stable lw)
    (IH: forall ea', size ea' < size ea -> forall lw G eb ren,
           alpha_eq ea' eb ren = true ->
           (forall x (IN: In x G), ~ In (rename_var ren x) (bvars ea')) ->
           uniq (bvars ea' ++ bvars eb ++ purge_lbl G) ->
           wfexp G eb ->
           stable lw ->
           mm_psim lw G ren ea' eb),
  mm_asim lw G ren ea eb.
Proof.
  i; ginduction ea; destruct eb; try (by vauto)
  ; try (by destruct op; vauto)
  ; try (by destruct bx; vauto); i
  ; auto with mm_compat; s in AEQ; s in WFB; des
  ; try (by apply beq_nat_true in AEQ; try (apply enum_inj in AEQ); clarify
         ; auto with mm_compat).
  - destruct op; destruct op0; vauto; destruct x; destruct x0; vauto
    ; destruct y; destruct y0; vauto
    ; try (apply andb_true_iff in AEQ; des; unfold teq in *; desf)
    ; try (rewrite ->beq_nat_true_iff in *; clarify)
    ; apply mm_compat_bop; auto.
  - unfold teq in *; desf; auto with mm_compat.
  - unfold teq in *; desf; auto with mm_compat.
  - apply mm_compat_fix.
    + ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
    + ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
    + ii; clarify; exfalso; s in UN; des; apply UN; vauto.
    + apply (IH ea); auto.
      {
        ii; apply in_consD in IN; des; clarify.
        {
          rewrite ren_ext_c_eq in FH; s in UN; des.
          apply UN1; apply in_or_app; vauto.
        }
        rewrite ren_c_ext_neq in FH; vauto.
        apply in_consD in IN; des; clarify.
        {
          rewrite ren_ext_eq in FH; s in UN; des.
          apply UN0; right; apply in_or_app; vauto.
        }
        rewrite ren_ext_neq in FH; vauto.
        apply in_consD in IN; des; clarify.
        {
          rewrite ren_ext_eq in FH; s in UN; des.
          apply UN; do !right; apply in_or_app; vauto.
        }
        rewrite ren_ext_neq in FH; vauto.
        assert (ING: In x G) by (eauto using sublist_in, uniq_sublist).
        apply FV in ING; auto; []; s; do !right; vauto.
      }
      {
        s; s in UN; des.
        eapply uniq_perm in UN2; [| eapply Permutation_trans]
        ; [| apply Permutation_app|]
        ; [| auto | eapply Permutation_trans|]
        ; [| rewrite cons_app3; apply Permutation_app | |]
        ; [| apply Permutation_rev | auto | |]
        ; [| do !apply Permutation_cons_app |]
        ; auto.
        apply (uniq_sublist UN2); do !(apply sublist_app; auto).
        eauto using purge_lbl_sublist, purge_sublist.
      }
  - apply mm_compat_gen; auto.
    apply (IH ea); auto.
    {
      ii; s in IN; des.
      - clarify; rewrite ren_ext_c_eq in FH.
        s in UN; des; apply UN; apply in_or_app; vauto.
      - assert (ING: In x G) by (eauto using sublist_in, uniq_sublist).
        apply FV in ING; vauto; []; s; right.
        destruct (LEM (x = inr bk0)); clarify.
        + rewrite ren_ext_c_eq in FH.
          exfalso; s in UN; des; apply UN; apply in_or_app; vauto.
        + rewrite ren_c_ext_neq in FH; auto.
    }
    {
      s; s in UN; des.
      eapply uniq_perm in UN0; [| eapply Permutation_trans]
      ; [| apply Permutation_app|]
      ; [| auto | eapply Permutation_cons_app|]
      ; auto.
      apply (uniq_sublist UN0); do !(apply sublist_app; auto).
      eauto using purge_lbl_sublist, purge_sublist.
    }
  - apply andb_true_iff in AEQ; des; unfold teq in *; desf
    ; auto with mm_compat.
Qed.

Lemma alpha_eq_ksim: forall lw G ea eb ren
    (AEQ: alpha_eq ea eb ren = true)
    (FV: forall x (IN: In x G), ~ In (rename_var ren x) (bvars ea))
    (UN: uniq (bvars ea ++ bvars eb ++ purge_lbl G))
    (WFB: wfatom G eb)
    (SLW: stable lw)
    (IH: forall ea', size ea' < size ea -> forall lw G eb ren,
           alpha_eq ea' eb ren = true ->
           (forall x (IN: In x G), ~ In (rename_var ren x) (bvars ea')) ->
           uniq (bvars ea' ++ bvars eb ++ purge_lbl G) ->
           wfexp G eb ->
           stable lw ->
           mm_psim lw G ren ea' eb),
  mm_ksim lw G ren ea eb.
Proof.
  i; ginduction ea; destruct eb; try (by vauto)
  ; try (by destruct op; vauto); try (by destruct bx; vauto); i
  ; s in WFB; des.
  apply mm_compat_cont.
  - ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
  - apply (IH ea); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_eq in FH.
          s in UN; des; apply UN; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right.
          destruct (LEM (x = inl (inr bx0))); clarify.
          + rewrite ren_ext_eq in FH.
            exfalso; s in UN; des; apply UN; apply in_or_app; vauto.
          + rewrite ren_ext_neq in FH; auto.
      }
      {
        s; s in UN; des.
        eapply (uniq_perm UN0); crush_perm.
      }
Qed.


Lemma alpha_eq_sim: forall ea lw G eb ren
    (AEQ: alpha_eq ea eb ren = true)
    (FV: forall x (IN: In x G), ~ In (rename_var ren x) (bvars ea))
    (UN: uniq (bvars ea ++ bvars eb ++ purge_lbl G))
    (WFB: wfexp G eb)
    (SLW: stable lw),
  mm_psim lw G ren ea eb.
Proof.
  induction ea as [? IH] using (well_founded_ind size_lt_well_founded)
  ; i; destruct ea; vauto; ii; destruct eb; vauto; try (destruct op; vauto)
  ; try (by destruct bx; vauto); s in AEQ
  ; try (apply andb_true_iff in AEQ; des); unfold teq in *; desf; s in WFB
  ; des.
  - apply mm_compat_new; auto.
    + ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
    + apply (IH ea); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_eq in FH.
          s in UN; des; apply UN; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right.
          destruct (LEM (x = inl (inr bx0))); clarify.
          + rewrite ren_ext_eq in FH.
          exfalso; s in UN; des; apply UN; apply in_or_app; vauto.
          + rewrite ren_ext_neq in FH; auto.
      }
      {
        s; s in UN; des.
        eapply (uniq_perm UN0); crush_perm.
      }
  - apply mm_compat_read; auto.
    + ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
    + apply (IH ea); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_eq in FH.
          s in UN; des; apply UN; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right.
          destruct (LEM (x = inl (inr bx0))); clarify.
          + rewrite ren_ext_eq in FH.
          exfalso; s in UN; des; apply UN; apply in_or_app; vauto.
          + rewrite ren_ext_neq in FH; auto.
      }
      {
        s; s in UN; des.
        eapply (uniq_perm UN0); crush_perm.
      }
  - apply mm_compat_write; auto.
  - apply mm_compat_input; auto.
    + ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
    + apply (IH ea); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_eq in FH.
          s in UN; des; apply UN; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right.
          destruct (LEM (x = inl (inr bx0))); clarify.
          + rewrite ren_ext_eq in FH.
          exfalso; s in UN; des; apply UN; apply in_or_app; vauto.
          + rewrite ren_ext_neq in FH; auto.
      }
      {
        s; s in UN; des.
        eapply (uniq_perm UN0); crush_perm.
      }
  - apply mm_compat_output; auto.
  - apply mm_compat_ifnz; auto.
    {
      apply (IH ea1); first (by s; rr; omega); auto.
      - ii; apply FV in IN; vauto; s; apply in_or_app; vauto.
      - s in UN; des; rewrite -!app_assoc in UN.
        apply (uniq_sublist UN); apply sublist_app; auto.
        apply sublist_r_app_r; apply sublist_app; auto.
    }
    {
      apply (IH ea2); first (by s; rr; omega); auto.
      - ii; apply FV in IN; vauto; s; apply in_or_app; vauto.
      - s in UN; des; rewrite -!app_assoc in UN
        ; eauto using uniq_sublist, sublist_r_app_r, sublist_app.
    }
  - apply mm_compat_case; auto.
    + ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
    + apply (IH ea1); first (by s; rr; omega); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_eq in FH.
          s in UN; des; apply UN; apply in_or_app; left
          ; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right; apply in_or_app; left.
          destruct (LEM (x = inl (inr bx0))); clarify.
          + rewrite ren_ext_eq in FH.
            exfalso; s in UN; des; apply UN; apply in_or_app; left
            ; apply in_or_app; vauto.
          + rewrite ren_ext_neq in FH; auto.        
      }
      {
        s; s in UN; des; rewrite -!app_assoc in UN0.
        eapply uniq_perm in UN0; [| apply Permutation_app]
        ; [| auto | apply Permutation_app]
        ; [| auto | eapply Permutation_trans]
        ; [| apply Permutation_cons_app|]
        ; [| apply Permutation_app | apply Permutation_app]
        ; [| auto | apply Permutation_app | | apply Permutation_cons_app]
        ; auto.
        eauto 6 using uniq_sublist, sublist_r_app_r, sublist_app.
      }
    + apply (IH ea2); first (by s; rr; omega); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_eq in FH.
          s in UN; des; apply UN; apply in_or_app; left
          ; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right; apply in_or_app; right.
          destruct (LEM (x = inl (inr bx0))); clarify.
          + rewrite ren_ext_eq in FH.
            exfalso; s in UN; des; apply UN; apply in_or_app; left
            ; apply in_or_app; vauto.
          + rewrite ren_ext_neq in FH; auto.        
      }
      {
        s; s in UN; des; rewrite -!app_assoc in UN0.
        eapply uniq_perm in UN0; [| apply Permutation_app]
        ; [| auto | apply Permutation_app]
        ; [| auto | eapply Permutation_trans]
        ; [| apply Permutation_cons_app|]
        ; [| apply Permutation_app | apply Permutation_app]
        ; [| auto | apply Permutation_app | | apply Permutation_cons_app]
        ; auto.
        eauto 6 using uniq_sublist, sublist_r_app_r, sublist_app.
      }
  - apply mm_compat_ret; auto.
  - apply mm_compat_app; auto.
  - apply mm_compat_inst; auto.
  - apply andb_true_iff in AEQ; des; apply mm_compat_let; auto.
    + ii; apply FV in IN; vauto; []; s; rewrite E; vauto.
    + apply alpha_eq_asim; auto.
      {
        ii; apply FV in IN; vauto; s; right; apply in_or_app; vauto.
      }
      {
        s in UN; des; rewrite -!app_assoc in UN0.
        apply (uniq_sublist UN0); apply sublist_app; auto.
        apply sublist_r_app_r; apply sublist_cons_drop.
        apply sublist_app; auto.
      }
      {
        ii; apply (IH ea'); first (by s; rr; omega); auto.
      }
    + apply (IH ea2); first (by s; rr; omega); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_eq in FH.
          s in UN; des; apply UN; apply in_or_app; left
          ; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right; apply in_or_app; right.
          destruct (LEM (x = inl (inr t0))); clarify.
          + rewrite ren_ext_eq in FH.
            exfalso; s in UN; des; apply UN; apply in_or_app; left
            ; apply in_or_app; vauto.
          + rewrite ren_ext_neq in FH; auto.        
      }
      {
        s; s in UN; des; rewrite -!app_assoc in UN0.
        eapply uniq_perm in UN0; [| apply Permutation_app]
        ; [| auto | apply Permutation_app]
        ; [| auto | eapply Permutation_trans]
        ; [| apply Permutation_cons_app|]
        ; [| apply Permutation_app | apply Permutation_app]
        ; [| auto | apply Permutation_app | | apply Permutation_cons_app]
        ; auto.
        eauto 6 using uniq_sublist, sublist_r_app_r, sublist_app.
      }
  - apply andb_true_iff in AEQ; des; apply mm_compat_letcont; auto.
    + ii; apply FV in IN; vauto.
    + apply alpha_eq_ksim; auto.
      {
        ii; apply FV in IN; vauto; s; right; apply in_or_app; vauto.
      }
      {
        s in UN; des; rewrite -!app_assoc in UN0.
        apply (uniq_sublist UN0); apply sublist_app; auto.
        apply sublist_r_app_r; apply sublist_cons_drop.
        apply sublist_app; auto.
      }
      {
        ii; apply (IH ea'); first (by s; rr; omega); auto.
      }
    + apply (IH ea2); first (by s; rr; omega); auto.
      {
        ii; s in IN; des.
        - clarify; rewrite ren_ext_c_eq in FH.
          s in UN; des; apply UN; apply in_or_app; left
          ; apply in_or_app; vauto.
        - apply FV in IN; vauto; []; s; right; apply in_or_app; right.
          destruct (LEM (x = inr c0)); clarify.
          + rewrite ren_ext_c_eq in FH.
            exfalso; s in UN; des; apply UN; apply in_or_app; left
            ; apply in_or_app; vauto.
          + rewrite ren_c_ext_neq in FH; auto.        
      }
      {
        s; s in UN; des; rewrite -!app_assoc in UN0.
        eapply uniq_perm in UN0; [| apply Permutation_app]
        ; [| auto | apply Permutation_app]
        ; [| auto | eapply Permutation_trans]
        ; [| apply Permutation_cons_app|]
        ; [| apply Permutation_app | apply Permutation_app]
        ; [| auto | apply Permutation_app | | apply Permutation_cons_app]
        ; auto.
        eauto 6 using uniq_sublist, sublist_r_app_r, sublist_app.
      }
Qed.

Lemma contify_related' lw G phia phib ec eb f mx yb k bf bx bk x: forall
    (WF: wfatom (defbefore (inl f) G) (IEfix bf bx bk eb))
    (INL1: lookup (inr mx) phia = Some (IEcont yb ec))
    (INL2: lookup (inl f) phib = Some (IEfix bf bx bk eb))
    (AEQ: alpha_eq ec eb
                    ren_id[f/bf][inr yb/bx][c k/bk] = true)
    (DB : forall y, In y (defbefore (inl f) G) ->
                    In y (defbefore (inr mx) G))
    (UN: uniq (bvars (IEfix bf bx bk eb) ++
               bvars (IEcont yb ec) ++
               purge_lbl G))
    (SLW: stable lw)
    (IN1: In (inr mx) G)
    (IN2: In (inl f) G)
    (INDB: In (inl f) (defbefore (inr mx) G))
    (IN3: ~ In (inl (inr yb)) G)
    (IN4: In (inl x) G)
    (IN5: In (inr k) (defbefore (inl f) G)),
  mm_test.mm_psim lw G ren_id phia phib (IEret mx x) (IEapp f x k).
Proof.
  ii; rr; splits; vauto; []; exists 0; i; s; des.
  set (ea := rename (((ren_id)[f/bf])[inr yb/bx])[c k/bk] eb).
  set (ren := (((ren_id)[f/bf])[inr yb/bx])[c k/bk]).

  assert (IN6: In (inr k) G)
               by (eauto using sublist_in, defbefore_sublist).

  
  assert (BVV: forall y
              (E: y = inl f \/ y = inl x \/ y = inr k \/
                  y = inl (inr yb)),
              ~ In y (bvars eb)).
  {
    i; apply uniq_appD in UN; des; apply uniq_appD in UN0; des; clarify.
    - eapply UN1.
      + s; do 3!right; eauto.
      + apply in_or_app; right; eauto using purge_lbl_nolbl, bvars_nolbl.
    - eapply UN1.
      + s; do 3!right; eauto.
      + apply in_or_app; right; eauto using purge_lbl_nolbl, bvars_nolbl.
    - eapply UN1.
      + s; do 3!right; eauto.
      + apply in_or_app; right; eauto using purge_lbl_nolbl, bvars_nolbl.
    - eapply UN1.
      + s; do 3!right; eauto.
      + apply in_or_app; left; s; vauto.
  }

  assert (SIM: mm.mm_psim lw
                       (inr bk :: inl (inr bx) :: inl (inr bf) ::
                        purge (defbefore (inl f) G))
                       ren
                       ec eb ).
  {
    apply alpha_eq_sim; auto.
    - ii; apply in_consD in IN; des; clarify.
      {
        rewrite ren_ext_c_eq in FH; s in UN; des
        ; apply uniq_appD in UN2; des; s in UN3; des
        ; apply uniq_appD in UN5; des.
        apply (UN7 (inr k)); auto.
      }
      rewrite ren_c_ext_neq in FH; vauto.
      apply in_consD in IN; des; clarify.
      {
        rewrite ren_ext_eq in FH; s in UN
        ; des; apply uniq_appD in UN2; des; s in UN3; des
        ; apply uniq_appD in UN5; des.
        apply UN3; apply in_or_app; vauto.
      }
      rewrite ren_ext_neq in FH; vauto.
      apply in_consD in IN; des; clarify.
      {
        rewrite ren_ext_eq in FH; s in UN; des
        ; apply uniq_appD in UN2; des; s in UN3; des
        ; apply uniq_appD in UN5; des.
        apply (UN7 (inl f)); eauto using purge_lbl_nolbl, bvars_nolbl.
      }
      rewrite ren_ext_neq in FH; vauto.
      rewrite ren_var_id in FH.
      s in UN; des
      ; apply uniq_appD in UN2; des; s in UN3; des
      ; apply uniq_appD in UN5; des.
      apply (UN7 x0); eauto using purge_lbl_nolbl, bvars_nolbl, purge_sublist
                      , sublist_in, defbefore_sublist.
    - s; simpl bvars in UN.
      
      replace (match f with
                 | inl l => lbls_upto l G
                 | inr _ => drop_until (inl f) G
               end)
              with (defbefore (inl f) G) by vauto.
      eapply uniq_perm in UN; [| eapply Permutation_trans]
      ; [| apply three_way_perm| apply Permutation_app]
      ; [| auto | rewrite -!app_comm_cons; eapply Permutation_trans]
      ; [| rewrite cons_app3; apply Permutation_app |]
      ; [| apply Permutation_rev | auto |]
      ; [| do !apply Permutation_cons_app]
      ; auto.
      s in UN; des; eapply uniq_sublist; eauto.
      apply sublist_app; auto; apply sublist_app; auto
      ; do !apply sublist_cons_keep.
      eauto using purge_lbl_sublist, sublist_trans, purge_sublist,
                  defbefore_sublist.
  }

  fold ren in AEQ.
  unfold ren in *.
  rrdes SIM; pfold; econs; i.
  rrdes RPa; specialize (RPa _ _ INL1); des.
  rrdes RPb; specialize (RPb _ _ INL2); des.
  apply RK in IG; apply RV in IG0; apply RV in IN4
  ; rewrite ren_itvar_id in IN4; apply RK in IN6.
  mm_destruct XTRa XTRb CNF.
  myclar.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb; do 3 eexists; eapply IRret; eauto.
  - esplits; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]. s in STa. invc STa.
    do 7 eexists; exists false; splits; first (by eauto)
    ; last (by s; eauto); last 1 first.
    + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
    + left; econs; eauto.
      * left. eapply IRok. eapply IRapp; eauto.
      * eapply gres_incl_id. left.
        rewrite mid_csplit. rewrite ->mid_csplit at 1.
        eapply pesim_release; s; eauto.
        myclar.
        eapply SIM1; vauto.

        clear CO CO0.
        clear SIM1.

        
        {
          intros x3 IN.
          assert (NEQbf: bx0 <> bf).
          {
            ii. simpl in UN; des; apply UN; right
            ; right; apply in_or_app; right; simpl; left; subst; auto.
          }
          
          apply in_consD in IN; des; vauto.
          apply in_consD in IN; des; clarify.
          - simpl rename_itvar; tmext_simpl; rewrite list_assoc_cons_eq
            ; rewrite list_assoc_cons_neq; vauto.
          - rewrite ren_ext_it_neq_c; dup IN2
            ; apply RV in IN2; des; apply in_consD in IN; des; clarify.
            + rewrite ren_ext_it_neq; [| by ii; vauto].
              simpl rename_itvar; tmext_simpl
              ; rewrite list_assoc_cons_neq; [| by ii; vauto]
              ; rewrite list_assoc_cons_neq; [| by ii; vauto]
              ; rewrite list_assoc_cons_neq; [| by ii; vauto]
              ; rewrite list_assoc_cons_eq.
              apply EN in INDB; rewrite ->ren_itvar_id in *
              ; eapply eq_trans in IG0; [|exact INDB]; clarify
              ; esplits; eauto.
            + 
              apply purge_inD in IN.
              specialize (EN _ (DB _ IN)); specialize (EN0 _ IN).
              assert (IN' : In (inl x3) G) by
                   eauto using sublist_in, defbefore_sublist.
              assert (NEQ: forall x, ~ In x (defbefore (inl f) G) ->
                                     x <> inl x3)
                     by (ii; vauto).

              exploit (RV x3); try done; intros (va' & vb' & La & Lb & Vc).
              eexists va', vb'; rewrite ->ren_itvar_id in *.
              splits; [| |apply Vc].
              * repeat first [ rewrite ren_itvar_id | rewrite ren_ext_it_neq
                             | rewrite ren_ext_it_neq_c]
                ; try (rewrite list_assoc_cons_neq
                       ; [by eapply eq_trans, eq_sym; eauto|])
                ; try (by ii; vauto).
              * repeat (rewrite list_assoc_cons_neq); try (by ii; vauto)
                ; eapply eq_trans; eauto.
        }
        
        {
          
          intros x3 IN.
          apply in_consD in IN; des; clarify.
          {
            specialize (EN _ (DB _ IN5)).
            simpl rename_contvar; tmext_simpl; rewrite list_assoc_cons_eq
            ; rewrite list_assoc_cons_neq; vauto
            ; rewrite ->ren_contvar_id in *
            ; esplits; [eapply eq_trans | auto | exact IN10]; [exact EN|auto].
          }
          {
            
            repeat (apply in_consD in IN; des; vauto).

            
            apply purge_inD in IN.
            specialize (EN _ (DB _ IN)); specialize (EN0 _ IN).
            assert (IN' : In (inr x3) G) by
                   eauto using sublist_in, defbefore_sublist.
            assert (NEQ: forall x, ~ In x (defbefore (inl f) G) ->
                                   x <> inr x3)
                   by (ii; vauto).

            hexploit (RK x3); try done; intros (ka' & kb' & La & Lb & Vc).
            eexists ka', kb'; rewrite ->ren_contvar_id in *.
            splits; [| |apply Vc].
            - repeat first [rewrite ren_contvar_id | rewrite ren_ext_c_neq
                            | rewrite ren_ext_c_neq_c ]
              ; try (rewrite list_assoc_cons_neq
                     ; [by eapply eq_trans, eq_sym; eauto|])
              ; try (by ii; vauto).
            - repeat rewrite list_assoc_cons_neq; auto
              ; by eapply eq_trans; eauto.
          }
        }
Qed.


Ltac snouniq :=
  match goal with
  | |- context[uniq] => idtac
  | |- _ => simpl
  end.


Lemma uniq_sp_perm (w x y z : list var) :
    uniq (w ++ x ++ y ++ z) ->
  uniq (y ++ w ++ z ++ x).
Proof.
  intros UN; eapply (uniq_perm UN)
  ; eapply Permutation_app_comm_useful.
  - rewrite app_assoc; instantiate (1 := y ++ (w ++ x) ++ z)
    ; apply Permutation_wf_bvar; auto.
  - rewrite app_assoc; instantiate (1 := (y ++ w) ++ (x ++ z))
    ; apply Permutation_app; auto using Permutation_app_comm.
  - repeat (rewrite <- app_assoc); auto.
Qed.

Lemma defbefore_neq: forall x y z G
   (IN: In x (defbefore y (z :: G)))
   (NL: nolbl z)
   (NEQ: z <> x),
  In x (defbefore y G).
Proof.
 i. destruct (LEM (nolbl y)) as [NLY|LY]
 ; destruct (LEM (In y G)) as [INY|NINY].
 - rewrite <- defbefore_cons in IN; auto.
 - rewrite defbefore_des'; auto. destruct (LEM (y = z)) as [EQY|NEQY].
   + clarify. rewrite defbefore_des in IN; auto.
   + rewrite defbefore_des' in IN; auto; ii; simpls; des; vauto.
 - destruct y as [[yl|yt]|yk]; try (by exfalso; eapply LY; eauto).
   unfold defbefore in IN.
   simpl lbls_upto in IN; destruct z as [[zl|zt]|zk]; vauto.
   + exfalso; rrdes NL; apply (NL zl); auto.
 - destruct y as [[yl|yt]|yk]; try (by exfalso; eapply LY; eauto).
   unfold defbefore in IN.
   simpl lbls_upto in IN; destruct z as [[zl|zt]|zk]; vauto.
   + exfalso; rrdes NL; apply (NL zl); auto.
Qed.

Lemma fresh_extend: forall x y G phi
    (FR: fresh G x phi)
    (NL: nolbl y)
    (NEQ: x <> y),
  fresh (y :: G) x phi.
Proof.
  ii; specialize (FR _ _ IN); des; splits; auto.
  intros IN2; eauto using defbefore_neq.
Qed.


Lemma fresh_extend_bvar: forall x y G phi e
    (FR: x = y \/ In x (bvars e) -> fresh G x phi)
    (BV: In x (bvars e))
    (NL: nolbl y)
    (NEQ: x <> y),
  fresh (y :: G) x phi.
Proof.
  i; eapply fresh_extend; eauto.
Qed.

Lemma fresh_extend_mult_bv: forall x y G phi e1 e2
    (FR: In x (bvars e1 ++ bvars e2) -> fresh G x phi)
    (BV: In x (bvars e1) \/ In x (bvars e2))
    (NL: nolbl y)
    (NEQ: x <> y),
  fresh (y :: G) x phi.
Proof.
  i; eapply fresh_extend; eauto.
  des.
  - assert (H: In x (bvars e1 ++ bvars e2)) by (apply in_or_app; left; done).
    auto.
  - assert (H: In x (bvars e1 ++ bvars e2)) by (apply in_or_app; right; done).
    auto.
Qed.

Lemma fresh_extend_mult_bv_or: forall x y G phi e1 e2
    (FR: y = x \/ In x (bvars e1 ++ bvars e2) -> fresh G x phi)
    (BV: In x (bvars e1) \/ In x (bvars e2))
    (NL: nolbl y)
    (NEQ: x <> y),
  fresh (y :: G) x phi.
Proof.
  i; eapply fresh_extend; eauto.
  des.
  - assert (H: y = x \/ In x (bvars e1 ++ bvars e2))
           by (right; apply in_or_app; left; done).
    auto.
  - assert (H: y = x \/ In x (bvars e1 ++ bvars e2))
           by (right; apply in_or_app; right; done).
    auto.
Qed.

Lemma fresh_no_extend_mult: forall x G phi e1 e2
    (FR: In x (bvars e1 ++ bvars e2) -> fresh G x phi)
    (BV: In x (bvars e1) \/ In x (bvars e2)),
  fresh G x phi.
Proof.
  i. des.
  - assert (H: In x (bvars e1 ++ bvars e2)) by (apply in_or_app; left; done).
    auto.
  - assert (H: In x (bvars e1 ++ bvars e2)) by (apply in_or_app; right; done).
    auto.
Qed.

Ltac fresh_crush :=
  intros x0 INx
  ; match goal with
    | [H: In x0 (bvars ?X),
       H': forall x, ?P -> fresh _ _ ?Z,
       UN: uniq (_ ++ bvars ?E1 ++ bvars ?E2 ++ _)
       |- context[fresh _  _ ?Z]] =>
       solve [specialize (H' x0); s in H'
       ; apply fresh_extend_mult_bv_or with (e1 := E1)(e2 := E2); auto
       ; s in UN; ii; clarify; apply UN; apply in_or_app; left; auto]
    end.


Lemma replace_sound : forall lw G e eb ec phia phib bf bx bk f k fc yb
    (WF: wfexp G e)
    (INL1: lookup (inr fc) phia =  Some (IEcont yb ec))
    (AEQ: alpha_eq ec eb ren_id[inr f/bf][inr yb/bx][c k/bk])
    (INL2: lookup (inl (inr f)) phib = Some (IEfix bf bx bk eb))
    (DB : forall y, In y (defbefore (inl (inr f)) G) ->
                  In y (defbefore (inr fc) G))
    (FRa: forall x (IN: In x (bvars e)), fresh G x phia)
    (FRb: forall x (IN: In x (bvars e)), fresh G x phib)
    (IN1: In (inr fc) G)
    (IN2: In (inl (inr f)) (defbefore (inr fc) G))
    (IN3: ~ In (inl (inr yb)) G)
    (IN5: In (inr k) (defbefore (inl (inr f)) G))
    (UN: uniq (bvars e ++ purge_lbl G))
    (UNf: uniq (bvars e ++ bvars (IEfix bf bx bk eb) ++
                bvars (IEcont yb ec) ++
                purge_lbl G))
    (WFa: wfatom (defbefore (inl (inr f)) G) (IEfix bf bx bk eb))
    (SLW: stable lw),
  mm_test.mm_psim lw G ren_id phia phib (replace f k fc e) e.
Proof.
  i.
  
  
  assert (INf: In (inl (inr f)) G)
         by (eauto using sublist_in, defbefore_sublist).
  
  revert G WF DB FRa FRb IN1 IN2 IN3 IN5 UN UNf INf WFa; induction e
  ; i; simpl wfexp in *; desc; vauto
  
  ;(match goal with
      | |- context[IElet] => destruct bx0
      | |- _ => idtac
    end)
  ; first [ eapply mm_test.mm_compat_new | eapply mm_test.mm_compat_read
            | eapply mm_test.mm_compat_write | eapply mm_test.mm_compat_let
            | eapply mm_test.mm_compat_input
            | eapply mm_test.mm_compat_letcont
            | eapply mm_test.mm_compat_output
            | eapply mm_test.mm_compat_ifnz
            | eapply mm_test.mm_compat_ret | eapply mm_test.mm_compat_case
            | eapply mm_test.mm_compat_inst
            | idtac ]
  ; eauto using ren_itvar_id, mm_test.extendable_t_same, ren_var_id
  
  ; try (try (rewrite ren_id_self_t || rewrite ren_id_self_c)
         ; first [ eapply IHe | eapply IHe1 | eapply IHe2])
  ; try (by simpl bvars in UN; eapply (uniq_perm UN)
         ; rewrite <- app_comm_cons; s; auto using Permutation_cons_app)
  ; try (by apply mm_asim_phi_weaken; apply mm_asim_refl; des; auto)
  ; try (by apply mm_ksim_phi_weaken; apply mm_ksim_refl; des; auto)
  
  ; try (by des; vauto)
  
  ; try (by simpl bvars in UN; apply (uniq_sublist UN); rewrite <- app_assoc
         ; auto using sublist_r_app_r, sublist_app)
  
  ; try (by rewrite <- !defbefore_cons; auto)
  
  ; try (simpl bvars in UN; rewrite <- app_comm_cons in UN
         ; rewrite <- app_assoc in UN; simpl
         ; rewrite cons_app; rewrite cons_app in UN)
  
  ; try (by apply uniq_sp_perm in UN; apply (uniq_sublist UN)
         ; auto using sublist_app)
  
  ; try (by intros FH; apply in_consD in FH; des; vauto
         ; simpl in UNf; des; apply UNf; apply in_or_app; right; s
         ; repeat right; apply in_or_app; right; s; auto)
  
  ; try (by simpl bvars in UNf; s bvars; s purge_lbl
         ; rhs3 rewrite <- app_comm_cons in UNf; rewrite cons_app in UNf
         ; first [ rewrite (app_assoc (bvars e)) in UNf
                 | rewrite (app_assoc (bvars e1 ++ bvars e2)) in UNf]
         ; rewrite (app_assoc _ _ (purge_lbl G)) in UNf
         ; rewrite app_assoc in UNf; eapply uniq_perm in UNf
         ; [| eapply Permutation_app]; [|eapply Permutation_app_comm|]; eauto
         ; instantiate; rewrite <- !app_assoc in UNf
         ; apply (uniq_sublist UNf); s
         ; auto 6 using sublist_app, sublist_cons_drop)
  
  ; try (by simpl bvars in UNf; s bvars; apply (uniq_sublist UNf)
         ; rhs3 rewrite <- app_comm_cons
         ; auto using sublist_app, sublist_cons_drop)
  ; try (by apply FRa; vauto); try (by apply FRb; vauto)
  
  ; try (by intros x0 INx; specialize (FRa x0); s in FRa; eapply fresh_extend
         ; eauto; s in UN; ii; clarify; apply UN; apply in_or_app; left; auto)
  ; try (by intros x0 INx; specialize (FRb x0); s in FRb; eapply fresh_extend
         ; eauto; s in UN; ii; clarify; apply UN; apply in_or_app; left
         ; auto)
  ; try (by intros x0 INx; specialize (FRa x0); s in FRa
         ; eapply fresh_no_extend_mult; eauto)
  ; try (by intros x0 INx; specialize (FRb x0); s in FRb
         ; eapply fresh_no_extend_mult; eauto).
  - fresh_crush.
  - fresh_crush.
  - rewrite <- (app_nil_l _); rewrite (app_assoc (bvars e1)) in UN.
    eapply uniq_perm in UN; [|eapply Permutation_wf_bvar; eauto]
    ; rewrite -> (app_nil_l _)
    ; apply (uniq_sublist UN); auto using sublist_app.
  - intros x0 INx
    ; specialize (FRa x0 (or_intror _ (in_or_app _ _ _ (or_intror _ INx))))
    ; apply fresh_extend_mult_bv_or with (e1 := e1)(e2 := e2); auto
    ; s in UN; ii; clarify; apply UN; apply in_or_app
    ; right; apply in_or_app; left; auto.
  - intros x0 INx; specialize (FRb x0); s in FRb
    ; apply fresh_extend_mult_bv_or with (e1 := e1)(e2 := e2); auto
    ; s in UN; ii; clarify; apply UN; apply in_or_app
    ; right; apply in_or_app; left; auto.
  - 
    simpl; destruct xf
    
    ; try (unfold teq; destruct (eqdec t f); destruct (eqdec xk k)); s
    
    ; try (by eapply mm_psim_phi_weaken; apply mm_compat_app
           ; auto using ren_itvar_id)
    
    ; []; s; try (eapply mm_test.mm_compat_app
              ; eauto using ren_itvar_id)
    ; subst; eapply contify_related'; try apply list_assoc_cons_eq
    ; eauto.
  - intros x0 INx; specialize (FRa x0); s in FRa
    ; apply fresh_extend_mult_bv_or with (e1 := e1)(e2 := e2); auto
    ; s in UN; ii; clarify; apply UN; apply in_or_app
    ; right; apply in_or_app; left; auto.
  - intros x0 INx; specialize (FRb x0); s in FRb
    ; eapply fresh_extend_mult_bv_or with (e1 := e1)(e2 := e2); auto
    ; s in UN; ii; clarify; apply UN; apply in_or_app
    ; right; apply in_or_app; left; auto.
  - intros x0 INx; specialize (FRb x0); s in FRb
    ; eapply fresh_extend_mult_bv_or with (e1 := e1)(e2 := e2); auto
    ; s in UN; ii; clarify; apply UN; apply in_or_app
    ; right; apply in_or_app; left; auto.
  - intros x0 INx; specialize (FRb x0); s in FRb
    ; eapply fresh_extend_mult_bv_or with (e1 := e1)(e2 := e2); auto
    ; s in UN; ii; clarify; apply UN; apply in_or_app
    ; right; apply in_or_app; left; auto.
Qed.

Lemma replace_nextbvar_equiv: forall e f k fc,
  otf.nextbvar (replace f k fc e) = otf.nextbvar e.
Proof.
  induction e; vauto; ii; try (by unfold replace; desf)
  ; try (rewrite ->!otf.nextbvar_equiv in *; s; rewrite <-(IHe f k fc)
         ; rewrite ->!otf.nextbvar_equiv in *; auto)
  ; try (rewrite ->!otf.nextbvar_equiv in *; s; rewrite <-(IHe1 f k fc)
         ; rewrite <-(IHe2 f k fc); rewrite ->!otf.nextbvar_equiv in *
         ; by auto).  
Qed.

Lemma replace_bvars: forall e f k fc,
  bvars (replace f k fc e) = bvars e.
Proof.
  induction e; vauto; ii; try (by unfold replace; desf)
  ; try (by s; rewrite IHe; auto)
  ; try (by s; try destruct bx; try rewrite IHe1; rewrite IHe2; auto).
Qed.


Lemma contify_subst_wf: forall e e' k G n n'
    (E: subst_cont k G n e = Some (e',n'))
    (WF: wfexp G e)
    (MX: (otf.nextbvar e) <= n)
    (UN: forall m (BIG: n <= m),
           uniq (m :: otf.vars2nat (bvars e) ++ otf.vars2nat (purge_lbl G))),
  <<MAX: max (otf.nextbvar e') n <= n' >> /\
  <<WF: wfexp G e'>> /\
  <<BV: subuniq (otf.vars2nat (bvars e'))
                (otf.list_between n (otf.nextbvar e') ++
                                  otf.vars2nat (bvars e))>> /\
  <<GT: is_good_thunk e -> is_good_thunk e'>> /\
  <<GE: good_exp e -> good_exp e'>>.
Proof.
  i; destruct e; vauto; s in E; des
  ; destruct bx; try (by ss)
  ; destruct e1; vauto
  ; destruct e2; vauto
  ; destruct e2_1; vauto
  ; destruct bx0; vauto
  ; destruct alpha_eq; vauto
  ; destruct (list_mem _ _ _) eqn:IN; try (by desf)
  ; destruct WF as [WF0 WF1]
  ; destruct WF1 as [WF1 WF2].

  rewrite ->list_mem_true in IN; s in IN.
  splits; vauto.
  - otf.rnbv; s in MX; otf.ssbvar; splits; unfold next in *; try omega
    ; []; by rewrite <-otf.nextbvar_equiv in *
    ; rewrite replace_nextbvar_equiv; auto.
  - s; splits; vauto.
    apply replace_wf; vauto.
  - otf.svbv; simpl map in *; rewrite ->!map_app in *.
    rewrite ->otf.lb_nil; s
    ; [| by otf.rnbv; ss; rewrite <-!otf.nextbvar_equiv in *
         ; rewrite replace_nextbvar_equiv; auto].
    rewrite replace_bvars; auto.
Qed.


Lemma contify_subst: forall lw ea eb k G n n' phia phib
    (WF: wfexp G eb)
    (E: subst_cont k G n eb  = Some (ea,n'))
    (MX: (otf.nextbvar eb) <= n)
    (UN: forall m (BIG: n <= m),
           uniq (m :: otf.vars2nat (bvars eb) ++ otf.vars2nat (purge_lbl G)))
    (SLW: stable lw),
  <<MAX: max (otf.nextbvar ea) n <= n' >> /\
  <<SIM: mm_test.mm_psim lw G ren_id phia phib ea eb>>.
Proof.
  i; destruct eb; vauto.

  
  destruct eb1; destruct bx; vauto.
  destruct eb2; vauto.
  unfold subst_cont in E; destruct eb2_1; destruct bx; vauto.

  destruct (alpha_eq _ _ _) eqn:AEQ; vauto
  ; destruct (list_mem _ _ _) eqn:INk; try (by desf); clarify.
  rewrite ->list_mem_true in INk; s in INk.
  splits; [by otf.rnbv; s in MX; otf.ssbvar; splits; unfold next in *
           ; try omega; []; by rewrite <-otf.nextbvar_equiv in *
           ; rewrite replace_nextbvar_equiv; auto|].
  s in WF; des.

  
  red; splits; try done.
  eexists; i; unfold esim.
  apply pesim_acquire; i.
  apply pesim_real; i.

  rrdes HR. dup WHX. apply mm_confrel_shape in WHX. rrdes WHX.
  subst; clarify; dup RLa; dup RLb.
  rrdes RLa; rrdes RLb; inversion Em; inversion Em0.
  destruct hay, hby; destruct ee, ee0; inv Ec; inv Ec0; subst.
  clean; simpl cplus in *; noemp.
  destruct e, e0; s; clarify.
  rename e1 into enva, e3 into envb.

  
  eapply (pesim_dtsteps (n := 1)).
  - eassumption.
  - econs; first by eapply mid_wdet. inst_pairs; econs; by econs.
    econs; first by eapply mid_wdet. inst_pairs; econs; by econs. 
    by econs.
  - i; econs; esplits; [| |exact PV| exact FIN]; inst_pairs; last reflexivity
    ; eauto.
  - eauto.
  - i. rrdes RLb; subst.
    destruct cfb; destruct ee; inv Ec1; subst.
    clean; simpl cplus in *; noemp.
    destruct e; s; clarify.
    rename e0 into envb.

    
    assert (NINc: ~ In (inr c) G).
    {
     
      specialize (UN _ (le_refl _)). s in UN; des.
      clear UN UN0 UN1 UN2 UN3.
      rename UN4 into UN; otf.svbv; simpl map in UN; rewrite !map_app in UN.
      rewrite -app_comm_cons in UN.
      eapply uniq_perm in UN; first last.
      - eapply Permutation_sym; eapply Permutation_cons_app.
        eapply Permutation_trans; [| eapply Permutation_app]
        ; [| auto | eapply Permutation_app_comm].
        apply three_way_perm.
      - s in UN; des; ii; apply UN; apply in_or_app; left.
        apply (in_map varid) with (x:=inr c); auto using purge_lbl_c.
    }
    assert (NIN: ~ In (inl (inr t)) G).
    {
      specialize (UN _ (le_refl _)). s in UN; des.
      clear UN UN1 UN2 UN3 UN4.
      rename UN0 into UN; otf.svbv; simpl map in UN; rewrite !map_app in UN.
      ii; apply UN; repeat right; repeat (apply in_or_app; right).
      apply (in_map varid) with (x:=inl (inr t)); auto using purge_lbl_t.
    }
    specialize (UN _ (le_refl _)).
    rewrite cons_app in UN.
    apply uniq_appD in UN; des.
    clear UN UN1; rename UN0 into UN.
    otf.svbv; simpl map in *; rewrite ->!map_app in UN; otf.svbv.
    assert (UN2: uniq (inl (inr t) :: inl (inr bf) :: inl (inr bx0) ::
                       inr bk :: bvars eb1 ++
                       inr c :: inl (inr bx1) :: bvars eb2_1 ++
                       bvars eb2_2 ++ purge_lbl G)).
    {
      assert (t::bf::bx0::bk::map varid (bvars eb1) ++
              c::bx1::map varid (bvars eb2_1) ++
              map varid (bvars eb2_2) ++
              map varid (purge_lbl G) =
              map varid (inl (inr t) :: inl (inr bf) :: inl (inr bx0) ::
                         inr bk :: bvars eb1 ++
                         inr c :: inl (inr bx1) :: bvars eb2_1 ++
                         bvars eb2_2 ++ purge_lbl G)).
      {
        repeat (s; rewrite !map_app); vauto.
      }
      rewrite H in UN.
      eapply uniq_mapD in UN; eauto.
    }
    clear UN; rename UN2 into UN.

    
    remember ((inl (inr t), (IEfix bf bx0 bk eb1)) :: phib) as phib'.
    remember ((inr c, (IEcont bx1 eb2_1)) :: phia) as phia'.
    remember (inr c :: inl (inr t) :: G) as G'.
    hexploit (@replace_sound lw G' eb2_2 eb1 eb2_1 phia' phib' bf bx0 bk t k
               c bx1)
    ; vauto; last first.
    i. rrdes H. unfold esim in H1.
    eexists ,((hb' ## h2), Just _),. dup RLb0; rrdes RLb0.
    destruct XTR as [Xh]; subst. inv Ec1; clean.
    esplits; first 1 last.
    
    left; econs 2; econs; inst_pairs; vauto.
    inst_pairs; red; unfold NW; repeat (eexists; eauto).

    rewrite mid_csplit; rewrite (mid_csplit (hb' ## h2)).
    eapply pesim_release; first last; vauto.
    rr; esplits; eauto.

    
    assert (RVf: vclos (wlift lw) s (g s) (IVclo enva (IEfix bf bx0 bk eb1))
                       (IVclo envb (IEfix bf bx0 bk eb1))).
    {
      simpl in WF; des
      ; hexploit (@mm_compat_fix lw G ren_id bf bf bx0 bx0 bk bk eb1 eb1)
      ; auto using extendable_t_same, ren_var_id.
      - repeat (rewrite ren_id_self_t); rewrite ren_id_self_c
        ; auto using mm_psim_refl.
      - intro H; rrdes H; specialize (H _ _ _ _ GK RV _ eq_refl); des; simpls
        ; by clarify.
    }

    eapply H1; clear H1; vauto.
    + apply respects_weaken; auto
      ; [apply respects_cons |  | ]
      ; [ exact RPa | vauto |  |  | ii; ss; des; vauto].
      {
        ii; splits; ii; clarify
        ; [ specialize (RPa _ _ IN); inversion RPa; des; vauto
          | eauto using sublist_in, defbefore_sublist].
      }
      {
        ii; splits; ii; clarify; specialize (RPa _ _ IN); inversion RPa; des
        ; [vauto |rewrite <- defbefore_cons in *; vauto
                  ; eauto using sublist_in, defbefore_sublist].
      }
    + apply respects_cons; [ apply respects_weaken; auto | vauto
                               | ].
      {
        ii; splits; ii; clarify
        ; [ specialize (RPb _ _ IN); inversion RPb; des; vauto
          | eauto using sublist_in, defbefore_sublist].
      }
      {
        
        ii; splits; ii; clarify; try (destruct (LEM (inl (inr t) = y)))
        ; try (by clarify; rewrite list_assoc_cons_eq in IN; clarify
               ; rewrite ->defbefore_des in *; vauto)
        ; rewrite list_assoc_cons_neq in IN; vauto
        ; specialize (RPb _ _ IN); inversion RPb; des
        ; try (rewrite <- defbefore_cons in *; vauto)
        ; eauto using sublist_in, defbefore_sublist.
      }
    + i. simpl in IN, WF. rrdes IN.
      * clarify.
      * clarify; esplits.
        rewrite ren_itvar_id; rewrite list_assoc_cons_neq
        ; [apply list_assoc_cons_eq | discriminate].
        rewrite list_assoc_cons_neq
        ; [apply list_assoc_cons_eq | discriminate].
        auto.
      * specialize (RV _ IN); des; exists va vb; splits.
        rewrite ->ren_itvar_id in *; rewrite list_assoc_cons_neq
        ; [|ii;clarify; vauto]; []; rewrite list_assoc_cons_neq
        ; [assumption | ii; clarify].
        repeat (rewrite list_assoc_cons_neq); try assumption
        ; ii; clarify; vauto.
        assumption.
    + i; simpl in IN; rrdes IN.
      * rewrite ren_contvar_id. rewrite IN. esplits; vauto.
        hexploit (@mm_compat_cont lw (inl (inr t) :: G)
                   ren_id bx1 bx1 eb2_1 eb2_1)
        ; auto using extendable_t_same, ren_var_id.
        simpl in WF; des.
        rewrite ren_id_self_t; auto using mm_psim_refl.
        i; rrdes H.
        remember (((inl (inr t),
                    IVclo enva (IEfix bf bx0 bk eb1)) :: enva)) as enva'.
        remember ((inl (inr t),
                   IVclo envb (IEfix bf bx0 bk eb1)) :: envb) as envb'.
        assert (RV1: forall x (IN: In (inl x) (inl (inr t) :: G)),
                       exists va vb,
                         << INa: lookup (inl (rename_itvar ren_id x)) enva' =
                             Some va >> /\
                         << INb: lookup (inl x) envb' = Some vb >> /\
                         << SIM: vclos (wlift lw) s (g s) va vb >>).
        {
          i; apply in_consD in IN; rrdes IN; [esplits; clarify; auto|].
          - specialize (RV _ IN); des; exists va vb; splits; clarify.
            rewrite ->ren_itvar_id in *; rewrite list_assoc_cons_neq
            ; [assumption|ii;clarify; vauto].
            repeat (rewrite list_assoc_cons_neq); try assumption
            ; ii; clarify; vauto.
        }
        assert (RK1: forall x (IN: In (inr x) (inl (inr t) :: G)),
                       exists ka kb,
               << INa: lookup (inr (rename_contvar ren_id x)) enva' =
                    Some ka >> /\
               << INb: lookup (inr x) envb' = Some kb >> /\
               << SIM: ksim (wlift lw) (wfo_nat i) kia kib g si s s ka kb >>).
        {
          i; apply in_consD in IN; rrdes IN; [esplits; clarify; auto|].
          specialize (RK _ IN); des; exists ka kb; splits; clarify.
        }
        specialize (H  _ _ _ _ (wfo_nat i)  _ _ _ _
                       bx1 eb2_1 GK EXT RV1 RK1 eq_refl).
        des; simpls; clarify; eauto.
      * discriminate.
      * specialize (RK _ IN); des; exists ka kb; splits.
        rewrite ->ren_contvar_id in *. rewrite list_assoc_cons_neq.
        rewrite list_assoc_cons_neq. assumption.
        ii; clarify.
        ii; clarify; vauto.
        repeat (rewrite list_assoc_cons_neq). assumption.
        ii; clarify.
        ii; clarify; vauto.
        eassumption.
    + rewrite <-defbefore_cons; vauto; rewrite defbefore_des; vauto.
    + simpl bvars; simpl purge_lbl.
      eapply (uniq_perm UN); crush_perm.
    + 
      eapply uniq_perm in UN
      ; [|instantiate (1 := bvars eb2_2 ++ [inr c] ++ [inl (inr bx1)] ++
                            bvars eb2_1 ++
                            [inl (inr t)] ++
                            [inl (inr bf); inl (inr bx0);
                             inr bk] ++ bvars eb1 ++  purge_lbl G)]
      ; s.
      apply (uniq_sublist UN); s; apply sublist_app
      ; auto 9 using sublist_cons_keep, sublist_cons_drop, sublist_r_app_r.
      crush_perm.
    + rewrite <-defbefore_cons; vauto; rewrite defbefore_des; vauto.
    + intros IN; s in IN; des; vauto.
      { simpls; des; apply UN; do !right; apply in_or_app; right
        ; simpls; right; vauto.
      }
      {
        ss; des; apply uniq_appD in UN3; des.
        s in UN4; des.
        apply UN6; apply in_or_app; right; apply in_or_app
        ; auto using purge_lbl_t.
      }
    + rewrite defbefore_des; vauto. ii; simpls; des; vauto.
    + i; unfold fresh. intros y e2 LU; destruct (LEM (y = inl (inr t))).
      {
        clarify. rewrite list_assoc_cons_eq in LU; vauto.
        splits.
        - ii; vauto. ss; des. apply UN; do !right; apply in_or_app; right.
          rewrite !app_comm_cons; apply in_or_app; right
          ; apply in_or_app; vauto.
        - rewrite <-defbefore_cons; vauto; rewrite defbefore_des; vauto.
          ii; ss; des. apply uniq_appD in UN3; des; ss; des.
          apply uniq_appD in UN7; des; apply uniq_appD in UN8; des.
          specialize (UN11 _ IN); eauto using purge_lbl_nolbl, bvars_nolbl.
      }
      {
        rewrite list_assoc_cons_neq in LU; [| by ii; vauto].
        apply RPb in LU; des; splits.
        - ii; clarify; ss; des; apply uniq_appD in UN3; des; s in UN4; des
          ; apply uniq_appD in UN7; des; apply uniq_appD in UN8; des.
          specialize (UN11 _ IN); eauto using purge_lbl_nolbl, bvars_nolbl.
        - do !(rewrite <- defbefore_cons; vauto).
          ii; clarify; ss; des; apply uniq_appD in UN3; des; s in UN4; des
          ; apply uniq_appD in UN7; des; apply uniq_appD in UN8; des.
          specialize (UN11 _ IN). apply UN11; apply purge_lbl_nolbl
          ; eauto using defbefore_sublist, bvars_nolbl, sublist_in.
      }
    + i; unfold fresh. intros y e LU; destruct (LEM (y = inr c)).
      {
        clarify. rewrite list_assoc_cons_eq in LU; vauto.
        splits.
        - ii; vauto; s in UN; des; apply uniq_appD in UN3; des; s in UN4; des.
          apply UN4; right; apply in_or_app; right; apply in_or_app; vauto.
        - rewrite defbefore_des; vauto; [| by ii; simpls; des; vauto]
          ; ii; simpls; des; clarify.
          + apply UN; do !right; apply in_or_app; right; s; do !right
            ; apply in_or_app; right; apply in_or_app; vauto.
          + ii; ss; des; apply uniq_appD in UN3; des; ss; des.
            apply uniq_appD in UN7; des; apply uniq_appD in UN8; des.
            specialize (UN11 _ IN); eauto using purge_lbl_nolbl, bvars_nolbl.
      }
      {
        rewrite list_assoc_cons_neq in LU; [| by ii; vauto].
        apply RPa in LU; des; splits.
        - ii; clarify; ss; des; apply uniq_appD in UN3; des; s in UN4; des
          ; apply uniq_appD in UN7; des; apply uniq_appD in UN8; des.
          specialize (UN11 _ IN); eauto using purge_lbl_nolbl, bvars_nolbl.
        - do !(rewrite <- defbefore_cons; vauto).
          ii; clarify; ss; des; apply uniq_appD in UN3; des; s in UN4; des
          ; apply uniq_appD in UN7; des; apply uniq_appD in UN8; des.
          specialize (UN11 _ IN). apply UN11; apply purge_lbl_nolbl
          ; eauto using defbefore_sublist, bvars_nolbl, sublist_in.
      }
    + intros y DB; rewrite <- defbefore_cons in DB; vauto
      ; rewrite defbefore_des in DB; vauto; rewrite defbefore_des
      ; vauto; []; ii; simpls; des; vauto.
Grab Existential Variables.
  exact IVunit. exact IVunit. exact 0.
Qed.



Inductive ContifyType :=
  | ContifyAdd
  | ContifySubst.

Definition contify_ann: otf.annT.
  refine {| otf._annT := ContifyType * contvar
         ;  otf.perform u :=
              match u with
              | (ContifyAdd, k) => add_cont k
              | (ContifySubst, k) => subst_cont k
              end
         |}.
 - ii; unnw; desf; by [eapply contify_add_wf; eauto
                      |eapply contify_subst_wf; eauto].
 - ii; unnw; desf; eapply otf.conj_app; try (by econs; exact H)
   ; by [eapply contify_add; eauto|eapply contify_subst; eauto].
Defined.


Fixpoint analyze (e : exp)
         : otf.aexp (otf._annT (otf.annT_list contify_ann)) :=
  match e with
  | IElet (IEfix bf y bk eb) (inl bx) ea
    => let anea := analyze ea in
       let aneb := analyze eb in
       match is_contifiable ea (inr bx) with
       | Some k =>
           let annot := [(ContifySubst, k); (ContifyAdd, k)] in
            (otf.IElet annot (otf.IEfix bf y bk aneb) (inl bx) anea)
       | _ =>
           (otf.IElet nil (otf.IEfix bf y bk aneb) (inl bx) anea)
       end
  | IElet ea bx eb
    => let anea := analyze ea in
       let aneb := analyze eb in
       (otf.IElet nil anea bx aneb)
  | IEnew x bx ea
    => let anea := analyze ea in
       (otf.IEnew x bx anea)
  | IEread x bx ea
    => let anea := analyze ea in
       (otf.IEread x bx anea)
  | IEwrite x y ea
    => let anea := analyze ea in
       (otf.IEwrite x y anea)
  | IEinput bx ea
    => let anea := analyze ea in
       (otf.IEinput bx anea)
  | IEoutput x ea
    => let anea := analyze ea in
       (otf.IEoutput x anea)
  | IEifnz x ea eb
    => let anea := analyze ea in
       let aneb := analyze eb in
       (otf.IEifnz x anea aneb)
  | IEcase x bx ea eb
    => let anea := analyze ea in
       let aneb := analyze eb in
       (otf.IEcase x bx anea aneb)
  | IEret xk k
    => (otf.IEret _ xk k)
  | IEapp xf x xk
    => (otf.IEapp nil xf x xk)
  | IEinst xf xk
    => (otf.IEinst nil xf xk)
  | IEpair x y
    => (otf.IEpair _ x y)
  | IEinl x
    => (otf.IEinl _ x)
  | IEinr x
    => (otf.IEinr _ x)
  | IEbop bop x y
    => (otf.IEbop _ bop x y)
  | IEfst x
    => (otf.IEfst _ x)
  | IEsnd x
    => (otf.IEsnd _ x)
  | IEfix bf bx bk ea
    => let anea := analyze ea in
       (otf.IEfix bf bx bk anea)
  | IEgen bk ea
    => let anea := analyze ea in
       (otf.IEgen bk anea)
  | IEcont bx ea
    => let anea := analyze ea in
       (otf.IEcont bx anea)
  | IEunit
    => (otf.IEunit _)
  | IEnat n
    => (otf.IEnat _ n)
  | IErefeq x y
    => (otf.IErefeq _ x y)
  end.

Definition contify_mod (G: list var) (m : module) : module :=
  otf.transform_mod analyze G (otf.nextvar m) m.

Lemma analyze_aerase: forall e,
  otf.aerase (analyze e) = e.
Proof.
  induction e; s; f_equal; auto.
  desf; s; f_equal; auto.
Qed.

Theorem contify_wfmod: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: forall n' (BIG: (otf.nextvar m) <= n'),
           uniq (n' :: otf.vars2nat (bvars_mod m))),
  mid_wfmod FI (contify_mod (map (inl<*>inl) FI) m) FO /\
  uniqvar (contify_mod (map (inl<*>inl) FI) m).
Proof.
  i; eapply otf.transform_wfmod; eauto using analyze_aerase.
Qed.

Theorem contify_freshen_wfmod: forall FI FO m
    (WF: mid_wfmod FI m FO),
  mid_wfmod FI (contify_mod (map (inl<*>inl) FI) (freshen_mod m)) FO /\
  uniqvar (contify_mod (map (inl<*>inl) FI) (freshen_mod m)).
Proof.
  i; unfold freshen_mod; hexploit freshen_uniq_aux; eauto.
  i; match goal with
     [H: context[freshen_mod' ?ne _] |- context[freshen_mod' ?n _]] =>
       unify ne n
     end; des; hexploit freshen_wfmod; eauto; []; clear WF; intros WF.
  apply contify_wfmod; auto.
  ii; s; splits; eauto using otf.nextvar_nin_mod.
Qed.

Theorem contify_msim: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: forall n' (BIG: (otf.nextvar m) <= n'),
           uniq (n' :: otf.vars2nat (bvars_mod m))),
  rtc (msim mm_gw FI FO) (contify_mod (map (inl<*>inl) FI) m) m.
Proof.
  i; eapply otf.transform_msim; eauto using analyze_aerase.
Qed.

Theorem contify_freshen_msim: forall FI FO m
    (WF: mid_wfmod FI m FO),
  rtc (msim mm_gw FI FO) (contify_mod (map (inl<*>inl) FI) (freshen_mod m))
                         (freshen_mod m).
Proof.
  i; unfold freshen_mod; hexploit freshen_uniq_aux; eauto.
  i; match goal with
     [H: context[freshen_mod' ?ne _] |- context[freshen_mod' ?n _]] =>
       unify ne n
     end; des; hexploit freshen_wfmod; eauto; []; clear WF; intros WF.
  apply contify_msim; auto.
  ii; s; splits; eauto using otf.nextvar_nin_mod.
Qed.

Theorem freshen_contify_correct:
  <<WF: modtf_wf
          (fun FI => contify_mod (map (inl <*> inl) FI) <*> freshen_mod)>> /\
  <<SIM: modtf_sim
           (fun FI => contify_mod (map (inl <*> inl) FI) <*> freshen_mod)>>.
Proof.
  split; ii.
  - eapply contify_freshen_wfmod; auto.
  - econs 3; [by eapply contify_freshen_msim; eauto|].
    eauto using freshen_msim.
Qed.
