(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import ssr countable vgtac basis lang_common lang_spec lang_mid lang_mid_renaming.
Require Import model_common model_untyped model_properties_untyped gw_untyped.
Require Import mm.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".


(** Rename all bound variables in order to (re-)establish the
    unique-variables condition.
*)


(** * Definition. *)


Fixpoint freshen n (tren: @tmap termvar _  termvar)
                   (cren: @tmap contvar _ contvar) (e: exp) : exp * nat :=
  match e with
    | IEunit => (IEunit, n)
    | IEnat m => (IEnat m, n)
    | IEpair x y => (IEpair (Inr tren x) (Inr tren y), n)
    | IEinl x => (IEinl (Inr tren x), n)
    | IEinr x => (IEinr (Inr tren x), n)
    | IEbop op x y => (IEbop op (Inl (Inr tren) x) (Inl (Inr tren) y), n)
    | IEfst x => (IEfst (Inr tren x), n)
    | IEsnd x => (IEsnd (Inr tren x), n)
    | IEfix bf bx bk e =>
      let (e', n') := freshen (3+n) ([1+n//bx]>>[n//bf]>>tren) ([2+n//bk]>>cren) e in
      (IEfix n (1+n) (2+n) e', n')
    | IEgen bk e =>
      let (e', n') := freshen (1+n) tren ([n//bk]>>cren) e in
      (IEgen n e', n')
    | IEcont bx e =>
      let (e', n') := freshen (1+n) ([n//bx]>>tren) cren e in
      (IEcont n e', n')
    | IEnew x bx e =>
      let (e', n') := freshen (1+n) ([n//bx]>>tren) cren e in
      (IEnew (Inr tren x) n e', n')
    | IEread x bx e =>
      let (e', n') := freshen (1+n) ([n//bx]>>tren) cren e in
      (IEread (Inr tren x) n e', n')
    | IEwrite x y e =>
      let (e', n') := freshen n tren cren e in
      (IEwrite (Inr tren x) (Inr tren y) e', n')
    | IEinput bx e =>
      let (e', n') := freshen (1+n) ([n//bx]>>tren) cren e in
      (IEinput n e', n')
    | IEoutput x e =>
      let (e', n') := freshen n tren cren e in
      (IEoutput (Inr tren x) e', n')
    | IEifnz x ea eb =>
      let (ea', n') := freshen n tren cren ea in
      let (eb', n'') := freshen n' tren cren eb in
      (IEifnz (Inr tren x) ea' eb', n'')
    | IEcase x bx ea eb =>
      let (ea', n') := freshen (1+n) ([n//bx]>>tren) cren ea in
      let (eb', n'') := freshen n' ([n//bx]>>tren) cren eb in
      (IEcase (Inr tren x) n ea' eb', n'')
    | IEret k x => (IEret (cren k) (Inr tren x), n)
    | IEapp f x k => (IEapp (Inr tren f) (Inr tren x) (cren k), n)
    | IEinst g k => (IEinst (Inr tren g) (cren k), n)
    | IElet ea bxk eb =>
      let '(tren', cren', nn) :=
          match bxk with
            | inl bx => ([n//bx]>>tren, cren, inl n)
            | inr bk => (tren, [n//bk]>>cren, inr n)
          end in
      let (ea', n') := freshen (1+n) tren cren ea in
      let (eb', n'') := freshen n' tren' cren' eb in
      (IElet ea' nn eb', n'')
    | IErefeq x y => (IErefeq (Inr tren x) (Inr tren y), n)
  end.

Fixpoint freshen_mod' n (m: module) : module :=
  match m with
    | nil => nil
    | (l, e) :: m' =>
      let (e', n') := freshen n id id e in
      (l, e') :: freshen_mod' n' m'
  end.

Definition freshen_mod: module -> module :=
  freshen_mod' 0.



(** * Syntactic correctness. *)


(** ** Preliminaries. *)

Definition varid (x: var) : nat :=
  match x with
    | inl (inl n)
    | inl (inr n)
    | inr n
      => n
  end.

Inductive renG: (termvar -> termvar) -> (contvar -> contvar) ->
                list var -> list var -> Prop :=
| renG_nil G
  : renG id id G G
| renG_cons_t tren cren G G' x y
    (RG: renG tren cren G G')
  : renG ([y//x]>>tren) cren (inl (inr x) :: G) (inl (inr y) :: G')
| renG_cons_c tren cren G G' x y
    (RG: renG tren cren G G')
  : renG tren ([y//x]>>cren) (inr x :: G) (inr y :: G')
| renG_cons_purge tren cren G G' x y
    (RG: renG tren cren (purge G) (purge G'))
  : renG tren ([y//x]>>cren) (purge G) (purge G')

.
Hint Constructors renG.

Lemma renG_In_c: forall x tren cren G G'
    (IN: In (inr x) G)
    (RG: renG tren cren G G'),
  In (inr (cren x)) G'.
Proof.
  i; ginduction RG; i.
  - destruct x; auto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
    s; rewrite tmext_neq; [|by ii; vauto].
    apply IHRG in IN; eauto.
  - exfalso; eauto using purge_inDr.
Qed.

Lemma renG_In_t: forall x tren cren G G'
    (IN: In (inl x) G)
    (RG: renG tren cren G G'),
  In (inl (Inr tren x)) G'.
Proof.
  i; ginduction RG; i.
  - destruct x; auto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
    destruct x0; vauto.
    + apply IHRG in IN; s in IN; eauto.
    + s; rewrite tmext_neq; [|by ii; vauto].
      apply IHRG in IN; eauto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
  - eauto.
Qed.

Lemma renG_purge: forall tren cren G G'
    (RG: renG tren cren G G'),
  renG tren cren (purge G) (purge G').
Proof.
  i; ginduction G; i; inv RG; s; eauto; rewrite ?purge_purge; auto.
Qed.


(** ** Uniqueness. *)

Lemma freshen_uniq_: forall e e' n n' tren cren
    (E: freshen n tren cren e = (e', n')),
  <<UN: uniq (map varid (bvars e'))>> /\
  <<WI: within n n' (map varid (bvars e'))>>.
Proof.
  induction e; i; s in E; clarify
  ; try (by destruct (freshen _ _ _ _) eqn:?; clarify
         ; hexploit IHe; eauto)
  ; try (by destruct (freshen _ _ _ _) eqn:?; clarify
         ; hexploit IHe; eauto; []; i; des
         ; simpl bvars; split; rewrite cons_app ?map_app; simpl map
         ; eauto using uniq_app_within, within_app).
  - destruct (freshen _ _ _ _) eqn:?; clarify.
    hexploit IHe; eauto; []; i; des.
    simpl bvars; split; rewrite cons_app3 map_app; simpl map
    ; eauto using uniq_app_within, within_app.
  - destruct (freshen _ _ _ e1) eqn:?; clarify.
    destruct (freshen _ _ _ e2) eqn:?; clarify.
    hexploit IHe1; eauto; []; i; des.
    hexploit IHe2; eauto; []; i; des.
    simpl bvars; split; rewrite map_app; simpl map
    ; eauto using uniq_app_within, within_app.
  - destruct (freshen _ _ _ e1) eqn:?; clarify.
    destruct (freshen _ _ _ e2) eqn:?; clarify.
    hexploit IHe1; eauto; []; i; des.
    hexploit IHe2; eauto; []; i; des.
    simpl bvars; split; rewrite cons_app ?map_app; simpl map
    ; eauto using uniq_app_within, within_app.
  - destruct bx
    ; (destruct (freshen _ _ _ e1) eqn:?; clarify
       ; destruct (freshen _ _ _ e2) eqn:?; clarify
       ; hexploit IHe1; eauto; []; i; des
       ; hexploit IHe2; eauto; []; i; des
       ; simpl bvars; split; rewrite cons_app ?map_app; simpl map
       ; eauto using uniq_app_within, within_app).
Qed.

Lemma freshen_uniq_aux: forall FI m FO n
    (WF: mid_wfmod FI m FO),
  uniq (map varid (bvars_mod (freshen_mod' n m))) /\
  forall x (IN: In x (map varid (bvars_mod (freshen_mod' n m)))), n <= x.
Proof.
  i; ginduction WF; vauto.
  s; i; destruct (freshen _ _ _ _) eqn:E.
  apply freshen_uniq_ in E; []; des.
  hexploit (IHWF n0); []; i; des.
  split; s; rewrite map_app; rrdes WI.
  - eapply uniq_app_le; eauto.
  - i; apply in_app_or in IN; des; eauto using le_trans.
Qed.

Lemma freshen_uniq: forall FI m FO
    (WF: mid_wfmod FI m FO),
  uniqvar (freshen_mod m).
Proof.
  i; eapply (uniq_mapD _ varid).
  eapply freshen_uniq_aux; eauto.
Qed.


(** ** Well-formedness. *)

Lemma freshen_good_thunk: forall e e' n n' F tren cren
    (IGT: is_good_thunk e)
    (WF: wfatom F e)
    (E: freshen n tren cren e = (e', n')),
  is_good_thunk e'.
Proof.
  i; destruct e; vauto
  ; s in E; destruct (freshen _ _ _ _) eqn:E'; clarify; ii; inv FH.
  destruct e; vauto; s in E'; try destruct bx
  ; repeat destruct (freshen _ _ _ _) in E'; vauto.
Qed.

Lemma freshen_wfatom_: forall e e' G G' n n' tren cren
    (WF: wfatom G e)
    (E: freshen n tren cren e = (e', n'))
    (RG: renG tren cren G G')
    (IH: forall e', size e' < size e -> forall e'' G G' n n' tren cren,
           wfexp G e' -> freshen n tren cren e' = (e'', n') ->
           renG tren cren G G' -> wfexp G' e''),
  wfatom G' e'.
Proof.
  i; destruct e; s in E; vauto
  ; try (by s; s in WF; des; splits; eauto using renG_In_t, renG_In_c).
  - destruct x, y; s; s in WF; des; splits; eauto using renG_In_t, renG_In_c.
  - s in WF.
    destruct (freshen _ _ _ _) eqn:E'; clarify.
    eapply IH; s; eauto using renG_purge; omega.
  - s in WF.
    destruct (freshen _ _ _ _) eqn:E'; clarify.
    eapply IH; s; eauto using renG_purge; omega.
  - s in WF.
    destruct (freshen _ _ _ _) eqn:E'; clarify.
    eapply IH; s; eauto using renG_purge; omega.
Qed.

Lemma freshen_wfexp: forall e e' G G' n n' tren cren
    (WF: wfexp G e)
    (E: freshen n tren cren e = (e', n'))
    (RG: renG tren cren G G'),
  wfexp G' e'.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i
  ; try (by s in WF; des; s in E; destruct (freshen _ _ _ _) eqn:E'; clarify
         ; s; splits; eauto using renG_In_t; eapply IH; s; eauto)
  ; try (by s in WF; des; s in E; clarify
         ; econs; eauto using renG_In_t, renG_In_c).
  - s in WF; des; s in E.
    destruct (freshen _ _ _ _) eqn:E1; destruct (freshen _ _ _ e2) eqn:E2.
    clarify; s; splits; eauto using renG_In_t.
    + eapply (IH e1); s; eauto; omega.
    + eapply (IH e2); s; eauto; omega.
  - s in WF; des; s in E.
    destruct (freshen _ _ _ _) eqn:E1; destruct (freshen _ _ _ e2) eqn:E2.
    clarify; s; splits; eauto using renG_In_t.
    + eapply (IH e1); s; eauto; omega.
    + eapply (IH e2); s; eauto; omega.
  - destruct bx.
    + s in WF; des; s in E.
      destruct (freshen _ _ _ _) eqn:E1; destruct (freshen _ _ _ e2) eqn:E2.
      clarify; s; splits.
      * eapply freshen_wfatom_; eauto.
        i; eapply IH; eauto; s; omega.
      * eapply (IH e2); s; eauto; omega.
    + s in WF; des; s in E.
      destruct (freshen _ _ _ _) eqn:E1; destruct (freshen _ _ _ e2) eqn:E2.
      clarify; s; splits.
      * eapply freshen_wfatom_; eauto.
        i; eapply IH; eauto; s; omega.
      * eapply (IH e2); s; eauto; omega.
Qed.

Lemma freshen_wfatom: forall G G' e e' n n' tren cren
    (WF: wfatom G e)
    (E: freshen n tren cren e = (e', n'))
    (RG: renG tren cren G G'),
  wfatom G' e'.
Proof.
  i; eapply freshen_wfatom_; eauto using freshen_wfexp.
Qed.

Lemma freshen_wfmod: forall FI m FO
    (WF: mid_wfmod FI m FO),
  mid_wfmod FI (freshen_mod m) FO.
Proof.
  unfold freshen_mod; generalize 0.
  i; ginduction WF; vauto.
  s; i; destruct (freshen _ _ _ _) eqn:E; econs
  ; eauto using freshen_good_thunk, freshen_wfatom.
Qed.



(** * Semantic correctness. *)


(** ** Preliminaries. *)

Definition mkren tren cren :=
  {| rename_termvar := inr <*> tren
   ; rename_contvar := cren
  |}.

Lemma mkren_ext_t: forall tren cren x y,
  (mkren tren cren)[inr y/x] = mkren ([y//x]>>tren) cren.
Proof.
  i; unfold mkren, ren_extend_t; f_equal.
  extensionalities; rewrite tmext_lift; auto.
Qed.

Lemma mkren_ext_c: forall tren cren x y,
  (mkren tren cren)[c y/x] = mkren tren ([y//x]>>cren).
Proof.
  i; destruct x; auto.
Qed.

Lemma extendable_t_fresh: forall G tren cren x n
    (FRt: forall y, In (inl (inr y)) G -> tren y < n),
  extendable_t G (mkren tren cren) n x.
Proof.
  ii; exfalso; hexploit FRt; eauto; []; i; s in E; clarify; omega.
Qed.

Lemma extendable_c_fresh: forall G tren cren x n
    (FRc: forall y, In (inr y) G -> cren y < n),
  extendable_c G (mkren tren cren) n x.
Proof.
  ii; exfalso; hexploit FRc; eauto; []; i; s in E; clarify; omega.
Qed.

Lemma fresh_t_tmext_t: forall (G: list var) tren x y n
    (IN: In (inl (inr y)) (inl (inr x) :: G))
    (FRt: In (inl (inr y)) G -> tren y < n),
  [n//x]>>tren y < S n.
Proof.
  i; apply in_consD in IN; des.
  - clarify; tmext_simpl; omega.
  - rewrite tmext_neq; [|by ii; vauto].
    hexploit FRt; eauto.
Qed.

Lemma fresh_t_tmext_c: forall (G: list var) tren x y n
    (IN: In (inl (inr y)) (inr x :: G))
    (FRt: In (inl (inr y)) G -> tren y < n),
  tren y < S n.
Proof.
  i; apply in_consD in IN; des; vauto.
  hexploit FRt; eauto.
Qed.

Lemma fresh_c_tmext_c: forall (G: list var) cren x y n
    (IN: In (inr y) (inr x :: G))
    (FRc: In (inr y) G -> cren y < n),
  [n//x]>>cren y < S n.
Proof.
  i; apply in_consD in IN; des.
  - clarify; tmext_simpl; omega.
  - rewrite tmext_neq; [|by ii; vauto].
    hexploit FRc; eauto.
Qed.

Lemma fresh_c_tmext_t: forall (G: list var) cren x y n
    (IN: In (inr y) (inl (inr x) :: G))
    (FRc: In (inr y) G -> cren y < n),
  cren y < S n.
Proof.
  i; apply in_consD in IN; des; vauto.
  hexploit FRc; eauto.
Qed.

Lemma freshen_le: forall n n' e e' tren cren
    (E: freshen n tren cren e = (e', n')),
  n <= n'.
Proof.
  i; hexploit freshen_uniq_; eauto; []; i; des; rrdes WI; auto.
Qed.


(** ** Actual proof. *)

Lemma freshen_asim_: forall lw eb G n n' tren cren ea
    (WF: wfatom G eb)
    (E: (ea, n') = freshen n tren cren eb)
    (FRt: forall x (IN: In (inl (inr x)) G), tren x < n)
    (FRc: forall x (IN: In (inr x) G), cren x < n)
    (STA: stable lw)
    (IH: forall eb', size eb' < size eb -> forall G n n' tren cren ea',
           wfexp G eb' -> (ea', n') = freshen n tren cren eb' ->
           (forall x : termvar, In (inl (inr x)) G -> tren x < n) ->
           (forall x : contvar, In (inr x) G -> cren x < n) ->
           stable lw -> mm_psim lw G (mkren tren cren) ea' eb'),
  mm_asim lw G (mkren tren cren) ea eb.
Proof.
  i; destruct eb; vauto
  ; try (by s in WF; des; simpl freshen in E; clarify; eauto with mm_compat).
  - destruct op, x, y; s in WF; des; simpl freshen in E; clarify; eauto with mm_compat.
  - simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
    s in WF; des; eapply mm_compat_fix; eauto.
    + eauto using extendable_t_fresh.
    + eapply extendable_t_fresh; i; eapply lt_le_trans; eauto.
    + omega.
    + rewrite 2!mkren_ext_t mkren_ext_c; eapply IH
      ; eauto 10 using fresh_t_tmext_c, fresh_t_tmext_t, purge_inD.
      i; eapply fresh_c_tmext_c; eauto.
      i; eapply fresh_c_tmext_t; eauto.
      i; eapply fresh_c_tmext_t; eauto.
      i; exfalso; eauto using purge_inDr.
  - simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
    s in WF; des; eapply mm_compat_gen; eauto.
    rewrite mkren_ext_c; eapply IH
    ; eauto 10 using fresh_t_tmext_c, fresh_t_tmext_t, purge_inD.
    i; eapply fresh_c_tmext_c; eauto.
    i; exfalso; eauto using purge_inDr.
Qed.

Lemma freshen_ksim_: forall lw eb G n n' tren cren ea
    (WF: wfatom G eb)
    (E: (ea, n') = freshen n tren cren eb)
    (FRt: forall x (IN: In (inl (inr x)) G), tren x < n)
    (FRc: forall x (IN: In (inr x) G), cren x < n)
    (STA: stable lw)
    (IH: forall eb', size eb' < size eb -> forall G n n' tren cren ea',
           wfexp G eb' -> (ea', n') = freshen n tren cren eb' ->
           (forall x : termvar, In (inl (inr x)) G -> tren x < n) ->
           (forall x : contvar, In (inr x) G -> cren x < n) ->
           stable lw -> mm_psim lw G (mkren tren cren) ea' eb'),
  mm_ksim lw G (mkren tren cren) ea eb.
Proof.
  ii; destruct eb; vauto.
  simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
  s in WF; eapply IH in WF; eauto; inst
  ; eauto using fresh_t_tmext_t, fresh_c_tmext_t.
  esplits; eauto.
  rewrite -mkren_ext_t in WF.
  hexploit mm_compat_cont; eauto using extendable_t_fresh; []; i.
  unfold mm_ksim in X; hexploit X; eauto; i; des; eauto.
Qed.

Theorem freshen_psim: forall lw eb G n n' tren cren ea
    (WF: wfexp G eb)
    (E: (ea, n') = freshen n tren cren eb)
    (FRt: forall x (IN: In (inl (inr x)) G), tren x < n)
    (FRc: forall x (IN: In (inr x) G), cren x < n)
    (STA: stable lw),
  mm_psim lw G (mkren tren cren) ea eb.
Proof.
  intros ? ?.
  induction eb as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct eb; vauto; i.
  - simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
    s in WF; des; eapply mm_compat_new; eauto using extendable_t_fresh.
    rewrite mkren_ext_t; eapply IH; eauto using fresh_t_tmext_t, fresh_c_tmext_t.
  - simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
    s in WF; des; eapply mm_compat_read; eauto using extendable_t_fresh.
    rewrite mkren_ext_t; eapply IH; eauto using fresh_t_tmext_t, fresh_c_tmext_t.
  - simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
    s in WF; des; eapply mm_compat_write; eauto.
  - simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
    s in WF; des; eapply mm_compat_input; eauto using extendable_t_fresh.
    rewrite mkren_ext_t; eapply IH; eauto using fresh_t_tmext_t, fresh_c_tmext_t.
  - simpl freshen in E; destruct (freshen _ _ _ _) eqn:E'; clarify.
    s in WF; des; eapply mm_compat_output; eauto.
  - simpl freshen in E; destruct (freshen _ _ _ eb1) eqn:E1
    ; destruct (freshen _ _ _ eb2) eqn:E2; clarify.
    s in WF; des; eapply mm_compat_ifnz; eauto.
    + eapply IH; eauto; s; omega.
    + eapply IH; eauto; first (s; omega)
      ; (i; eapply lt_le_trans; eauto using freshen_le with nocore).
  - simpl freshen in E; destruct (freshen _ _ _ eb1) eqn:E1
    ; destruct (freshen _ _ _ eb2) eqn:E2; clarify.
    s in WF; des; eapply mm_compat_case; eauto using extendable_t_fresh.
    + rewrite mkren_ext_t; eapply IH; eauto using fresh_t_tmext_t, fresh_c_tmext_t; s; omega.
    + rewrite mkren_ext_t; eapply IH; eauto; first (s; omega)
      ; (i; eapply lt_le_trans; eauto using fresh_t_tmext_t, fresh_c_tmext_t,
                                            freshen_le with nocore).
  - simpl freshen in E; clarify.
    s in WF; des; eapply mm_compat_ret; eauto.
  - simpl freshen in E; clarify.
    s in WF; des; eapply mm_compat_app; eauto.
  - simpl freshen in E; clarify.
    s in WF; des; eapply mm_compat_inst; eauto.
  - destruct bx.
    + simpl freshen in E; destruct (freshen _ _ _ eb1) eqn:E1
      ; destruct (freshen _ _ _ eb2) eqn:E2; clarify.
      s in WF; des; eapply mm_compat_let; eauto using extendable_t_fresh.
      * eapply freshen_asim_; eauto using lt_le_trans.
        i; eapply IH; eauto; s; omega.
      * rewrite mkren_ext_t; eapply IH; eauto; first (s; omega)
      ; (i; eapply lt_le_trans; eauto using fresh_t_tmext_t, fresh_c_tmext_t,
                                            freshen_le with nocore).
    + simpl freshen in E; destruct (freshen _ _ _ eb1) eqn:E1
      ; destruct (freshen _ _ _ eb2) eqn:E2; clarify.
      s in WF; des; eapply mm_compat_letcont; eauto using extendable_c_fresh.
      * eapply freshen_ksim_; eauto using lt_le_trans.
        i; eapply IH; eauto; s; omega.
      * rewrite mkren_ext_c; eapply IH; eauto; first (s; omega)
      ; (i; eapply lt_le_trans; eauto using fresh_c_tmext_c, fresh_t_tmext_c,
                                            freshen_le with nocore).
Qed.

Theorem freshen_msim: forall FI FO m
    (WF: mid_wfmod FI m FO),
  rtc (msim mm_gw FI FO) (freshen_mod m) m.
Proof.
  unfold freshen_mod; generalize 0.
  i; econs 1; ginduction WF; eauto using mm_msim_nil.
  i; hexploit IHWF; eauto; []; i.
  s; destruct (freshen _ _ _ _) eqn:E.
  eapply mm_msim_consB; eauto using lwemp_stable.
  - destruct e; vauto.
    + s in E; destruct (freshen _ _ _ _) eqn:EE; clarify.
      econs 1.
      * rewrite ?mkren_ext_t ?mkren_ext_c.
        s in WFE; rewrite -map_map purge_map_inl map_map in WFE.
        eapply freshen_psim; eauto using lwemp_stable.
        { i; apply in_consD in IN; des; vauto.
          i; apply in_consD in IN; des; vauto.
          { tmext_simpl; omega. }
          rewrite tmext_neq; [|by ii; vauto].
          i; apply in_consD in IN; des; vauto.
          { tmext_simpl; omega. }
          rewrite tmext_neq; [|by ii; vauto].
          
          apply in_map_iff in IN; des; vauto.
        }
        { i; apply in_consD in IN; des; vauto.
          { tmext_simpl; omega. }
          repeat (i; apply in_consD in IN; des; vauto).
          apply in_map_iff in IN; des; vauto.
        }
      * omega.
    + s in E; destruct (freshen _ _ _ _) eqn:EE; clarify.
      econs 2.
      * rewrite ?mkren_ext_t ?mkren_ext_c.
        s in WFE; rewrite -map_map purge_map_inl map_map in WFE.
        eapply freshen_psim; eauto using lwemp_stable.
        { i; apply in_consD in IN; des; vauto.
          apply in_map_iff in IN; des; vauto.
        }
        { i; apply in_consD in IN; des; vauto.
          { tmext_simpl; omega. }
          repeat (i; apply in_consD in IN; des; vauto).
          apply in_map_iff in IN; des; vauto.
        }
  - inst_pairs; i; rr; s; unfold mid_cemp; eauto.
Grab Existential Variables.
  exact 0. exact tt.
Qed.





