(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

open Big_int
open Lexer
open Parser
open Printf

open Basics
open Basis
open Datatypes
open Lang_common
open Lang_tgt
open Lang_tgt_interp
open Lang_tgt_syntax
open Pilsner


(** Showing machine syntax. *)

let show_binop = function
  | Plus -> "Plus"
  | Minus -> "Minus"
  | Mult -> "Mult"
  | Equal -> "Equal"
;;

let show_reg = function
  | Coq_r_sp -> "sp"
  | Coq_r_env -> "env"
  | Coq_r_arg -> "arg"
  | Coq_r_clo -> "clo"
  | Coq_r_ret -> "ret"
  | Coq_r_aux -> "aux"
  | Coq_r_i -> "i"
  | Coq_r_j -> "j"
;;

let show_offset = function
  | Coq_o_pos n -> "+" ^ string_of_big_int n
  | Coq_o_neg n -> "-" ^ string_of_big_int n
;;

let show_operand = function
  | Coq_o_imm n -> string_of_big_int n
  | Coq_o_reg r -> show_reg r
  | Coq_o_sind (r, o) -> "{" ^ show_reg r ^ show_offset o ^ "}"
  | Coq_o_hind (r, o) -> "[" ^ show_reg r ^ show_offset o ^ "]"
;;

let show_instr = function
  | Coq_i_jmp o ->
    "jmp " ^ show_operand o
  | Coq_i_jnz (r, o) ->
    "jnz " ^ show_reg r ^ " " ^ show_operand o
  | Coq_i_ld (r, o) ->
    "ld " ^ show_reg r ^ " " ^ show_operand o
  | Coq_i_sto (o, r) ->
    "sto " ^ show_operand o ^ " " ^ show_reg r
  | Coq_i_lpc r ->
    "lpc " ^ show_reg r
  | Coq_i_opr (op, r, o1, o2) ->
    "opr " ^ show_binop op ^ " " ^ show_reg r ^ " "
           ^ show_operand o1 ^ " " ^ show_operand o2
  | Coq_i_inp r ->
    "inp " ^ show_reg r
  | Coq_i_out r ->
    "out " ^ show_reg r
  | Coq_i_alc (r1, r2) ->
    "alc " ^ show_reg r1 ^ " " ^ show_reg r2
;;


(** Misc. *)

let big =
  big_int_of_int
;;
  

(** Modules: Loading, Storing, and Printing. *)

let print_module m =
  let print_function l v w =
    printf "Function \"%s\" has value %s (code starts at %s).\n"
           (string_of_big_int l)
           (string_of_big_int v)
           (string_of_big_int w) in
  let (h, exports) = cloadh m zero_big_int [] in
  let print_cell (a, w) =
    (*
    try let p = List.find (fun p -> eq_big_int (snd p) a) exports in
        (* This address is an exported function value. *)
        print_function (fst p) a w
    with Not_found ->
    *)
    let s = match decode_instr w with
              | Some x -> show_instr x
              | None -> "INVALID" in
    printf "%3s: %10s  (%s)\n" (string_of_big_int a) (string_of_big_int w) s in
  printf "Loaded %n functions:\n\n" (List.length exports);
  List.iter (fun (l, v) -> let p = List.find (fun p -> eq_big_int (fst p) v) h in
                           print_function l v (snd p)) exports;
  print_newline ();
  printf "Dump of initial memory:\n\n";
  printf "loc: value      (instruction)\n";
  printf "-------------------------------------------------------------------\n";
  List.iter print_cell h;
  print_newline ()
;;

let save_module m filename : unit =
  let cout = open_out_bin filename in
  Marshal.to_channel cout m [];
  close_out cout
;;

let load_module filename : Lang_tgt.coq_module =
  let cin = open_in_bin filename in
  let m = (Marshal.from_channel cin : Lang_tgt.coq_module) in
  close_in cin;
  m
;;
