(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis lang_common lang_src_lsi lang_tgt.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


Notation index x ctx := (withd 0 (list_index x ctx)).

Definition push n := i_opr Plus r_sp (o_reg r_sp) (o_imm n).
Definition pop n := i_opr Minus r_sp (o_reg r_sp) (o_imm n).

Definition stack_top := o_sind r_sp o_mone.
Definition stack_over_top := o_sind r_sp o_zero.


Definition getLabel (l: label) n imports :=
  let i := index l imports in
  i_lpc r_aux ::
  i_opr Minus r_aux (o_reg r_aux) (o_imm n) ::
  i_ld r_aux (o_hind r_aux (o_pos i)) ::
  push 1 ::
  i_sto stack_top r_aux ::
  nil.


Definition getVariable (x: termvar) vars :=
  let index := index x vars in
  i_sto (o_reg r_clo) r_env ::
  repeat (i_ld r_clo (o_hind r_clo o_one)) index ++
  push 1 :: i_ld r_clo (o_hind r_clo o_zero) :: i_sto stack_top r_clo :: nil.


Fixpoint tr_exp e n imports vars :=
  
  match e with
    | Evar x =>
        getVariable x vars

    | Elbl l =>
        getLabel l n imports

    | Eunit =>
        push 1 :: nil 

    | Enat n => 
        push 1 ::
        i_ld r_aux (o_imm n) ::
        i_sto stack_top r_aux :: nil

    | Epair e1 e2 =>
        let c1 := tr_exp e1 n imports vars in
        let c2 := tr_exp e2 (length c1 + n) imports vars in
        c1 ++ c2 ++
        pop 1 ::
        i_ld r_clo (o_imm 2) ::
        i_alc r_clo r_clo ::
        i_ld r_aux stack_top ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_ld r_aux stack_over_top ::
        i_sto (o_hind r_clo o_one) r_aux ::
        i_sto stack_top r_clo :: nil

    | Efst e =>
        tr_exp e n imports vars ++ 
        i_ld r_clo stack_top ::
        i_ld r_aux (o_hind r_clo o_zero) ::
        i_sto stack_top r_aux :: nil

    | Esnd e =>
        tr_exp e n imports vars ++ 
        i_ld r_clo stack_top ::
        i_ld r_aux (o_hind r_clo o_one) ::
        i_sto stack_top r_aux :: nil

    | Einl e =>
        tr_exp e n imports vars ++
        i_ld r_clo (o_imm 2) ::
        i_alc r_clo r_clo ::
        i_ld r_aux (o_imm 0) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_ld r_aux stack_top ::
        i_sto (o_hind r_clo o_one) r_aux ::
        i_sto stack_top r_clo :: nil

    | Einr e =>
        tr_exp e n imports vars ++
        i_ld r_clo (o_imm 2) ::
        i_alc r_clo r_clo ::
        i_ld r_aux (o_imm 1) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_ld r_aux stack_top ::
        i_sto (o_hind r_clo o_one) r_aux ::
        i_sto stack_top r_clo :: nil

    | Ebop op e1 e2 =>
        let c1 := tr_exp e1 n imports vars in
        let c2 := tr_exp e2 (length c1 + n) imports vars in
        c1 ++ c2 ++
        pop 1 ::
        i_opr op r_aux stack_top stack_over_top ::
        i_sto stack_top r_aux :: nil

    | Einput =>
        i_inp r_clo ::
        push 1 ::
        i_sto stack_top r_clo :: nil

    | Eoutput e =>
        tr_exp e n imports vars ++
        i_ld r_clo stack_top ::
        i_out r_clo :: nil

    | Eroll e | Eunroll e | Epack e =>
        tr_exp e n imports vars

    | Eifnz e ea eb =>
        let c := tr_exp e n imports vars in
        let cb := tr_exp eb (5 + length c + n) imports vars in
        let ca := tr_exp ea (8 + length cb + length c + n) imports vars in
        c ++ 
        pop 1 ::
        i_lpc r_clo ::
        i_opr Plus r_clo (o_reg r_clo) (o_imm (length cb + 7)) ::
        i_ld r_arg stack_over_top ::
        i_jnz r_arg (o_reg r_clo) ::
        cb ++
        i_lpc r_clo ::
        i_opr Plus r_clo (o_reg r_clo) (o_imm (length ca + 3)) ::
        i_jmp (o_reg r_clo) :: 
        ca

    | Ecase e x ea eb =>
        let c := tr_exp e n imports vars in
        let ca := tr_exp ea (12 + length c + n) imports (x :: vars) in
        let cb := tr_exp eb (15 + length ca + length c + n) imports (x :: vars) in
        c ++ 
        pop 1 :: 
        i_ld r_clo (o_reg r_env) ::
        i_ld r_env (o_imm 2) ::
        i_alc r_env r_env ::
        i_sto (o_hind r_env o_one) r_clo ::
        i_ld r_arg stack_over_top ::
        i_ld r_aux (o_hind r_arg o_one) ::
        i_sto (o_hind r_env o_zero) r_aux ::
        i_ld r_aux (o_hind r_arg o_zero) ::
        i_lpc r_clo ::
        i_opr Plus r_clo (o_reg r_clo) (o_imm (length ca + 6)) ::
        i_jnz r_aux (o_reg r_clo) ::
        ca ++
        i_lpc r_clo ::
        i_opr Plus r_clo (o_reg r_clo) (o_imm (length cb + 3)) ::
        i_jmp (o_reg r_clo) ::
        cb ++ 
        i_ld r_env (o_hind r_env o_one) :: nil

    | Efix f x e =>
        let c := tr_exp e (22 + n) imports (x :: f :: vars) in
        i_ld r_clo (o_imm 2) ::
        i_alc r_clo r_clo ::
        i_lpc r_arg ::
        i_opr Plus r_arg (o_reg r_arg) (o_imm 8) ::
        i_sto (o_hind r_clo o_zero) r_arg ::
        i_sto (o_hind r_clo o_one) r_env ::
        push 1 ::
        i_sto stack_top r_clo ::
        i_opr Plus r_arg (o_reg r_arg) (o_imm (length c + 16)) ::
        i_jmp (o_reg r_arg) ::
        
        push 1 ::
        i_sto stack_top r_env ::
        i_ld r_env (o_imm 2) ::
        i_alc r_env r_env ::
        i_sto (o_hind r_env o_zero) r_clo :: 
        i_ld r_aux (o_hind r_clo o_one) :: 
        i_sto (o_hind r_env o_one) r_aux :: 
        i_ld r_clo (o_imm 2) ::
        i_alc r_clo r_clo ::
        i_sto (o_hind r_clo o_zero) r_arg :: 
        i_sto (o_hind r_clo o_one) r_env ::
        i_ld r_env (o_reg r_clo) ::
        c ++
        i_ld r_arg stack_top :: 
        pop 2 ::
        i_ld r_env stack_over_top ::
        i_jmp (o_reg r_ret) :: nil

    | Eapp e1 e2 => 
        let c1 := tr_exp e1 n imports vars in
        let c2 := tr_exp e2 (length c1 + n) imports vars in
        c1 ++ c2 ++
        pop 1 :: 
        i_ld r_clo stack_top ::
        i_sto stack_top r_ret ::
        i_ld r_arg stack_over_top ::
        i_lpc r_ret ::
        i_opr Plus r_ret (o_reg r_ret) (o_imm 3) ::
        i_jmp (o_hind r_clo o_zero) ::
        
        i_ld r_ret stack_top ::
        i_sto stack_top r_arg :: nil

    | Egen e =>
        let c := tr_exp e (13 + n) imports vars in
        i_ld r_clo (o_imm 2) ::
        i_alc r_clo r_clo ::
        i_lpc r_arg ::
        i_opr Plus r_arg (o_reg r_arg) (o_imm 8) ::
        i_sto (o_hind r_clo o_zero) r_arg ::
        i_sto (o_hind r_clo o_one) r_env ::
        push 1 ::
        i_sto stack_top r_clo ::
        i_opr Plus r_arg (o_reg r_arg) (o_imm (length c + 7)) ::
        i_jmp (o_reg r_arg) ::
        
        push 1 :: 
        i_sto stack_top r_env ::
        i_ld r_env (o_hind r_clo o_one) ::
        c ++
        i_ld r_arg stack_top :: 
        pop 2 ::
        i_ld r_env stack_over_top ::
        i_jmp (o_reg r_ret) :: nil

    | Einst e => 
        let c := tr_exp e n imports vars in
        c ++
        i_ld r_clo stack_top ::
        i_sto stack_top r_ret ::
        i_lpc r_ret ::
        i_opr Plus r_ret (o_reg r_ret) (o_imm 3) ::
        i_jmp (o_hind r_clo o_zero) ::
        
        i_ld r_ret stack_top ::
        i_sto stack_top r_arg :: nil

    | Eunpack x e1 e2 => 
        let c1 := tr_exp e1 n imports vars in
        let c2 := tr_exp e2 (7 + length c1 + n) imports (x :: vars) in
        c1 ++ 
        pop 1 :: 
        i_ld r_clo (o_reg r_env) ::
        i_ld r_env (o_imm 2) ::
        i_alc r_env r_env ::
        i_ld r_aux stack_over_top ::
        i_sto (o_hind r_env o_zero) r_aux ::
        i_sto (o_hind r_env o_one) r_clo ::
        c2 ++
        i_ld r_env (o_hind r_env o_one) :: nil

    | Enew e => 
        tr_exp e n imports vars ++
        i_ld r_clo (o_imm 1) :: 
        i_alc r_clo r_clo ::
        i_ld r_aux stack_top ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_sto stack_top r_clo :: nil

    | Eread e =>
        tr_exp e n imports vars ++
        i_ld r_arg stack_top ::
        i_ld r_aux (o_hind r_arg o_zero) ::
        i_sto stack_top r_aux :: nil

    | Ewrite e1 e2 =>
        let c1 := tr_exp e1 n imports vars in
        let c2 := tr_exp e2 (length c1 + n) imports vars in
        c1 ++ c2 ++
        pop 1 ::
        i_ld r_clo stack_top ::
        i_ld r_aux stack_over_top ::
        i_sto (o_hind r_clo o_zero) r_aux :: nil
    
    | Erefeq e1 e2 =>
        let c1 := tr_exp e1 n imports vars in
        let c2 := tr_exp e2 (length c1 + n) imports vars in
        c1 ++ c2 ++
        pop 1 ::
        i_opr Equal r_aux stack_top stack_over_top ::
        i_sto stack_top r_aux :: nil
        
    | Eloc _ => nil 
    end.

Fixpoint tr_toplvl v n imports :=
  match v with
    | Efix f x e =>
      push 1 ::
      i_sto stack_top r_env ::
      i_ld r_aux (o_imm 2) ::
      i_alc r_aux r_aux ::      
      i_sto (o_hind r_aux o_zero) r_clo ::
      i_ld r_env (o_imm 2) ::
      i_alc r_env r_env ::      
      i_sto (o_hind r_env o_zero) r_arg ::
      i_sto (o_hind r_env o_one) r_aux ::
      tr_exp e (9 + n) imports (x :: f :: nil) ++
      i_ld r_arg stack_top :: 
      pop 2 ::
      i_ld r_env stack_over_top ::
      i_jmp (o_reg r_ret) :: nil
    | Egen e =>
      push 1 ::
      i_sto stack_top r_env ::
      tr_exp e (2 + n) imports nil ++
      i_ld r_arg stack_top :: 
      pop 2 ::
      i_ld r_env stack_over_top ::
      i_jmp (o_reg r_ret) :: nil
    | _ => nil
  end.

Fixpoint tr_mod_aux m n F (M: lang_src.module) : group :=
  match M with
    | nil =>
      {| g_cptrs := nil; g_data := nil |}
    | (f, v) :: M' =>
      let code := tr_toplvl v (m + n) F in
      let g := tr_mod_aux m (n + length code) (F ++ [f]) M' in
      {| g_cptrs := (f, n) :: g_cptrs g;
         g_data := map enc code ++ g_data g
      |}
  end.

Definition tr_mod F (M: lang_src.module) : module :=
  let ni := length F in
  let ne := length M in
  [tr_mod_aux (ni + ne + ne) 0 F M].
