(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Max Relations.
Require Import vgtac basis myclassical ssr.
Require Import lang_common lang_spec lang_mid lang_tgt lang_src_lsi.
Require Import model_common model_untyped model model_properties_untyped model_properties.
Require Import gw_common gw_untyped gw vcomp_aux.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Definition bullet (sga: tm_gw.(state)) (sgb: ms_gw.(state)) : ts_gw.(state) :=
  let '(sta, sma, sra) := sga in
  let '(smb, ssb, srb) := sgb in
  (sta, ssb, refst_compP sra srb).


Definition minusa (sga: tm_gw.(state)) (sg: ts_gw.(state)) : tm_gw.(state) :=
  let '(sta, sma, sra) := sga in
  let '(_, _, sr) := sg in
  (sta, sma, refst_minusa sra sr).

Definition minusb (sgb: ms_gw.(state)) (sg: ts_gw.(state)) : ms_gw.(state) :=
  let '(stb, smb, srb) := sgb in
  let '(_, _, sr) := sg in
  (stb, smb, refst_minusb srb sr).


Lemma minusa_ext: forall sga sgb,
  ext _ sga (minusa sga (bullet sga sgb)).
Proof.
  i; s; splits; auto
  ; destruct sga, sgb; destruct s, s1; destruct s1; s; i; des; auto.
Qed.

Lemma minusb_ext: forall sgb sga,
  ext _ sgb (minusb sgb (bullet sga sgb)).
Proof.
  i; s; splits; auto
  ; destruct sgb, sga; destruct s, s1; destruct s1; s; i; des; auto.
Qed.


Lemma bullet_ext: forall
    sga sga' (EXTa: ext tm_gw sga' sga)
    sgb sgb' (EXTb: ext ms_gw sgb' sgb),
  ext ts_gw (bullet sga' sgb') (bullet sga sgb).
Proof.
  i; destruct sga as [[[]]], sga' as [[[]]], sgb as [[[]]], sgb' as [[[]]].
  rrdes EXTa EXTa EXTb EXTb; clean; unfold bullet.
  split; [split|]; eauto using refst_compP_ext.
Qed.

Lemma bullet_extpub: forall
    sga sga' (EXTa: extpub tm_gw sga' sga)
    sgb sgb' (EXTb: extpub ms_gw sgb' sgb),
  extpub ts_gw (bullet sga' sgb') (bullet sga sgb).
Proof.
  i; destruct sga as [[[]]], sga' as [[[]]], sgb as [[[]]], sgb' as [[[]]].
  rrdes EXTa EXTa EXTb EXTb; clean; unfold bullet.
  split; [split|]; eauto using refst_compP_extpub.
Qed.

Lemma bullet_tgt: forall sga sgb,
  fst (fst sga) = fst (fst (bullet sga sgb)).
Proof.
  i; destruct sga as [[[]]], sgb as [[]]; auto. 
Qed.

Lemma bullet_src: forall sga sgb,
  snd (fst sgb) = snd (fst (bullet sga sgb)).
Proof.
  i; destruct sga as [[[]]], sgb as [[]]; auto.
Qed.



Lemma bullet_ref_comp: forall (sga: state tm_gw) (sgb: state ms_gw) ty lt lm ls
    (REGa: snd sga lt lm)
    (REGb: snd sgb ty lm ls),
  snd (bullet sga sgb) ty lt ls.
Proof.
  unfold bullet; i; destruct sga as [[[]]], sgb as [[[]]].
  s; unfold refst_comp; eauto.
Qed.

Lemma bullet_ref_decomp: forall sga sgb ty lt ls
    (REG: snd (bullet sga sgb) ty lt ls),
  exists lm,
    <<REGa: snd sga lt lm>> /\
    <<REGb: snd sgb ty lm ls>>.
Proof.
  i; destruct sga as [[[]]], sgb as [[[]]]; rrdes REG; eauto.
Qed.



Section vcomp_tms.

  Variable mt: lang_spec.module tgt_lsi.
  Variable mm: lang_spec.module mid_lsi.
  Variable ms: lang_spec.module src_lsi.
  Variable ienv oenv: list (label * flextyp).
  Variable names: typename -> Prop.
  Variable lwa: A.world_loca tm_gw.
  Variable lwb: B.world_loca ms_gw.


  Definition lw_sts :=
    tsprod (tsprod GA.ref_sts (A.lwsts lwa)) (tsprod GB.ref_sts (B.lwsts lwb)).

  

  Definition sdecomp (s: state (tsprod ts_gw lw_sts))
    : state (A.wlift lwa) * state (B.wlift lwb) :=
    let '(sg, ((sra, sla), (srb, slb))) := s in
    let '(sct, scs, sr) := sg in
    ((sct, tt, sra, sla), (tt, scs, srb, slb)).

  Definition scomp (sa: state (A.wlift lwa)) (sb: state (B.wlift lwb))
    : state (tsprod ts_gw lw_sts) :=
    let '(sga, sla) := sa in
    let '(sgb, slb) := sb in
    (bullet sga sgb, ((snd sga, sla), (snd sgb, slb))).

  Lemma scomp_inj: forall sa sb sa' sb'
      (EQ: scomp sa sb = scomp sa' sb'),
    sa = sa' /\ sb = sb'.
  Proof.
    i; destruct sa as [[[]]], sb as [[[]]], sa' as [[[]]], sb' as [[[]]].
    destruct s0, s3, s8, s11; s in EQ; clarify. 
  Qed.


  Definition mybad (ty: flextyp) (vt: tgt_val) (vs: src_val) :=
    match flext_of ty with
      | Tarrow _ _ => 
          IVclo nil (IEfix 0 0 0 (IEnat (enum (ty, vt, vs))))
      | Tforall _ => 
          IVclo nil (IEgen 0 (IEnat (enum (ty, vt, vs))))
      | Tname _ => 
          IVclo nil (IEnat (enum (ty, vt, vs)))
      | _ => 
          IVunit
    end.

  Lemma mybad_inj: forall (ty ty': flextyp) vt vt' vs vs'
      (E: mybad ty vt vs = mybad ty' vt' vs'),
    ty = ty' /\ vt = vt' /\ vs = vs'.
  Proof.
    unfold mybad; i; destruct ty as [[]], ty' as [[]]; ss; clarify
    ; apply enum_inj in H0; inversion H0; subst
    ; apply (f_equal (fun x => (fst (fst x)))) in H0; auto.
  Qed.


  Lemma enc_valid_fun_a: forall sa tyx tyy vt vs,
    A.vqueryb (A.wlift lwa) sa valid_fun (mybad (FTarrow tyx tyy) vt vs).
  Proof.
    econs.
  Qed.

  Lemma enc_valid_fun_b: forall sb tyx tyy vt vs,
    B.vquerya (B.wlift lwb) sb valid_fun (mybad (FTarrow tyx tyy) vt vs).
  Proof.
    econs.
  Qed.

  Lemma enc_valid_gen_a: forall sa ty vt vs,
    A.vqueryb (A.wlift lwa) sa valid_gen (mybad (FTforall ty) vt vs).
  Proof.
    econs.
  Qed.

  Lemma enc_valid_gen_b: forall sb ty vt vs,
    B.vquerya (B.wlift lwb) sb valid_gen (mybad (FTforall ty) vt vs).
  Proof.
    econs.
  Qed.


  Inductive gka_enc (g: state (tsprod ts_gw lw_sts) -> @vrelf tgt_lsi src_lsi)
                    (sb: state (B.wlift lwb)) (sa: state (A.wlift lwa)) : @A.vrelf tgt_lsi mid_lsi :=
  | gka_enc_ sb' sa' ty vt vm vs
      (R: g (scomp sa' sb') ty vt vs)
      (EXTb: ext _ sb sb')
      (EXTa: ext _ sa sa')
      (N: forall tn (IN: lwb.(lwname).(supp) tn), ty <> FTname tn)
      (Evm: vm = mybad  ty vt vs)
    : gka_enc g sb sa vt vm.
  Hint Constructors gka_enc.
  
  


  Inductive gkb_enc (g: state (tsprod ts_gw lw_sts) -> @vrelf tgt_lsi src_lsi)
                    (sa: state (A.wlift lwa)) (sb: state (B.wlift lwb)) : @B.vrelf mid_lsi src_lsi :=
  | gkb_enc_ sa' sb' ty vt vm vs
      (R: g (scomp sa' sb') ty vt vs)
      (EXTa: ext _ sa sa')
      (EXTb: ext _ sb sb')
      (N: forall tn (IN: lwb.(lwname).(supp) tn), ty <> FTname tn)
      (Evm: vm = mybad  ty vt vs)
    : gkb_enc g sa sb ty vm vs.
  Hint Constructors gkb_enc.

  Definition gka (g: state (tsprod ts_gw lw_sts) -> @vrelf tgt_lsi src_lsi) (sb: state (B.wlift lwb))
    : state (A.wlift lwa) -> @A.vrelf tgt_lsi mid_lsi :=
    PA.gkmin _ (gka_enc g sb).

  Definition gkb (g: state (tsprod ts_gw lw_sts) -> @vrelf tgt_lsi src_lsi) (sa: state (A.wlift lwa))
    : state (B.wlift lwb) -> @B.vrelf mid_lsi src_lsi :=
    PB.gkmin _ (gkb_enc g sa).




  Lemma sdecomp_ext: forall s s'
      (EXT: ext (tsprod ts_gw lw_sts) s' s),
    <<EXTa: ext (A.wlift lwa) (fst (sdecomp s')) (fst (sdecomp s))>> /\
    <<EXTb: ext (B.wlift lwb) (snd (sdecomp s')) (snd (sdecomp s))>>.
  Proof.
    clear; i; rrdes EXT EXT EXT; rrdes EXT0 EXT0 EXT3.
    destruct s as [[[]] [[] []]], s' as [[[]] [[] []]].
    unfold sdecomp, scomp in *; do 4 (try split); s; auto.
  Qed.

  Lemma sdecomp_extpub: forall s s'
      (EXT: extpub (tsprod ts_gw lw_sts) s' s),
    <<EXTa: extpub (A.wlift lwa) (fst (sdecomp s')) (fst (sdecomp s))>> /\
    <<EXTb: extpub (B.wlift lwb) (snd (sdecomp s')) (snd (sdecomp s))>>.
  Proof.
    clear; i; rrdes EXT EXT EXT; rrdes EXT0 EXT0 EXT3.
    destruct s as [[[]] [[] []]], s' as [[[]] [[] []]].
    unfold sdecomp, scomp in *; do 4 (try split); s; auto.
  Qed.

  Lemma scomp_ext: forall
      sa sa' (EXTa: ext (A.wlift lwa) sa' sa)
      sb sb' (EXTb: ext (B.wlift lwb) sb' sb),
    ext (tsprod ts_gw lw_sts) (scomp sa' sb') (scomp sa sb).
  Proof.
    clear; i; rrdes EXTa EXTb.
    hexploit bullet_ext; eauto; []; i.
    destruct sa as [[[]]], sb as [[[]]], sa' as [[[]]], sb' as [[[]]].
    split; [by auto|by rrdes EXTa EXTb; split; auto].
  Qed.

  Lemma scomp_extpub: forall
      sa sa' (EXTa: extpub (A.wlift lwa) sa' sa)
      sb sb' (EXTb: extpub (B.wlift lwb) sb' sb),
    extpub (tsprod ts_gw lw_sts) (scomp sa' sb') (scomp sa sb).
  Proof.
    clear; i; rrdes EXTa EXTb.
    hexploit bullet_extpub; eauto; []; i.
    destruct sa as [[[]]], sb as [[[]]], sa' as [[[]]], sb' as [[[]]].
    split; [by auto|by rrdes EXTa EXTb; split; auto].
  Qed.

  Lemma gka_enc_mon: forall
      g g' (LE: g <g= g')
      sb sb' (EXTb: ext _ sb' sb)
      sa sa' (EXTa: ext _ sa' sa),
    gka_enc g sb sa <Ur= gka_enc g' sb' sa'.
  Proof.
    clear; i; invc PR; eauto using ext_trans.
  Qed.

  Lemma gkb_enc_mon: forall
      g g' (LE: g <g= g')
      sa sa' (EXTa: ext _ sa' sa)
      sb sb' (EXTb: ext _ sb' sb),
    gkb_enc g sa sb <r= gkb_enc g' sa' sb'.
  Proof.
    clear; i; invc PR; eauto using ext_trans.
  Qed.

  Lemma gka_mon: forall g sb sa vt vm
      (IN: gka g sb sa vt vm)
      g' (LE: g <g= g')
      sb' (EXTb: ext _ sb' sb)
      sa' (EXTa: ext _ sa' sa),
    gka g' sb' sa' vt vm.
  
  
  
  
  
  Proof.
    clear; i; ginduction IN; i.
    - econs 1; eauto using gka_enc_mon.
    - econs 2.
      + eapply A.lsim_mon; eauto.
      + eapply PA.gkmin_mon; eauto using gka_enc_mon.
  Qed.

  Lemma gkb_mon: forall g sa sb ty vm vs
      (IN: gkb g sa sb ty vm vs)
      g' (LE: g <g= g')
      sb' (EXTb: ext _ sb' sb)
      sa' (EXTa: ext _ sa' sa),
    gkb g' sa' sb' ty vm vs.
  Proof.
    clear; i; ginduction IN; i.
    - econs 1; eauto using gkb_enc_mon.
    - econs 2; i.
      + eapply B.lsim_mon; eauto.
      + eapply PB.gkmin_mon; last first; eauto using gkb_enc_mon.
    - econs 3; eauto.
      eapply B.name_mon; eauto; i; eapply H; eauto.
  Qed.
  


  Inductive lw_namerel g (s: state (tsprod ts_gw lw_sts)) tn vt vs : Prop :=
  | lw_namerel_ (SUPP: lwb.(supp) tn) vm
             (Ra: A.vclos (A.wlift lwa) (fst (sdecomp s))
                          (gka g (snd (sdecomp s)) (fst (sdecomp s))) vt vm)
             (Rb: B.vclos (B.wlift lwb) (snd (sdecomp s))
                          (gkb g (fst (sdecomp s)) (snd (sdecomp s))) (Tname tn) vm vs)
  .                             

  Lemma lw_namerel_mon: forall
      g s tn va vb (NAM: lw_namerel g s tn va vb)
      g' (LE: g <g= g')
      s' (EXT: ext _ s' s),
    lw_namerel g' s' tn va vb.
  Proof.
    i; invc NAM; hexploit sdecomp_ext; eauto; []; i; des.
    econs; eauto.
    - eapply A.vclos_mon; eauto using gka_mon.
    - eapply B.vclos_mon; eauto using gkb_mon.
  Qed.

  Program Definition lw_name: method_name tgt_lsi src_lsi (tsprod ts_gw lw_sts) :=
    {| supp := lwb.(supp)
     ; name := lw_namerel
    |}.
  Next Obligation.
    invc WV.
  Qed.
  Next Obligation.
    clean; subst; eapply lw_namerel_mon; repeat (rr; splits; s; eauto).
  Qed.


  Inductive lw_confrel g (s: state (tsprod ts_gw lw_sts)) : @crel tgt_lsi src_lsi :=
  | lw_confrel_
      (sga: state tm_gw) (sla: state lwa)
      (sgb: state ms_gw) (slb: state lwb)
      (SC: s = scomp (sga, sla) (sgb, slb))

      (crt: conf tgt_lsi) (crma crmb: conf mid_lsi) (crs: conf src_lsi)
      (CRga: tm_refconfrel' (snd (minusa sga (fst s)))
                            (fun sga' => gka g (sgb, slb) (sga', sla)) sga crt crma)
      (CRgb: ms_refconfrel' (snd (minusb sgb (fst s)))
                            (fun sgb' => gkb g (sga, sla) (sgb', slb)) sgb crmb crs)

      (clt: conf tgt_lsi) (clma clmb: conf mid_lsi) (cls: conf src_lsi)
      (CRla: A.confrel lwa (gka g (sgb, slb)) (sga, sla) clt clma)
      (CRlb: B.confrel lwb (gkb g (sga, sla)) (sgb, slb) clmb cls)

      (E: crma # clma = crmb # clmb)

      (RL: forall em hm (CO: core em) (VLD: pmvalid hm)
               (DOM: forall lm vm (RD: pmread hm lm vm), exists lt, snd sga lt lm),
             exists mm, real mid_lsi (em # ((hm, Nothing) : conf mid_lsi) # clma) mm)
     

      ct cs (Ect: ct = crt # clt) (Ecs: cs = crs # cls)

    : lw_confrel g s ct cs.

  
  



  Program Definition lw: world_loca ts_gw :=
    {| lwwfo := awfo_lexprod lwb lwa
     ; lwsts := lw_sts
     ; lwname := lw_name
     ; lwconf := {| confrel := lw_confrel |}
    |}.
  Next Obligation.
    invc PR; econs; clean; eauto; des; esplits; eauto.
    - eapply tm_refconfrel'_mon; eauto.
      s; i; eapply gka_mon; eauto.
    - eapply ms_refconfrel'_mon; eauto.
      s; i; eapply gkb_mon; eauto.
    - eapply A.confrel_mon; eauto using gka_mon.
    - eapply B.confrel_mon; eauto using gkb_mon.
  Qed.
  Next Obligation.
    invc CR.
    apply A.lwconf_extra in CRla.
    apply B.lwconf_extra in CRlb.
    des; rrdes CRla CRla0 CRlb CRlb0; destruct crma, crmb, clt, clma, clmb.
    s in CRla; s in CRla1; clarify; noemp; clarify.
    rrdes CRga CRgb; clarify; s; noemp; split; rr; eauto.
  Qed.


  Lemma gka_gknow: forall sb g
      (GK: gknow (wlift lw) g),
    A.gknow (A.wlift lwa) (gka g sb).
  Proof.
    i; eapply PA.gkmin_gknow.
    i; eapply gka_enc_mon; last first; eauto.
  Qed.

  Lemma gkb_gknow: forall sa g
      (GK: gknow (wlift lw) g),
    B.gknow (B.wlift lwb) (gkb g sa).
  Proof.
    i; eapply PB.gkmin_gknow.
    - i; inv IN; s in FH.
      apply GK in R; auto; eapply N; eauto.
    - i; eapply gkb_enc_mon; last first; eauto.
  Qed.



  Lemma scomp_vquerya: forall sa sb q vt,
    A.vquerya (A.wlift lwa) sa q vt
    <->
    vquerya (wlift lw) (scomp sa sb) q vt.
  Proof.
    i; destruct sa, sb; s; rewrite -bullet_tgt; eauto.
  Qed.

  Lemma scomp_vqueryb: forall sa sb q vs,
    B.vqueryb (wlift lwb) sb q vs
    <->
    vqueryb (wlift lw) (scomp sa sb) q vs.
  Proof.
    i; s; eauto.
  Qed.                          

  Lemma scomp_cquerya: forall sa sb q et,
    A.cquerya (A.wlift lwa) sa q et
    <->
    cquerya (wlift lw) (scomp sa sb) q et.
  Proof.
    i; s; destruct sa, sb; rewrite -bullet_tgt; eauto.
  Qed.

  Lemma scomp_cqueryb: forall sa sb q es,
    B.cqueryb (wlift lwb) sb q es
    <->
    cqueryb (wlift lw) (scomp sa sb) q es.
  Proof.
    i; s; eauto.
  Qed.                          

  Lemma minusa_scomp: forall (sga: state tm_gw) sla (sgb: state ms_gw) slb lt lm
      (NIN: ~ exists (ty: typ) (ls: lang_mid.location), (snd sgb) ty lm ls)
      (REG: (snd sga) lt lm),
     snd (minusa sga (fst (scomp (sga, sla) (sgb, slb)))) lt lm.
  Proof.
    i; unfold scomp, minusa; s.
    destruct sga, sgb; destruct s, s1; clean; s; split; auto.
    i; rrdes FH; vauto.
    hexploit (GA.refst_bij FH REG); []; i; des; subst.
    specialize (X eq_refl); des; subst; eauto.
  Qed.

  Lemma minusb_scomp: forall (sga: state tm_gw) sla (sgb: state ms_gw) slb ty lm ls
      (NIN: ~ exists lt, (snd sga) lt lm)
      (REG: (snd sgb) ty lm ls),
     snd (minusb sgb (fst (scomp (sga, sla) (sgb, slb)))) ty lm ls.
  Proof.
    i; unfold scomp, minusa; s.
    destruct sga, sgb; destruct s, s1; clean; s; split; auto.
    i; rrdes FH; vauto.
    hexploit (GB.refst_bij FH0 REG); []; i; des; subst.
    specialize (X0 eq_refl); des; subst; eauto.
  Qed.

  Lemma rquery_comp: forall sa sb ty lt lm ls
      (FHa: A.rquery (GA.tm_gw : A.world_glob _ _) (fst sa) lt lm)
      (FHb: B.rquery (GB.ms_gw : B.world_glob _ _) (fst sb) ty lm ls),
    model.rquery ts_gw (fst (scomp sa sb)) ty lt ls.
  Proof.
    unfold scomp; i; rrdes FHa FHb; clarify; destruct sa, sb.
    rr; esplits; s; eauto using bullet_ref_comp.
  Qed.

  Lemma rquery_decomp: forall sa sb ty lt ls
      (FH: model.rquery ts_gw (fst (scomp sa sb)) ty lt ls),
    exists lm,
      <<FHa: A.rquery (GA.tm_gw : A.world_glob _ _) (fst sa) lt lm>> /\
      <<FHb: B.rquery (GB.ms_gw : B.world_glob _ _) (fst sb) ty lm ls>>.
  Proof.
    unfold scomp; i; subst; rrdes FH; destruct sa, sb.
    hexploit bullet_ref_decomp; eauto; i; des.
    esplits; s; r; esplits; eauto.
  Qed.


  Lemma vclos_thunk_decomp: forall sa sb g ty vt vs
      (VCL: vclos_thunk (wlift lw) (scomp sa sb) (g (scomp sa sb)) ty vt vs)
      r (GK: gknow_ (wlift lw) r g),
    <<VCLa: A.vclos_thunk (A.wlift lwa) sa (gka g sb sa) vt (mybad ty vt vs)>> /\
    <<VCLb: B.vclos_thunk (B.wlift lwb) sb (gkb g sa sb) ty (mybad ty vt vs) vs>>.
  Proof.
    clear; i; subst; inv VCL.
    - split; econs 1.
      + eapply scomp_vquerya; eauto.
      + apply enc_valid_fun_a.
      + apply PA.gkmin_fold; left; econs; eauto; vauto.
      + apply enc_valid_fun_b.
      + eauto.
      + apply PB.gkmin_fold; left; left; econs; eauto; vauto.
    - split; econs 2.
      + eapply scomp_vquerya; eauto.
      + apply enc_valid_gen_a.
      + apply PA.gkmin_fold; left; econs; eauto; vauto.
      + apply enc_valid_gen_b.
      + eauto.
      + apply PB.gkmin_fold; left; left; econs; eauto; vauto.
  Qed.


  Lemma vclos_decomp: forall g sa sb ty vt vs
      (RV: vclos (wlift lw) (scomp sa sb) (g (scomp sa sb)) ty vt vs)
      r (GK: gknow_ (wlift lw) r g),
    exists vm,
      <<RVa: A.vclos (A.wlift lwa) sa (gka g sb sa) vt vm>> /\
      <<RVb: B.vclos (B.wlift lwb) sb (gkb g sa sb) ty vm vs>>.
  Proof.
    i; apply (vclos_induction RV); i; invc IH; des.
    - esplits; econs
      ; [eapply A.vclgen_unit|eapply B.vclgen_unit]; s; eauto.
    - esplits; econs
      ; [eapply A.vclgen_nat|eapply B.vclgen_nat]; last first; s; eauto.
    - esplits; econs
      ; [eapply (A.vclgen_pair _ _ _ _ vax vay va)
        |eapply (B.vclgen_pair _ _ _ _ _ _ vm0 vm)]; try (by s; eauto).
      eapply scomp_vquerya; eauto. 
    - esplits; econs
      ; [eapply A.vclgen_inl|eapply B.vclgen_inl]; try (by s; eauto).
      eapply scomp_vquerya; eauto. 
    - esplits; econs
      ; [eapply A.vclgen_inr|eapply B.vclgen_inr]; try (by s; eauto).
      eapply scomp_vquerya; eauto. 
    - esplits; econs
      ; [eapply A.vclgen_roll|eapply B.vclgen_roll]; last first; try (by s; eauto).
    - esplits; econs
      ; [eapply A.vclgen_pack|eapply B.vclgen_pack]; last first; try (by s; eauto).
    - hexploit rquery_decomp; ss; eauto; []; i; des.
      esplits; econs
      ; [eapply A.vclgen_ref|eapply B.vclgen_ref]; try (by s; eauto).
    - destruct (LEM (supp lwb tn)).
      + destruct sa as [[[]]], sb as [[[]]]; destruct s1, s4.
        apply (gk_name GK) in VR; eauto; []; inv VR; eauto.
      + esplits; econs
        ; [eapply A.vclgen_name|eapply B.vclgen_name]; try (by s; eauto).
        * apply PA.gkmin_fold; left.
          econs; eauto; first by (ii; vauto).
          unfold mybad; s; eauto.
        * apply PB.gkmin_fold; left; left.
          econs; eauto; ii; vauto.
    - hexploit vclos_thunk_decomp; eauto; []; i; des.
      esplits; econs
      ; [eapply A.vclgen_thunk|eapply B.vclgen_thunk]; last first; try (by s; eauto).
  Qed.


  Lemma confrel_decomp: forall sa sb r g ct cs
      (GK: gknow_ (wlift lw) r g)
      (CR: confrel (wlift lw) g (scomp sa sb) ct cs),
    exists cm,
      <<CRa: A.confrel (A.wlift lwa) (gka g sb) sa ct cm>> /\
      <<CRb: B.confrel (B.wlift lwb) (gkb g sa) sb cm cs>> /\
      <<RL: forall em (CO: core em), exists mm, real _ (em # cm) mm>>.
  Proof.
    clear; i; rrdes CR WHX WHX0; subst.
    rename hay into clt; rename hby into cls.
    rename hay0 into crt; rename hby0 into crs.
    rename hax0 into cct; rename hbx0 into ccs.
    rewrite <-!cplus_assoc; invc WHY.
    apply scomp_inj in SC; des; subst.

    assert (MED:
      forall lm, exists vmo,
        (forall lt ty ls (RRa: snd sga lt lm) (RRb: snd sgb ty lm ls),
         exists vt vm vs,
           vmo = Some vm /\
           pmread (chp crt) lt vt /\
           pmread (fst (fst crs)) ls vs /\
           A.vclos (A.wlift lwa) (sga, sla) (gka g (sgb, slb) (sga, sla)) vt vm /\
           B.vclos (B.wlift lwb) (sgb, slb) (gkb g (sga, sla) (sgb, slb)) ty vm vs)
        /\
        (forall vm (E: vmo = Some vm), exists lt ty ls,
           <<RRa: snd sga lt lm>> /\ <<RRb: snd sgb ty lm ls>>)).
    { rrdes WHY0; rename ha' into hrt; rename hb' into hrs; i.
      destruct (LEM (exists lt ty ls, snd sga lt lm /\ snd sgb ty lm ls)).
      - des; hexploit RELV; [by eapply bullet_ref_comp; eauto|]; i; des.
        
        hexploit (@vclos_decomp g (sga, sla) (sgb, slb))
        ; first (apply vclos_lift); eauto; []; i; des.
        esplits; i; eauto.
        hexploit (GA.refst_bij e RRa); []; i; des.
        specialize (X0 eq_refl); des; subst.
        hexploit (GB.refst_bij e0 RRb); []; i; des.
        specialize (X0 eq_refl); des; subst.
        esplits; eauto.
      - exists None; split; i; exfalso.
        + apply f; eauto.
        + vauto.
    }

    set (hrmf := fun lm => proj1_sig (cid _ (MED lm))).
    set (crm := (Some hrmf, Nothing) : conf mid_lsi).

    assert (tm_refconfrel (fun _ : state tm_gwsts => gka g (sgb, slb) (sga, sla))
                          sga (crt0 # crt) (crma # crm)).
    { rrdes WHY0 CRga; subst crt0 crma.

      assert (VLDt: pmvalid (ha'0 ## ha')).
      { apply pmvalid_plus_read; auto; []; i.
        apply DOMa0 in RDa; apply DOMa in RDa0; []; des.
        destruct sga as [[[]]], sgb as [[[]]].
        rrdes RDa0 RDa; eapply RDa2; rr; eauto.
      }

      assert (VLDm: pmvalid (hb'0 ## Some hrmf)).
      { apply pmvalid_plus_read; vauto; []; i.
        apply DOMb0 in RDa; []; des.
        apply pmread_some in RDa0.
        unfold hrmf in RDa0; destruct (cid _ (MED x)); s in RDa0; des.
        apply a0 in RDa0; des.
        destruct sga as [[[]]], sgb as [[[]]].
        rrdes RDa; eapply RDa0; rr; eauto.
      }

      s; noemp; rr; esplits; eauto.
      - rewrite Ea; noemp; eauto.
      - i; apply pmread_plusD in ALC; des.
        + hexploit DOMa0; eauto; []; i; des.
          s in X; apply minusa_ext in X; eauto.
        + hexploit DOMa; eauto; []; i; des.
          s in X; apply bullet_ref_decomp in X; des; eauto.
      - i; apply pmread_plusD in ALC; des.
        + hexploit DOMb0; eauto; []; i; des.
          s in X; apply minusa_ext in X; eauto.
        + apply pmread_some in ALC.
          unfold hrmf in ALC.
          destruct (cid _ (MED lb)); s in ALC; des.
          apply a0 in ALC; des; eauto.
      - i; rename la into lt, lb into lm.
        destruct (LEM (exists ty ls, snd sgb ty lm ls)).
        + assert (RD: hrmf lm = hrmf lm) by auto; unfold hrmf at 2 in RD.
          destruct (cid _ (MED lm)) as (vm & P); s in RD; des.
          hexploit P; eauto; []; i; des; subst vm; clear P P0.
          esplits.
          * rewrite Ea in X0; s in X0.
            eapply pmread_plus; eauto; (rhs rewrite pmplus_comm); eauto.
          * eapply pmread_some in X.
            eapply pmread_plus; eauto; (rhs rewrite pmplus_comm); eauto.
          * eapply PA.vclos_unlift in X2; eauto.
        + hexploit RELV0; eauto using minusa_scomp; []; i; des.
          esplits; eauto using pmread_plus, A.vclos_mon.
    }

    assert (ms_refconfrel (fun _ : state ms_gwsts => gkb g (sga, sla) (sgb, slb))
                          sgb (crmb # crm) (crs0 # crs)).
    { rrdes WHY0 CRgb.
      subst crs0 crmb; rewrite ->Ea, Eb in *.

      assert (VLDm: pmvalid (ha'0 ## Some hrmf)).
      { apply pmvalid_plus_read; vauto; []; i.
        apply DOMa0 in RDa; []; des.
        apply pmread_some in RDa0.
        unfold hrmf in RDa0; destruct (cid _ (MED x)); s in RDa0; des.
        apply a0 in RDa0; des.
        destruct sga as [[[]]], sgb as [[[]]].
        rrdes RDa; eapply RDa0; rr; eauto.
      }

      assert (VLDs: pmvalid (hb'0 ## hb')).
      { apply pmvalid_plus_read; auto; []; i.
        apply DOMb0 in RDa; apply DOMb in RDa0; []; des.
        destruct sga as [[[]]], sgb as [[[]]].
        rrdes RDa0 RDa.
        hexploit (GB.refst_bij RDa1 RDa); []; i; des.
        specialize (X0 eq_refl); des; subst ty lm.
        eapply RDa2; rr; eauto.
      }

      s; noemp; rr; esplits; eauto.
      - i; apply pmread_plusD in ALC; des.
        + hexploit DOMa0; eauto; []; i; des.
          s in X; apply minusb_ext in X; eauto.
        + apply pmread_some in ALC.
          unfold hrmf in ALC.
          destruct (cid _ (MED la)); s in ALC; des.
          apply a0 in ALC; des; eauto.
      - i; apply pmread_plusD in ALC; des.
        + hexploit DOMb0; eauto; []; i; des.
          s in X; apply minusb_ext in X; eauto.
        + hexploit DOMb; eauto; []; i; des.
          s in X; apply bullet_ref_decomp in X; des; eauto.
      - i; rename la into lm, lb into ls.
        destruct (LEM (exists lt, snd sga lt lm)).
        + assert (RD: hrmf lm = hrmf lm) by auto; unfold hrmf at 2 in RD.
          destruct (cid _ (MED lm)) as (vm & P); s in RD; des.
          hexploit P; eauto; []; i; des; subst vm; clear P P0.
          esplits.
          * eapply pmread_some in X.
            eapply pmread_plus; eauto; (rhs rewrite pmplus_comm); eauto.
          * rewrite Eb in X1; s in X1.
            eapply pmread_plus; eauto; (rhs rewrite pmplus_comm); eauto.
          * eapply PB.vclos_unlift in X3; eauto.
        + hexploit RELV0; eauto using minusb_scomp; []; i; des.
          esplits; eauto using pmread_plus, B.vclos_mon.
    }

    esplits.
    - rr; esplits; [|by eauto|by rewrite 2!cplus_assoc; eauto|reflexivity].
      clean; rewrite -?cplus_assoc.
      rr; esplits; eauto.
      + split; [|by rr; eauto].
        s in CPa; erewrite bullet_tgt; eauto.
      + rewrite (cplus_comm crt); eauto.
    - noemp; rewrite -?cplus_assoc.
      rr; esplits; [|by eauto| |by rewrite 2!cplus_assoc; eauto]; last first.
      { rewrite cplus_swap E cplus_assoc; eauto. }
      clean; rewrite -?cplus_assoc.
      rr; esplits; eauto.
      + split; [by rr; eauto|].
        s in CPb; erewrite bullet_src; eauto.
      + noemp; rewrite cplus_comm; auto.
      + rewrite (cplus_comm crs); eauto.
    - noemp; i; rrdes H.
      destruct crma, crm; s in Eb; destruct e, e0; vauto.
      hexploit (RL em (h ## h0)); eauto.
  Qed.


  

  



  Variable ancht: anchor tgt_lsi.
  Variable subt: list (label * lang_spec.val tgt_lsi).
  Variable cigt: conf tgt_lsi.
  Variable cilt: conf tgt_lsi.
  Variable LMt: cload tgt_lsi mt ancht subt cigt cilt.
  Variable SUBt: map fst subt = map fst ienv.

  Variable anchs: anchor src_lsi.
  Variable subs: list (label * lang_spec.val src_lsi).
  Variable cigs: conf src_lsi.
  Variable cils: conf src_lsi.
  Variable LMs: cload src_lsi ms anchs subs cigs cils.
  Variable SUBs: map fst subs = map fst ienv.


  Definition anchm: anchor mid_lsi := tt.

  Definition subm: list (label * lang_spec.val mid_lsi) :=
    let dummy := IVunit in
    let g p :=
        let '(f, ty) := p in
        match lookup f subt, lookup f subs with
          | Some vt, Some vs => (f, mybad ty vt vs)
          | _, _ => (f, IVunit)
        end in
    map g ienv.

  Definition cigm: conf mid_lsi := cemp.
  Definition cilm: conf mid_lsi := cemp.

  Lemma LMm: cload mid_lsi mm anchm subm cigm cilm.
  Proof.
    rr; auto.
  Qed.

  Lemma SUBm: map fst subm = map fst ienv.
  Proof.
    clear; unfold subm; induction ienv; auto.
    destruct a; s. 
    destruct (@list_assoc _ tgt_val _ _ _)
    ; try destruct (@list_assoc _ src_val _ _ _); rewrite IHl; auto.
  Qed.


  Lemma subm_lookupD: forall f vm
      (LU: lookup f subm = Some vm),
    exists ty vt vs,
      lookup f ienv = Some ty /\
      lookup f subt = Some vt /\
      lookup f subs = Some vs /\
      vm = mybad ty vt vs.
  Proof.
    i; dup LU; eapply list_assoc_someD in LU.
    hexploit (proj2 (in_map_fst f subm)); eauto; i.
    dup X; rewrite SUBm -SUBs in X; rewrite SUBm -SUBt in X0.
    eapply (list_assoc_map eqdec) in X; des.
    eapply (list_assoc_map eqdec) in X0; des.
    clear -LU0 X X0; unfold subm in *.
    ginduction ienv; i; [by auto|].
    destruct a; simpl map in LU0.
    match type of LU0 with lookup _ (?x :: _) = _ => destruct x eqn:E end.
    assert (l0 = l1).
    { destruct (@list_assoc _ tgt_val) in E
      ; try destruct (@list_assoc _ src_val) in E; clarify.
    }
    subst; apply list_assoc_cons_or in LU0; des; subst.
    - subst; rewrite X0 X in E; clarify; esplits; eauto.
    - hexploit IHl; eauto; i; des.
      rewrite list_assoc_cons_neq; eauto.
  Qed.

  
  Variable sia: state tm_gw * state lwa.
  Variable STAa: A.stable lwa.
  Variable INIga:
    forall (g: state (A.wlift lwa) -> A.vrelf)
           (GK: A.gknow (A.wlift lwa) g),
      A.confrel (GA.tm_gw : A.world_glob _ _) 
                (fun sga => g (sga, snd sia)) (fst sia) cigt cigm.
  Variable INIla:
    forall (g: state (A.wlift lwa) -> A.vrelf)
           (GK: A.gknow (A.wlift lwa) g),
      A.confrel lwa g sia cilt cilm.
  Variable REFa:
    forall (lt: lang_spec.val tgt_lsi) (lm: lang_spec.val mid_lsi),
      ~ A.rquery (GA.tm_gw: A.world_glob _ _) 
                 (fst sia) lt lm.
  Variable SIMa:
    forall (f: label) (IN: In f (map fst oenv)), exists vt vm,
      <<Et: vload tgt_lsi mt ancht subt f vt>> /\
      <<Em: vload mid_lsi mm anchm subm f vm>> /\
      <<SIM: A.tlsim (map fst ienv) subt subm sia vt vm>>.

  Variable sib: state ms_gw * state lwb.
  Variable STAb: stable lwb.
  Variable INIgb:
    forall (g: state (B.wlift lwb) -> B.vrelf)
           (GK: B.gknow (B.wlift lwb) g),
      B.confrel (GB.ms_gw: B.world_glob _ _)
                (fun sgb => g (sgb, snd sib)) (fst sib) cigm cigs.
  Variable INIlb:
    forall (g: state (B.wlift lwb) -> B.vrelf)
           (GK: B.gknow (B.wlift lwb) g),
      B.confrel lwb g sib cilm cils.
  Variable REFb:
    forall ty (lm: lang_spec.val mid_lsi) (ls: lang_spec.val src_lsi),
      ~ rquery (GB.ms_gw: B.world_glob _ _)
               (fst sib) ty lm ls.
  Variable SIMb:
    forall (f: label) ty (IN: lookup f oenv = Some ty), exists vm vs,
      <<Em: vload mid_lsi mm anchm subm f vm>> /\
      <<Es: vload src_lsi ms anchs subs f vs>> /\
      <<SIM: B.tlsim ienv subm subs sib vm vs ty>>.


  Definition si: state (wlift lw) :=
    scomp sia sib.


  Lemma gugua: forall (sga: state tm_gw) sgb lt lm
      (REG: (snd sga) lt lm)
      (X: ~ snd (minusa sga (bullet sga sgb)) lt lm),
    exists ty ls, snd sgb ty lm ls.
  Proof.
    i; apply NNPP; ii; apply X.
    destruct sga as [[[]]], sgb as [[[]]].
    rr; split; auto; i; rrdes FH; apply H.
    hexploit (GA.refst_bij REG FH); []; i; des.
    specialize (X0 eq_refl); des; subst; eauto. 
  Qed.

  Lemma gugub: forall (sgb: state ms_gw) sga ty lm ls
      (REG: (snd sgb) ty lm ls)
      (X: ~ snd (minusb sgb (bullet sga sgb)) ty lm ls),
    exists lt, snd sga lt lm.
  Proof.
    i; apply NNPP; ii; apply X.
    destruct sga as [[[]]], sgb as [[[]]].
    rr; split; auto; i; rrdes FH; apply H.
    hexploit (GB.refst_bij REG FH0); []; i; des.
    specialize (X1 eq_refl); des; subst; eauto. 
  Qed.

  Lemma STA:
    stable lw.
  Proof.
    clear - STAa STAb; ii; invc CR.
    destruct freshloc; des; rename x into freshloc.
    rewrite ->prod_eta in SC; clarify.
    rrdes COMPAT; invc COMPAT; des.
    rrdes RH; subst.
    rename hay into crtn, hby into crsn.
    rename hax into cctn, hbx into ccsn.

    dup RL.
    set (dummy := (pmemp, Just (nil, IEnat 42)) : conf mid_lsi).
    specialize (RL0 dummy (fst crma)); hexploit RL0.
    { unfold dummy; rr; eauto. }
    { rrdes CRga; subst; eauto. }
    { rrdes CRga; i; subst; apply DOMb in RD; des.
      destruct sga as [[[]]], sgb as [[[]]]; rrdes RD; eauto.
    }
    intro TMP; des; unfold dummy in TMP; rrdes TMP.
    rewrite (prod_eta clma) in Ec; s in Ec; clarify; noemp; dup E.
    rewrite (prod_eta clma) (prod_eta crma) (prod_eta clmb) (prod_eta crmb) in E0.
    s in E0; clarify.
    dup FIN; rename FIN into FINa; rename FIN0 into FINb; rewrite H2 in FINb.
    dup PV; rename PV into PVa; rename PV0 into PVb; rewrite H2 in PVb.
    clear ee H0 RL0 H1 H2.
    apply pmfinite_plusD in FINa; eauto; []; des.
    apply pmfinite_plusD in FINb; eauto; []; des.
    dup FINa0; dup FINb0; rrdes FINa0 FINb0.

    hexploit (GA.refst_fin (snd sga)); []; i; des.
    hexploit (GB.refst_fin (snd sgb)); []; i; des.
    rename l into sra_fin; rename l0 into srb_fin.

    set (ctx := map fst sra_fin ++ map snd sra_fin ++
                map (snd<*>fst) srb_fin ++ map snd srb_fin ++
                findom ++ findom0).

    set (sra'f lt lm := snd sga lt lm \/ 
                        exists ty ls, lm = freshloc ctx ty lt ls
                                   /\ snd sp' ty lt ls
                                   /\ ~ snd (bullet sga sgb) ty lt ls).
    set_prop sra'_ (GA.Build_refst sra'f) sra'_bij.
    { i; unfold sra'f in *; des.
      - eapply GA.refst_bij in IN; eauto; []; des.
        split; i; subst; symmetry; eauto.
      - split; i; subst; exfalso.
        + rrdes WHY.
          dup IN0; apply RELV in IN0; des.
          assert (forall vb, ~ snd (minusa sga (bullet sga sgb)) va vb).
          { ii; rrdes CRga; apply RELV0 in FH; des; subst.
            clear - RDa RDa0 MA.
            eapply (@pmvalidD_plus_read nat); [|exact RDa|exact RDa0]. 
            destruct ea, cctn, clt, cfa; rrdes MA; s in HP.
            rewrite -HP -!pmplus_assoc in HV; decomp_pmvalid HV; auto.
          }
          assert (exists ty vc, snd (bullet sga sgb) ty va vc).
          { apply NNPP; ii; eapply H.
            destruct sga as [[]], sgb as [[]]; s; split; eauto.
          }
          des; dup H0; apply EXT in H0.
          hexploit (refst_bij IN2 H0); []; i; des.
          specialize (X eq_refl); des; subst; vauto.
        + apply FIN, (in_map snd) in IN'; s in IN'.
          eapply (freshloc_fresh ctx ty va ls).
          unfold ctx at 2; apply in_or_app; right; apply in_or_app; auto.
      - split; i; subst; exfalso.
        + rrdes WHY.
          dup IN'0; apply RELV in IN'0; des.
          assert (forall vb, ~ snd (minusa sga (bullet sga sgb)) va vb).
          { ii; rrdes CRga; apply RELV0 in FH; des; subst.
            clear - RDa RDa0 MA.
            eapply (@pmvalidD_plus_read nat); [|exact RDa|exact RDa0]. 
            destruct ea, cctn, clt, cfa; rrdes MA; s in HP.
            rewrite -HP -!pmplus_assoc in HV; decomp_pmvalid HV; auto.
          }
          assert (exists ty vc, snd (bullet sga sgb) ty va vc).
          { apply NNPP; ii; eapply H.
            destruct sga as [[]], sgb as [[]]; s; split; eauto.
          }
          des; dup H0; apply EXT in H0.
          hexploit (refst_bij IN'2 H0); []; i; des.
          specialize (X eq_refl); des; subst; vauto.
        + subst.                
          apply FIN, (in_map snd) in IN; s in IN.
          eapply (freshloc_fresh ctx ty va' ls).
          unfold ctx at 2; apply in_or_app; right; apply in_or_app; auto.
      - split; i; subst.
        + eapply refst_bij in IN0; eauto; []; des.
          specialize (IN0 eq_refl); des; subst; auto.
        + hexploit freshloc_inj; eauto; i; []; des; subst; auto.
    }
    set_prop sra' sra'_ sra'_fin.
    { hexploit (refst_fin (snd sp')); []; i; des.
      exists (sra_fin ++ map (fun p => let '(ty, lt, ls) := p in
                                       (lt, freshloc ctx ty lt ls)) l).
      ii; unfold sra'f in IN; des.
      - eauto using in_or_app, FIN.
      - subst; apply FIN1 in IN0.
        apply in_or_app; right.
        apply in_map_iff; econs; inst_pairs; eauto.
    }
    unfold sra'_ in sra'; clear sra'_. 

    set (srb'f ty lm ls := snd sgb ty lm ls \/ 
                        exists lt, lm = freshloc ctx ty lt ls
                                /\ snd sp' ty lt ls
                                /\ ~ snd (bullet sga sgb) ty lt ls).
    set_prop srb'_ (GB.Build_refst srb'f) srb'_bij.
    { i; unfold srb'f in *; des.
      - eapply GB.refst_bij in IN; eauto; []; des.
        split; i; subst
        ; [specialize (IN eq_refl)|specialize (IN0 eq_refl)]; des; subst; auto.
      - split; i; subst; exfalso.
        + apply FIN0, (in_map (snd<*>fst)) in IN'; s in IN'.
          eapply (freshloc_fresh ctx ty lt vb).
          unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
        + rrdes WHY.
          dup IN0; apply RELV in IN0; des.
          assert (forall ty va, ~ snd (minusb sgb (bullet sga sgb)) ty va vb).
          { ii; rrdes CRgb; apply RELV0 in FH; des; subst.
            clear - RDb RDb0 MB.
            eapply (@pmvalidD_plus_read nat); [|exact RDb|exact RDb0].
            destruct eb as [[]], ccsn as [[]], cls as [[]], cfb as [[]].
            rrdes MB; s in Ec; clarify.
            rewrite -!pmplus_assoc in PV; decomp_pmvalid PV; auto.
          }
          assert (exists ty va, snd (bullet sga sgb) ty va vb).
          { apply NNPP; ii; eapply H.
            destruct sga as [[]], sgb as [[]]; s; split; eauto.
          }
          des; dup H0; apply EXT in H0.
          hexploit (refst_bij IN2 H0); []; i; des.
          specialize (X0 eq_refl); des; subst; vauto.
      - split; i; subst; exfalso.
        + subst.
          apply FIN0, (in_map (snd<*>fst)) in IN; s in IN.
          eapply (freshloc_fresh ctx ty' lt vb').
          unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
        + rrdes WHY.
          dup IN'0; apply RELV in IN'0; des.
          assert (forall ty va, ~ snd (minusb sgb (bullet sga sgb)) ty va vb).
          { ii; rrdes CRgb; apply RELV0 in FH; des; subst.
            clear - RDb RDb0 MB.
            eapply (@pmvalidD_plus_read nat); [|exact RDb|exact RDb0].
            destruct eb as [[]], ccsn as [[]], cls as [[]], cfb as [[]].
            rrdes MB; s in Ec; clarify.
            rewrite -!pmplus_assoc in PV; decomp_pmvalid PV; auto.
          }
          assert (exists ty va, snd (bullet sga sgb) ty va vb).
          { apply NNPP; ii; eapply H.
            destruct sga as [[]], sgb as [[]]; s; split; eauto.
          }
          des; dup H0; apply EXT in H0.
          hexploit (refst_bij IN'2 H0); []; i; des.
          specialize (X0 eq_refl); des; subst; vauto.
      - split; i; subst.
        + hexploit freshloc_inj; eauto; i; []; des; subst; auto.
        + eapply refst_bij in IN0; eauto; []; des.
          specialize (IN2 eq_refl); des; subst; auto.
    }
    set_prop srb' srb'_ srb'_fin.
    { hexploit (refst_fin (snd sp')); []; i; des.
      exists (srb_fin ++ map (fun p => let '(ty, lt, ls) := p in
                                       (ty, freshloc ctx ty lt ls, ls)) l).
      ii; unfold srb'f in IN; des.
      - eauto using in_or_app, FIN0.
      - subst; apply FIN1 in IN0.
        apply in_or_app; right.
        apply in_map_iff; econs; inst_pairs; eauto.
    }
    unfold srb'_ in srb'; clear srb'_. 


    assert (EXTa: extpub _ (sra': state GA.ref_sts) (snd sga)).
    { rr; eauto. }

    assert (EXTb: extpub _ (srb': state GB.ref_sts) (snd sgb)).
    { rr; eauto. }

    set (sga' := (fst (fst sp'), tt, sra') : state tm_gw).
    assert (EXTa': ext _ sga' sga).
    { split; [|by auto].
      rrdes EXT EXT; destruct sga as [[[]]], sgb as [[[]]]; destruct s.
      split; auto.
    }

    set (sgb' := (tt, snd (fst sp'), srb') : state ms_gw).
    assert (EXTb': ext _ sgb' sgb).
    { split; [|by auto].
      rrdes EXT EXT; destruct sga as [[[]]], sgb as [[[]]]; destruct s.
      split; auto.
    }

    assert (Esp': sp' = bullet sga' sgb').
    { unfold sga', sgb'; destruct sp' as [[[]]]; s; f_equal.
      apply refst_extensionality; s; split; intro X.
      - unfold sra'f, srb'f; rr.
        destruct (LEM (exists lm, snd sga la lm /\ snd sgb ty lm lb)).
        + des; eauto.
        + econs; split; right; esplits; eauto
          ; i; apply f; destruct sga as [[[]]], sgb as [[[]]]; rrdes FH; eauto.
      - rrdes X; unfold sra'f, srb'f in X, X0; des; subst.
        + destruct sga as [[[]]], sgb as [[[]]]; apply EXT; rr; eauto.
        + exfalso.
          apply FIN0, (in_map (snd<*>fst)) in X0; s in X0.
          eapply (freshloc_fresh ctx ty0 la ls).
          unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
        + exfalso.
          apply FIN, (in_map snd) in X; s in X.
          eapply (freshloc_fresh ctx ty lt lb).
          unfold ctx at 2; do 1 (apply in_or_app; right); apply in_or_app; auto.
        + apply freshloc_inj in X0; des; subst; auto.
    }

    assert (MED:
      forall lm, exists vmo,
        (forall lt ty ls (RRa: sra' lt lm) (RRb: srb' ty lm ls),
         exists vt vm vs,
           vmo = Some vm /\
           pmread (chp crtn) lt vt /\
           pmread (fst (fst crsn)) ls vs /\
           A.vclos (A.wlift lwa) (sga', sla) (gka g (sgb', slb) (sga', sla)) vt vm /\
           B.vclos (B.wlift lwb) (sgb', slb) (gkb g (sga', sla) (sgb', slb)) ty vm vs)
        /\
        (forall vm (E: vmo = Some vm), exists lt ty ls,
           <<RRa: sra' lt lm>> /\ <<RRb: srb' ty lm ls>>)).
    { rewrite Esp' in WHY; rrdes WHY; i.
      destruct (LEM (exists lt ty ls, sra' lt lm /\ srb' ty lm ls)).
      - des; hexploit RELV; [by eapply bullet_ref_comp; eauto|]; i; des.
        hexploit (@vclos_decomp g (sga', sla) (sgb', slb)); eauto.
        { apply vclos_lift; eapply vclos_mon; eauto.
          i; eapply gk_mon; eauto; repeat (split; auto).
        }
        i; des; esplits; i; eauto.
        hexploit (GA.refst_bij e RRa); []; i; des.
        specialize (X0 eq_refl); des; subst lt.
        hexploit (GB.refst_bij e0 RRb); []; i; des.
        specialize (X0 eq_refl); des; subst ty ls.
        rewrite Ea Eb; esplits; eauto.
      - exists None; split; i; exfalso.
        + apply f; eauto.
        + vauto.
    }

    set (hrmf := fun lm => proj1_sig (cid _ (MED lm))).
    set (crm := (Some hrmf, Nothing) : conf mid_lsi).

    set (crt' := crt # crtn).
    set (crma' := crma # crm).
    set (crmb' := crmb # crm).
    set (crs' := crs # crsn).

    assert (VLDt: pmvalid (chp crt')).
    { unfold crt'; rrdes MA; rewrite -HP in HV.
      destruct ea, cctn, crtn, crt, clt, cfa; ss.
      rewrite -!pmplus_assoc in HV.
      rewrite pmplus_comm; decomp_pmvalid HV; auto.
    }

    assert (VLDma: pmvalid (fst crma')).
    { unfold crma', crm; destruct crma; s.
      apply pmvalid_plus_read.
      - rrdes CRga; clarify; subst h; auto.
      - ii; vauto.
      - i; apply pmread_some in RDa0.
        unfold hrmf in RDa0; destruct (cid _ (MED x)); des; apply a0 in RDa0.
        des; s in RRa; s in RRb; unfold sra'f, srb'f in *; des.
        + rrdes CRga; clarify; subst e h.
          apply DOMb in RDa; des.
          destruct sga as [[[]]], sgb as [[[]]]; rrdes RDa.
          eapply RDa0; rr; eauto.
        + rewrite RRa in RRb.
          apply FIN0, (in_map (snd<*>fst)) in RRb; s in RRb.
          eapply (freshloc_fresh ctx ty0 lt ls0).
          unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
        + rewrite RRb in RRa.
          apply FIN, (in_map snd) in RRa; s in RRa.
          eapply (freshloc_fresh ctx ty lt0 ls).
          unfold ctx at 2; do 1 (apply in_or_app; right); apply in_or_app; auto.
        + rewrite RRa in RRb; apply freshloc_inj in RRb; des.
          subst x; rrdes CRga; clarify; subst e h.
          apply DOMb in RDa; des.
          destruct sga as [[[]]], sgb as [[[]]]; rrdes RDa.
          apply FIN, (in_map snd) in RDa; s in RDa.
          eapply (freshloc_fresh ctx ty0 lt ls0).
          unfold ctx at 2; do 1 (apply in_or_app; right); apply in_or_app; auto.
    }

    assert (VLDmb: pmvalid (fst crmb')).
    { unfold crmb', crm; destruct crmb; s.
      apply pmvalid_plus_read.
      - rrdes CRgb; clarify; subst h; auto.
      - ii; vauto.
      - i; apply pmread_some in RDa0.
        unfold hrmf in RDa0; destruct (cid _ (MED x)); des; apply a0 in RDa0.
        des; s in RRa; s in RRb; unfold sra'f, srb'f in *; des.
        + rrdes CRgb; clarify; subst e h.
          apply DOMa in RDa; des.
          destruct sga as [[[]]], sgb as [[[]]]; rrdes RDa.
          eapply RDa0; rr; eauto.
        + rewrite RRa in RRb.
          apply FIN0, (in_map (snd<*>fst)) in RRb; s in RRb.
          eapply (freshloc_fresh ctx ty0 lt ls0).
          unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
        + rewrite RRb in RRa.
          apply FIN, (in_map snd) in RRa; s in RRa.
          eapply (freshloc_fresh ctx ty lt0 ls).
          unfold ctx at 2; do 1 (apply in_or_app; right); apply in_or_app; auto.
        + rewrite RRa in RRb; apply freshloc_inj in RRb; des.
          subst x; rrdes CRgb; clarify; subst e h.
          apply DOMa in RDa; des.
          destruct sga as [[[]]], sgb as [[[]]]; rrdes RDa.
          apply FIN0, (in_map (snd<*>fst)) in RDa; s in RDa.
          eapply (freshloc_fresh ctx ty0 lt ls0).
          unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
    }

    assert (VLDs: pmvalid (fst (fst crs'))).
    { unfold crs'; rrdes MB.
      destruct eb as [[]], ccsn as [[]], crsn as [[]], crs as [[]], cls as [[]], cfb as [[]].
      ss; clarify; rewrite -H2 -pmplus_assoc in PV.
      rewrite pmplus_comm; decomp_pmvalid PV; auto.
    }

    clean.

    assert (TMP: forall ty lt ls
                (IN: snd sp' ty lt ls)
                (NIN: snd (bullet sga sgb) ty lt ls -> False),
              <<INa: snd sga' lt (freshloc ctx ty lt ls)>> /\
              <<INb: snd sgb' ty (freshloc ctx ty lt ls) ls>>).
    { i; rewrite Esp' in IN; rrdes IN.
      destruct sga as [[[]]], sgb as [[[]]].
      unfold sga', sgb', sra', srb', sra'f, srb'f in *; s in IN; s in IN0; s; des.
      - exfalso; apply NIN; rr; eauto.
      - exfalso.
        subst lm; apply FIN0, (in_map (snd<*>fst)) in IN0; s in IN0.
        eapply (freshloc_fresh ctx ty0 lt ls0).
        unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
      - exfalso.
        subst lm; apply FIN, (in_map snd) in IN; s in IN.
        eapply (freshloc_fresh ctx ty lt0 ls).
        unfold ctx at 2; do 1 (apply in_or_app; right); apply in_or_app; auto.
      - subst lm; apply freshloc_inj in IN0; des.
        subst ty lt ls; eauto; split; right; eauto.
    }
    

    assert (CRga_: tm_refconfrel (fun sga'' => gka g (sgb', slb) (sga'', sla))
                                 sga' crt' crma').
    { rr; unfold crt', crma', crm in *; rrdes CRga WHY.
      rewrite ->Ea, Eb, Ea0 in *; esplits.
      - s; noemp; eauto.
      - s; eauto.
      - eauto.
      - eauto.
      - destruct sga as [[] sra], sgb as [[] srb].
        i; apply pmread_plusD in ALC; des.
        + apply DOMa in ALC; des; rrdes ALC; eauto.
        + apply DOMa0 in ALC; des.
          destruct (LEM (exists lm, sra la lm /\ srb ty lm lb)); des; eauto.
          unfold sga', sra', sra'f; s; econs; right; esplits; eauto.
      - destruct sga as [[] sra], sgb as [[] srb].
        i; apply pmread_plusD in ALC; des.
        + apply DOMb in ALC; des; rrdes ALC; eauto.
        + apply pmread_some in ALC; unfold hrmf in ALC; destruct (cid _ (MED lb)).
          s in ALC; des; apply a0 in ALC; des; eauto. 
      - i; unfold sga', sra', sra'f in REG; s in REG.
        destruct (LEM (snd (minusa sga (bullet sga sgb)) la lb)).
        + apply RELV in r; []; nbdes.
          esplits; [M|M|]; Mdo by eapply pmread_plus; eauto.
          eapply A.vclos_mon; eauto.
          i; eapply gka_mon; try split; eauto.
        + des.
          { hexploit gugua; eauto; []; i; des.
            destruct sga as [[[]]], sgb as [[[]]].
            hexploit RELV0; [by apply EXT; rr; eauto|]; i; nbdes.
            assert (RD: hrmf lb = hrmf lb) by auto; unfold hrmf at 2 in RD.
            destruct (cid _ (MED lb)) as (vm & PP); s in RD; nbdes.
            hexploit PP; eauto; []; i; nbdes; subst vm; clear PP PP0.
            
            rewrite Ea0 in X1; s in X1; hexploit (pmread_uniq X1 RDa); []; i; subst vt.
            econs; exists vm0; esplits.
            - eapply pmread_plus; eauto using pmplus_comm.
            - eapply pmread_plus with (ha := Some hrmf); eauto.
              + eapply pmread_some; eauto.
              + eauto using pmplus_comm.
            - eapply PA.vclos_unlift in X3; eauto.
          }
          { hexploit TMP; eauto; []; i; des.
            assert (RD: hrmf lb = hrmf lb) by auto; unfold hrmf at 2 in RD.
            destruct (cid _ (MED lb)) as (vm & PP); s in RD; nbdes; subst lb.
            hexploit PP; eauto; []; i; nbdes; subst vm; clear PP PP0; esplits.
            - rewrite Ea0 in X0; s in X0.
              eapply pmread_plus; eauto using pmplus_comm.
            - eapply pmread_some in X.
              eapply pmread_plus; eauto using pmplus_comm.
            - eapply PA.vclos_unlift in X2; eauto.
          }
    }

    assert (CRgb_: ms_refconfrel (fun sgb'' => gkb g (sga', sla) (sgb'', slb))
                                 sgb' crmb' crs').
    { rr; unfold crs', crmb', crm in *; rrdes CRgb WHY.
      rewrite ->Ea, Eb, Eb0 in *; esplits.
      - s; noemp; eauto.
      - s; eauto.
      - eauto.
      - eauto.
      - destruct sga as [[] sra], sgb as [[] srb].
        i; apply pmread_plusD in ALC; des.
        + apply DOMa in ALC; des; rrdes ALC; eauto.
        + apply pmread_some in ALC; unfold hrmf in ALC; destruct (cid _ (MED la)).
          s in ALC; des; apply a0 in ALC; des; eauto.
      - destruct sga as [[] sra], sgb as [[] srb].
        i; apply pmread_plusD in ALC; des.
        + apply DOMb in ALC; des; rrdes ALC; eauto.
        + apply DOMb0 in ALC; des.
          destruct (LEM (exists lm, sra la lm /\ srb ty lm lb)); des; eauto.
          unfold sgb', srb', srb'f; s; do 2 econs; right; esplits; eauto.
      - i; unfold sgb', srb', srb'f in REG; s in REG.
        destruct (LEM (snd (minusb sgb (bullet sga sgb)) ty la lb)).
        + apply RELV in r; []; nbdes.
          esplits; [M|M|]; Mdo by eapply pmread_plus; eauto.
          eapply B.vclos_mon; eauto.
          i; eapply gkb_mon; try split; eauto.
        + des.
          { hexploit gugub; eauto; []; i; des.
            destruct sga as [[[]]], sgb as [[[]]].
            hexploit RELV0; [by apply EXT; rr; eauto|]; i; nbdes.
            assert (RD: hrmf la = hrmf la) by auto; unfold hrmf at 2 in RD.
            destruct (cid _ (MED la)) as (vm & PP); s in RD; nbdes.
            hexploit PP; eauto; []; i; nbdes; subst vm; clear PP PP0.
            
            rewrite Eb0 in X2; s in X2; hexploit (pmread_uniq X2 RDb); []; i; subst vs.
            exists vm0; econs; esplits.
            - eapply pmread_plus with (ha := Some hrmf); eauto.
              + eapply pmread_some; eauto.
              + eauto using pmplus_comm.
            - eapply pmread_plus; eauto using pmplus_comm.
            - eapply PB.vclos_unlift in X4; eauto.
          }
          { hexploit TMP; eauto; []; i; des.
            assert (RD: hrmf la = hrmf la) by auto; unfold hrmf at 2 in RD.
            destruct (cid _ (MED la)) as (vm & PP); s in RD; nbdes; subst la.
            hexploit PP; eauto; []; i; nbdes; subst vm; clear PP PP0; esplits.
            - eapply pmread_some in X.
              eapply pmread_plus; eauto using pmplus_comm.
            - rewrite Eb0 in X1; s in X1.
              eapply pmread_plus; eauto using pmplus_comm.
            - eapply PB.vclos_unlift in X3; eauto.
          }
    }

    assert (RGSa: A.realizable_global_state clt clma tm_gw
                    (fun sga'' => gka g (sgb', slb) (sga'', sla)) sga').
    { dup CRga_; rrdes CRga_.
      hexploit A.lwconf_extra; eauto; []; i; des.
      rrdes X0; subst clma; unfold sga'; rrdes WHX.
      econs; exists dummy; esplits; econs; first 1 last; first last.
      - rr; esplits; first 1 last; eauto; split; rr; eauto.
      - unfold dummy; rewrite Eb; instantiate (2 := cemp); s; noemp.
        rr; esplits; eauto.
        + eapply pmvalid_plus_read; eauto.
          { apply pmvalid_plusD in PVa; des; auto. }
          i; apply DOMb in RDa; des.
          unfold sga', sra', sra'f in RDa; s in RDa; des.
          * specialize (RL dummy {x |-> yb}).
            hexploit RL; unfold dummy; eauto.
            { unfold dummy; rr; eauto. }
            { i; apply pmread_cellD in RD; des; subst x yb; eauto. }
            i; des; rrdes H; s in Ec; clarify; rewrite -H1 in PV; noemp.
            eapply (pmvalidD_plus_read PV); eauto.
          * apply FINa0 in RDa0. subst x.
            apply (freshloc_fresh ctx ty la ls).
            unfold ctx at 2; do 4 (apply in_or_app; right); apply in_or_app; auto.
        + apply tm_refconfrel_fin in CRga_0; rewrite Eb in CRga_0; des; eauto.
      - unfold crt'; rewrite <-!cplus_assoc, (cplus_swap _ crtn) in MA.
        rewrite (cplus_assoc cctn) (cplus_assoc _ crtn) -(cplus_assoc cctn) in MA; eauto. 
      - unfold dummy; split; rr; eauto.
    }

    assert (RGSb: B.realizable_global_state clmb cls ms_gw
                    (fun sgb'' => gkb g (sga', sla) (sgb'', slb)) sgb').
    { dup CRga_; rrdes CRga_.
      hexploit A.lwconf_extra; eauto; []; i; des.
      rrdes X0; subst clma; unfold sgb'; rrdes WHX.
      exists dummy; esplits; econs; first 1 last; first last.
      - rr; esplits; first 1 last; eauto; split; rr; eauto.
      - unfold crs'; rewrite <-!cplus_assoc, (cplus_swap _ crsn) in MB.
        rewrite (cplus_assoc ccsn) (cplus_assoc _ crsn) -(cplus_assoc ccsn) in MB; eauto. 
      - 
        noemp; unfold crmb'; rewrite (cplus_comm crmb) -(cplus_assoc crm).
        rewrite ->(cplus_assoc crmb), <-E, <-cplus_assoc.
        rewrite (cplus_assoc crm) (cplus_comm crm); fold crma'.
        
        unfold dummy; rewrite Eb; instantiate (2 := cemp); s; noemp.
        rr; esplits; eauto.
        + eapply pmvalid_plus_read; eauto.
          { apply pmvalid_plusD in PVa; des; auto. }
          i; apply DOMb in RDa; des.
          unfold sga', sra', sra'f in RDa; s in RDa; des.
          * specialize (RL dummy {x |-> yb}).
            hexploit RL; unfold dummy; eauto.
            { unfold dummy; rr; eauto. }
            { i; apply pmread_cellD in RD; des; subst x yb; eauto. }
            i; des; rrdes H; s in Ec; clarify; rewrite -H1 in PV; noemp.
            eapply (pmvalidD_plus_read PV); eauto.
          * apply FINa0 in RDa0. subst x.
            apply (freshloc_fresh ctx ty la ls).
            unfold ctx at 2; do 4 (apply in_or_app; right); apply in_or_app; auto.
        + apply tm_refconfrel_fin in CRga_0; rewrite Eb in CRga_0; des; eauto.
      - unfold dummy; split; rr; eauto.
    }

    assert (CRla': A.confrel lwa (gka g (sgb', slb)) (sga, sla) clt clma).
    { eapply A.confrel_mon; eauto.
      i; eapply gka_mon; rr; eauto.
    }

    assert (CRlb': B.confrel lwb (gkb g (sga', sla)) (sgb, slb) clmb cls).
    { eapply B.confrel_mon; eauto.
      i; eapply gkb_mon; rr; eauto.
    }

    dup EXTa; apply extpub_in_ext in EXTa.
    exploit STAa; eauto 10 using gka_gknow; []; i; des.
    dup EXTb; apply extpub_in_ext in EXTb.
    exploit STAb; eauto using gkb_gknow; []; i; des.
    rename s' into sla', s'0 into slb'.


    assert (old_or_new:
      forall ty lt lm ls (SRa: sra' lt lm) (SRb: srb' ty lm ls),
        <<OLD: snd sga lt lm /\ snd sgb ty lm ls>> \/
        <<NEW: lm = freshloc ctx ty lt ls /\ snd sp' ty lt ls /\
               ~ snd (bullet sga sgb) ty lt ls>>).
    { clear - freshloc_fresh freshloc_inj FIN0 FIN.
      destruct sga as [[[]]], sgb as [[[]]].
      unfold sra', srb', sra'f, srb'f in *; s; i; des; eauto; subst lm.
      - exfalso.
        apply FIN0, (in_map (snd<*>fst)) in SRb; s in SRb.
        eapply (freshloc_fresh ctx ty0 lt ls0).
        unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
      - exfalso.
        apply FIN, (in_map snd) in SRa; s in SRa.
        eapply (freshloc_fresh ctx ty lt0 ls).
        unfold ctx at 2; do 1 (apply in_or_app; right); apply in_or_app; auto.
      - apply freshloc_inj in SRb; des.
        subst ty lt ls; eauto; split; right; eauto.
    }                           


    assert (Ea: snd (minusa sga (bullet sga sgb)) =
                snd (minusa sga' (bullet sga' sgb'))).
    { destruct sga as [[[]]], sgb as [[[]]]; s.
      apply GA.refst_extensionality; split; intro IN; rrdes IN.
      - rr; split; eauto.
        i; rrdes FH; hexploit old_or_new; eauto; []; i; des.
        + eapply IN0; rr; eauto.
        + dup IN; apply EXTa0 in IN1.
          hexploit (GA.refst_bij IN1 FH); []; i; des.
          specialize (X eq_refl); subst lm. 
          subst lb; apply FIN, (in_map snd) in IN; s in IN.
          eapply (freshloc_fresh ctx ty la ls).
          unfold ctx at 2; do 1 (apply in_or_app; right); apply in_or_app; auto.
      - unfold sra', sra'f in IN; s in IN; des.
        + rr; split; auto.
          i; eapply IN0; rrdes FH; rr; esplits; eauto.
        + exfalso; subst lb; eapply IN0.
          rr; econs; eapply TMP; eauto.
    }
          
    assert (Eb: snd (minusb sgb (bullet sga sgb)) =
                snd (minusb sgb' (bullet sga' sgb'))).
    { destruct sga as [[[]]], sgb as [[[]]]; s.
      apply refst_extensionality; split; intro IN; rrdes IN.
      - rr; split; eauto.
        i; rrdes FH; hexploit old_or_new; eauto; []; i; des.
        + eapply IN0; rr; eauto.
        + dup IN; apply EXTb0 in IN1.
          hexploit (GB.refst_bij IN1 FH0); []; i; des.
          specialize (X0 eq_refl); des; subst lm. 
          subst la; apply FIN0, (in_map (snd<*>fst)) in IN; s in IN.
          eapply (freshloc_fresh ctx ty lt lb).
          unfold ctx at 2; do 2 (apply in_or_app; right); apply in_or_app; auto.
      - unfold srb', srb'f in IN; s in IN; des.
        + rr; split; auto.
          i; eapply IN0; rrdes FH; rr; esplits; eauto.
        + exfalso; subst la; eapply IN0.
          rr; econs; eapply TMP; eauto.
    }

    assert (CRga': tm_refconfrel' (snd (minusa sga' (bullet sga' sgb')))
                                  (fun sga'' => gka g (sgb', slb') (sga'', sla')) sga'
                                  crt crma).
    { rewrite -Ea.
      eapply tm_refconfrel'_mon.
      - eapply tm_refconfrel'_smon; eauto.
        s; i; eapply gka_mon; rr; eauto.
      - s; i; eapply gka_mon; rr; eauto.
    }

    assert (CRgb': ms_refconfrel' (snd (minusb sgb' (bullet sga' sgb')))
                                  (fun sgb'' => gkb g (sga', sla') (sgb'', slb')) sgb'
                                  crmb crs).
    { rewrite -Eb.
      eapply ms_refconfrel'_mon.
      - eapply ms_refconfrel'_smon; eauto.
        s; i; eapply gkb_mon; rr; eauto.
      - s; i; eapply gkb_mon; rr; eauto.
    }

    rrdes EXTa0 EXTb0; destruct sp' as [[[]]].
    exists ((sra', sla'), (srb', slb')); splits; [by split; split; auto|].
    rewrite Esp'; econs.
    - unfold sga', sgb', scomp; inst_pairs; s; eauto.
    - eauto.
    - eauto.
    - eapply A.confrel_mon; eauto.
      i; eapply gka_mon; rr; eauto.
    - eapply B.confrel_mon; eauto.
      i; eapply gkb_mon; rr; eauto.
    - eauto.
    - i; hexploit (pmplus_from_pred hm (fun lm => exists lt, snd sga lt lm)); [].
      i; des. subst hm; hexploit (RL em hP); eauto.
      { apply pmvalid_plusD in VLD; des; auto. }
      { i; apply RDP in RD; des; eauto. }
      i; des. rrdes H; subst mm; rrdes CO1; subst em.
      hexploit A.lwconf_extra; eauto; []; i; des; rrdes X0.
      subst clma; s in Ec; noemp; clarify; s; noemp.
      econs; rr; esplits; eauto.
      + rewrite -pmplus_assoc.
        apply pmvalid_plus_plus; auto.
        dup PV; apply pmvalid_plusD in VLD; apply pmvalid_plusD in PV0; des.
        apply pmvalid_plus_read; auto.
        i; apply RDNP in RDa. des; apply FINa0 in RDa0; apply DOM in RDa; des.
        unfold sra', sra'f in RDa. s in RDa. des; [by exfalso; apply RDa1; vauto|].
        subst x; apply (freshloc_fresh ctx ty lt ls).
        unfold ctx at 2; do 4 (apply in_or_app; right); apply in_or_app; auto. 
      + apply pmfinite_plus; auto.
        hexploit (GA.refst_fin sra'); []; i; des.
        exists (map snd l1); i; apply DOM in RD; des.
        apply FIN4 in RD; apply in_map_snd; eauto.
    - eauto.
    - eauto.
  Qed.


  Lemma INIg: forall g
      (GK: gknow (wlift lw) g),
    model.confrel ts_gw (fun sg => g (sg, snd si)) (fst si) cigt cigs. 
  Proof.
    clear -REFa REFb INIga INIgb; i.
    hexploit INIga; eauto using gka_gknow; []; i; rrdes X.
    hexploit INIgb; eauto using gkb_gknow; []; i; rrdes X.
    subst; unfold si; clean.
    rr; esplits; eauto.
    - rrdes WHX WHX0; destruct sia as [[[]]], sib as [[[]]].
      rr; esplits; rewrite -?bullet_tgt -?bullet_src; eauto.
    - apply tm_refconfrel_emp in WHY; [|by i; eapply REFa; rr; eauto].
      apply ms_refconfrel_emp in WHY0; [|by i; eapply REFb; rr; eauto 10].
      des; subst.
      do 2 exists pmemp; splits; eauto using pmvalid_emp. 
      + i; apply pmread_empD in ALC; vauto.
      + i; apply pmread_empD in ALC; vauto.
      + destruct sia as [[[]]], sib as [[[]]].
        i; hexploit bullet_ref_decomp; [exact REG|]; i; des.
        exfalso; eapply REFa; rr; eauto.
  Grab Existential Variables.
    exact sia. exact sib.
  Qed.


  Lemma INIl: forall g
      (GK: gknow (wlift lw) g),
    confrel lw g si cilt cils.
  Proof.
    clear -REFa REFb INIga INIgb INIla INIlb; i.
    hexploit INIga; eauto using gka_gknow; []; i; rrdes X.
    hexploit INIgb; eauto using gkb_gknow; []; i; rrdes X.
    hexploit INIla; eauto using gka_gknow; []; clear INIla; i.
    hexploit INIlb; eauto using gkb_gknow; []; clear INIlb; i.
    subst; unfold si; clean.

    hexploit tm_refconfrel_emp; [by i; eapply REFa; rr; eauto|]
    ; eauto; []; i; des; hexploit X; eauto; []; i; des; subst.
    clear X X0.
    hexploit ms_refconfrel_emp; [by i; eapply REFb; rr; eauto 10|]
    ; eauto; []; i; des; hexploit X; eauto; []; i; des; subst.
    clear X X0.

    hexploit A.lwconf_extra; eauto; []; i; des.
    hexploit B.lwconf_extra; eauto; []; i; des.
    destruct cilt; rrdes X X1 X2; clean; s in X; s in X3; clarify.

    destruct sia, sib; clean.
    rewrite -(pmplus_emp chp) -(pmplus_emp h0); noemp.

    econs.
    - unfold scomp; eauto.
    - insts $ (@cemp mid_lsi) (@cemp tgt_lsi).
      unfold scomp. clean. esplits; eauto.
      clear -WHY. rrdes WHY.
      s in Ea; unfold tgt_cemp in Ea; clarify.
      s in Eb; unfold mid_cemp in Eb; clarify.
      rr; esplits; s; unfold tgt_cemp, mid_cemp; eauto.
      + i; eapply pmread_empD in ALC; vauto.
      + i; eapply pmread_empD in ALC; vauto.
      + destruct s as [[[]]]; destruct s1 as [[[]]]; clean.
        i; hexploit RELV; rrdes REG; eauto.
    - insts $ (@cemp src_lsi) (@cemp mid_lsi).
      unfold scomp. clean. esplits; eauto.
      clear -WHY0. rrdes WHY0.
      s in Ea; unfold mid_cemp in Ea; clarify.
      s in Eb; unfold src_cemp in Eb; clarify.
      rr; esplits; s; unfold mid_cemp, src_cemp; eauto.
      + i; eapply pmread_empD in ALC; vauto.
      + i; eapply pmread_empD in ALC; vauto.
      + destruct s as [[[]]]; destruct s1 as [[[]]]; clean.
        i; hexploit RELV; rrdes REG; eauto.
    - esplits; eauto; inst; s; eauto.
    - esplits; eauto; inst; s; eauto.
    - auto.
    - unfold cilm; i; rrdes CO; subst; s; noemp.
      econs; rr; esplits; s; eauto.
      hexploit (GA.refst_fin (snd s)); []; i; des.
      exists (map snd l); i. apply DOM in RD; des.
      apply FIN in RD; apply in_map_snd; eauto. 
    - noemp; auto.
    - noemp; auto.
  Qed.


  Lemma REF: forall ty lt ls,
    ~ model.rquery ts_gw (fst si) ty lt ls.
  Proof.
    unfold si; i.
    hexploit rquery_decomp; eauto; i; des; eauto.
  Qed.





  (** ** Bad stuff. *)

  Lemma mybad_bad_fun: forall (fty: flextyp) ty ty' vt vs
      (E: Tarrow ty ty' = fty),
    bad_fun (mybad fty vt vs).
  Proof.
    unfold mybad; i; destruct (flext_of fty); vauto.
  Qed.

  Lemma mybad_bad_gen: forall (fty: flextyp) ty vt vs
      (E: Tforall ty = fty),
    bad_gen (mybad fty vt vs).
  Proof.
    unfold mybad; i; destruct (flext_of fty); vauto.
  Qed.

  Definition badcontm: lang_spec.cont mid_lsi :=
    IVclo nil (IEcont 0 (IEnat 42)).

  (** Next lemma relies on [badcontm] taking one step to a [halted] machine. *)
  Lemma badcontm_ret_error: forall sb vm em cm mm mm'
      (CQ: B.cquerya (B.wlift lwb) sb (ret vm badcontm) em)
      (RL: real mid_lsi (em # cm) mm)
      (ST: rtc (step mid_lsi etau) mm mm'),
    ~ error mid_lsi mm'.
  Proof.
    clear; i; invc CQ; destruct cm; simpl cplus in RL; noemp.
    dup RL; rrdes RL. subst; destruct e; vauto.
    eapply clos_rt_rt1n_iff in ST; invc ST; first by eauto.
    inversion H; subst; last by (eapply STK; econs; eauto).
    invc STEP; rewrite GETk in GET; rewrite GETv in GETV; clarify.
    destruct H0; last (by apply halted_step in H0; rr; eauto).
    eapply FH; rr; esplits; eauto.
  Qed.

  Lemma bad_fun_not_error: forall
      mm mm' (STm: rtc (step mid_lsi etau) mm mm')
      em csm (RLm: real _ (em # csm) mm)
      s fm vm km (APPm: B.cquerya (B.wlift lwb) s (app fm vm km) em)
      (BFm: bad_fun fm),
    ~ error _ mm'.
  Proof. clear.
    i; invc BFm APPm; rrdes RLm.
    destruct csm; destruct e; vauto; []; s in Ec; clarify; noemp.
    apply halted_tausteps in STm; [|by rr; auto].
    subst; eapply FH; rr; esplits; eauto.
    
  Qed.

  Lemma bad_gen_not_error: forall
      mm mm' (STm: rtc (step mid_lsi etau) mm mm')
      em csm (RLm: real _ (em # csm) mm)
      s gm km (INSTm: B.cquerya (B.wlift lwb) s (inst gm km) em)
      (BGm: bad_gen gm),
    ~ error _ mm'.
  Proof. clear.
    i; invc BGm INSTm; rrdes RLm.
    destruct csm; destruct e; vauto; []; s in Ec; clarify; noemp.
    apply halted_tausteps in STm; [|by rr; auto].
    subst; eapply FH; rr; esplits; eauto.
  Qed.

  Lemma ret_halted: forall
      sa vm km em (RET: A.cqueryb (A.wlift lwa) sa (ret vm km) em)
      csm mm (RL: real _ (em # csm) mm),
    ~ halted _ mm.
  Proof. clear.
    i; rrdes RL; invc RET.
    destruct csm; s in Ec; destruct e; vauto.
  Qed.                            

  Lemma bad_ret_app: forall
      sb fm vm' km em (APP: B.cquerya (B.wlift lwb) sb (app fm vm' km) em)
      (BF: bad_fun fm)
      csm mm (RL: real _ (em # csm) mm)
      mm' (ST: rtc (step mid_lsi etau) mm mm')
      em' csm' (RL': real _ (em' # csm') mm')
      sa vm (RET: A.cqueryb (A.wlift lwa) sa (ret vm badcontm) em'),
    False.
  Proof. clear.
    i; invc BF APP; rrdes RL; destruct csm; s in Ec; destruct e; vauto.
    hexploit halted_tausteps; eauto; inst; rr; eauto; []; i; des; clarify.
    invc RET; rrdes RL'; destruct csm'; s in Ec; destruct e; vauto.
  Qed.

  Lemma bad_ret_inst: forall
      sb gm km em (INST: B.cquerya (B.wlift lwb) sb (inst gm km) em)
      (BG: bad_gen gm)
      csm mm (RL: real _ (em # csm) mm)
      mm' (ST: rtc (step mid_lsi etau) mm mm')
      em' csm' (RL': real _ (em' # csm') mm')
      sa vm (RET: A.cqueryb (A.wlift lwa) sa (ret vm badcontm) em'),
    False.
  Proof. clear.
    i; invc BG INST; rrdes RL; destruct csm; s in Ec; destruct e; vauto.
    hexploit halted_tausteps; eauto; inst; rr; eauto; []; i; des; clarify.
    invc RET; rrdes RL'; destruct csm'; s in Ec; destruct e; vauto.
  Qed.

  Lemma esim_halteda: forall i ka kb g si s sync ty ea eb
      (SIM: B.esim (wlift lwb) i ka kb g si s sync ty ea eb)
      csa csb cfa cfb ma mb (CNF: B.configure (wlift lwb) g s sync ea eb csa csb cfa cfb ma mb)
      (HA: halted _ ma)
      (GK: B.gknow _ g)
      (XTRa: extra cfa) (XTRb: extra cfb),
    exists mb', <<ST: rtc (step _ etau) mb mb'>> /\ <<ERR: error _ mb'>>.
  Proof. clear.
    i; apply esim_unfold in SIM; invc SIM.
    hexploit ESIM; eauto; []; clear ESIM; i; des; invc H; eauto.
    invc RETa CNF0; rrdes MA; destruct cfa, csa'; destruct e, e0; s in Ec; vauto.
  
  Qed.

  Lemma bad_app_ret: forall sa sb vm fm vm' km em em' csm csm' mm mm'
      (RET: B.cquerya (B.wlift lwb) sb (ret vm badcontm) em)
      (RL: real _ (em # csm) mm)
      (ST: tc (step mid_lsi etau) mm mm')
      (RL': real _ (em' # csm') mm')
      (APP: A.cqueryb (A.wlift lwa) sa (app fm vm' km) em')
      (BF: bad_fun fm),  
    False.
  Proof. clear.
    i; invc RET; rrdes RL; destruct csm; destruct e; s in Ec; clarify; noemp.
    apply tc_first in ST; des.
    invc ST; [|by exfalso; eapply STK; econs; eauto].
    invc STEP; rewrite GET in GETk; clarify.
    hexploit halted_tausteps; eauto; rr; eauto; []; i; inst; subst.
    invc APP; rrdes RL'; destruct csm'; destruct e0; s in Ec; clarify.
  Qed.

  Lemma bad_inst_ret: forall sa sb vm gm km em em' csm csm' mm mm'
      (RET: B.cquerya (B.wlift lwb) sb (ret vm badcontm) em)
      (RL: real _ (em # csm) mm)
      (ST: tc (step mid_lsi etau) mm mm')
      (RL': real _ (em' # csm') mm')
      (INST: A.cqueryb (A.wlift lwa) sa (inst gm km) em')
      (BG: bad_gen gm),  
    False.
  Proof. clear.
    i; invc RET; rrdes RL; destruct csm; destruct e; s in Ec; clarify; noemp.
    apply tc_first in ST; des.
    invc ST; [|by exfalso; eapply STK; econs; eauto].
    invc STEP; rewrite GET in GETk; clarify.
    hexploit halted_tausteps; eauto; rr; eauto; []; i; inst; subst.
    invc INST; rrdes RL'; destruct csm'; destruct e0; s in Ec; clarify.
  Qed.

  Lemma esim_haltedb: forall La Lb (w: A.world La Lb) i ka kb g si s sync ea eb
      (SIM: A.esim w i ka kb g si s sync ea eb)
      csa csb cfa cfb ma mb (CNF: A.configure w g s sync ea eb csa csb cfa cfb ma mb)
      (HA: halted _ mb)
      (GK: A.gknow _ g)
      (XTRa: extra cfa) (XTRb: extra cfb),
    <<ERR: error _ mb>> \/
    <<RET: exists eb' csb' vb s', real _ (eb' # csb') mb
                               /\ A.cqueryb w s' (ret vb kb) eb'>>.
    
    
    
  Proof. clear.
    i; depgen sync; revert s ea eb. depgen ma. depgen csa. depgen csb.
    induction i using (well_founded_ind (tc_well_founded (wfo_wf w))).
    i; apply PA.esim_unfold in SIM; invc SIM.
    hexploit ESIM; eauto; []; clear ESIM; i; des; invc H0.
    - hexploit halted_tausteps; eauto; []; i; inst; subst; eauto.
    - hexploit progress_if_not_halted; eauto; []; i; des.
      hexploit EST; eauto; []; i; des.
      + 
        apply rtc_tc in STP; des; subst; last first.
        { exfalso.
          assert (tc (step _ etau) mb mb') by (eapply rtc_tc_tc; eauto).
          apply tc_first in H0; des; rrdes H0.
          eapply halted_step; eauto.
        }
        invc ES; des.
        * exfalso.
          assert (tc anystep mb' mbt).
          { eapply rtc_tc_tc; last by econs 1; eauto.
            eapply rtc_mon; [econs 3|]; eauto.
          }
          apply tc_first in H0; des; rrdes H0.
          eapply halted_step; eauto.
        * subst; eapply H; eauto.
      + 
        exfalso.
        assert (ST': rtc (step _ etau) mb mb') by (econs 3; eauto).
        hexploit halted_tausteps; eauto; []; i; inst; subst.
        invc ES; eapply halted_step; eauto.
    - right.
      hexploit halted_tausteps; eauto; []; i; inst; subst.
      inv CNF0; esplits; eauto.
  Qed.

  Lemma mess_fun: forall y z g sb sa s' sync evt mbt ea' eb' csa' csb' cfa ma'
                         s'0 ea'0 csa'0 va ka kt ia ff
      (ES: A.esim_progress (PA.esim_ext _ (gka_enc g sb)) ia kt badcontm
                           (gka g sb) sa s' sync evt z mbt ea' eb' \/
           PA.esim_call_ext kt badcontm (gka_enc g sb) (gka g sb)
                            sa s' sync evt z mbt ea' eb')
      (APPa : B.cquerya (B.wlift lwb) s'0 (app ff va ka) ea'0)
      (CNF : A.configure (A.wlift lwa) (gka g sb) s' sync ea' eb' csa' csb' cfa cemp ma' mbt)
      (GK: gknow (wlift lw) g)
      (ST: rtc (step mid_lsi etau) y z)
      (MA: real mid_lsi (ea'0 # csa'0) y)
      (BF: bad_fun ff)
      (XTRa: extra cfa),
    False.
  Proof.
    i; invc BF APPa; rrnbdes MA; destruct csa'0; destruct e; vauto; [].
    s in Ec; clarify; noemp.
    hexploit halted_steps; first last; first 1 last.
    { eapply rtc_mon; eauto. }
    { rr; eauto. }
    i; subst; des.
    + invc ES; des.
      * eapply halted_step in STEP; eauto; rr; eauto.
      * subst.
        eapply PA.esim_ext_impl_esim in ES0; last first.
        { i; eapply PA.gkmin_fold; eauto. }
        hexploit esim_haltedb; eauto; rr; eauto.
        { eapply gka_gknow; eauto. }
        { s; unfold mid_cemp; eauto. }
        i; des.
        { invc CNF; eapply ERR; eauto. }
        { invc RET0; rrdes RET; destruct csb'0; destruct e; s in Ec; clarify. }
    + invc ES; eapply halted_step in ST0; eauto; rr; eauto.
  Qed.

  Lemma mess_gen: forall y z g sb sa s' sync evt mbt ea' eb' csa' csb' cfa ma'
                         s'0 ea'0 csa'0 ka kt ia gg
      (ES: A.esim_progress (PA.esim_ext _ (gka_enc g sb)) ia kt badcontm
                           (gka g sb) sa s' sync evt z mbt ea' eb' \/
           PA.esim_call_ext kt badcontm (gka_enc g sb) (gka g sb)
                            sa s' sync evt z mbt ea' eb')
      (INSTa : B.cquerya (B.wlift lwb) s'0 (inst gg ka) ea'0)
      (CNF : A.configure (A.wlift lwa) (gka g sb) s' sync ea' eb' csa' csb' cfa cemp ma' mbt)
      (GK: gknow (wlift lw) g)
      (ST: rtc (step mid_lsi etau) y z)
      (MA: real mid_lsi (ea'0 # csa'0) y)
      (BG: bad_gen gg)
      (XTRa: extra cfa),
    False.
  Proof.
    
    i; invc BG INSTa; rrnbdes MA; destruct csa'0; destruct e; vauto; [].
    s in Ec; clarify; noemp.
    hexploit halted_steps; first last; first 1 last.
    { eapply rtc_mon; eauto. }
    { rr; eauto. }
    i; subst; des.
    + invc ES; des.
      * eapply halted_step in STEP; eauto; rr; eauto.
      * subst.
        eapply PA.esim_ext_impl_esim in ES0; last first.
        { i; eapply PA.gkmin_fold; eauto. }
        hexploit esim_haltedb; eauto; rr; eauto.
        { eapply gka_gknow; eauto. }
        { s; unfold mid_cemp; eauto. }
        i; des.
        { invc CNF; eapply ERR; eauto. }
        { invc RET0; rrdes RET; destruct csb'0; destruct e; s in Ec; clarify. }
    + invc ES; eapply halted_step in ST0; eauto; rr; eauto.
  Qed.



  Lemma mess_ret: forall y z g sb' sb''' sa sa' sa'' sync evt mbt
      ea' eb' ma' cfa csb' csa' ea'0 csa'0 va kt ia mb' x cfb csb'0 eb'0
      (ES: A.esim_progress (PA.esim_ext _ (gka_enc g sb')) ia kt badcontm
                           (gka g sb') sa sa'' sync evt z mbt ea' eb' \/
           PA.esim_call_ext kt badcontm (gka_enc g sb') (gka g sb')
                            sa sa'' sync evt z mbt ea' eb')
      (RETa : B.cquerya (B.wlift lwb) sb''' (ret va badcontm) ea'0)
      (CNFb : B.configure (B.wlift lwb) (gkb g sa') sb''' true ea'0 eb'0 csa'0 csb'0 cemp
             cfb x mb')
      (CNFa : A.configure (A.wlift lwa) (gka g sb') sa'' sync ea' eb' csa'
            csb' cfa cemp ma' mbt)
      (GK: gknow (wlift lw) g)
      (ST : step mid_lsi etau x y)
      (STP : rtc (step mid_lsi etau) y z)
      (XTRa: extra cfa),
    False.
  Proof.
    i.
    inv RETa CNFb. rrnbdes MA. destruct csa'0. destruct e; vauto. s in Ec. clarify. noemp.
    invc ST; [|by eapply STK; econs; eauto].
    invc STEP.
    rewrite GET in GETk; clarify.
    hexploit halted_steps; last first; first 1 last.
    { eapply rtc_mon; eauto. }
    { rr; eauto. }
    i; subst; des; invc ES; des.
    - eapply halted_step; last first; eauto; rr; eauto.
    - clarify.
      eapply PA.esim_ext_impl_esim in ES0; last first.
      { i; eapply PA.gkmin_fold; eauto. }
      hexploit esim_haltedb; eauto using gka_gknow, extra_emp; rr; eauto.
      i; des.
      { invc CNFa; eapply ERR; eauto. } 
      { invc RET0; rrdes RET; destruct csb'1; destruct e; s in Ec; clarify. } 
    - eapply halted_step; last first; eauto; rr; eauto.
  Qed.


  Lemma bad_ret_call: forall sb vm em cm mm mm' em' cm' sa ga ga' krel et' evt
      (RET: B.cquerya (B.wlift lwb) sb (ret vm badcontm) em)
      (RL: real _ (em # cm) mm)
      (ST: step mid_lsi evt mm mm')
      (RL: real _ (em' # cm') mm')
      (SU: A.call (A.wlift lwa) sa ga ga' krel et' em'),
    False.
  Proof.
    clear; i.
    invc RET; rrdes RL; destruct cm; destruct e; s in Ec; vauto; noemp; [].
    invc ST; [|by exfalso; eapply STK; econs; eauto]; invc STEP. 
    rewrite GET in GETk; clarify.
    invc SU; [invc APPb|invc INSTb]; destruct cm'; rrdes RL0; destruct e0; vauto.
  Qed.










  (** ** Composition and decomposition. *)


  Lemma good_thunk_comp: forall sa sb ty vt vm vs
      (GTa: A.good_thunk (A.wlift lwa) sa vt vm)
      (GTb: B.good_thunk (B.wlift lwb) sb ty vm vs),
    good_thunk (wlift lw) (scomp sa sb) ty vt vs.
  Proof.
    i; invc GTa GTb; eauto.
  Qed.

  Definition esim_comp_CIH (r : (wlift lw).(wwfo) ->
      lang_spec.cont tgt_lsi ->
      lang_spec.cont src_lsi ->
      (state (wlift lw) -> vrelf) ->
      state (wlift lw) ->
      state (wlift lw) -> bool -> typ -> conf tgt_lsi -> conf src_lsi -> Prop)
 := forall (g : state (wlift lw) -> vrelf) (em : conf mid_lsi) (sa : state (A.wlift lwa))
          (sb : state (wlift lwb)) (sa' : state (A.wlift lwa))
          (sb' : state (wlift lwb)) sa'' sb'' (sync : bool) ia
          (kt : lang_spec.cont tgt_lsi) (et : conf tgt_lsi),
        A.esim (A.wlift lwa) ia kt badcontm (gka g sb') sa sa'' sync et em ->
        forall ib (ks : lang_spec.cont src_lsi)
          (ty : typ) (es : conf src_lsi),
        esim (wlift lwb) ib badcontm ks (gkb g sa') sb sb'' sync ty em es ->
        (if sync
         then core em
         else exists mm : mach mid_lsi, real mid_lsi em mm) ->
        ext (A.wlift lwa) sa' sa ->
        ext (wlift lwb) sb' sb ->
        (if sync then sa'' = sa' else ext (A.wlift lwa) sa'' sa') ->
        (if sync then sb'' = sb' else ext (B.wlift lwb) sb'' sb') ->
        r (existT _ ib ia) kt ks g (scomp sa sb) (scomp sa'' sb'') sync ty et es.



  Lemma vclos_comp_name: forall g sa sb tn vt vm vs
      (RVa: A.vclos (A.wlift lwa) sa (gka g sb sa) vt vm)
      (VR: gkb g sa sb (FTname tn) vm vs)
      (VQ: vqueryb (B.wlift lwb) sb valid_name vs)
      (GK: gknow (wlift lw) g),
    vclos (wlift lw) (scomp sa sb) (g (scomp sa sb)) (Tname tn) vt vs.
  Proof.
    i; destruct (LEM (supp lwb tn)).
    { econs; econs; eauto.
      destruct sa as [[[]]], sb as [[[]]]; destruct s1, s4.
      eapply (gk_name GK); eauto; []; econs; eauto.
      econs; econs; rr; eauto.
    }
    revert vs VR VQ; eapply (A.vclos_induction RVa).
    i; invc IH; try (by invc CA CB; eauto)
    ; (econs; econs; rr; eauto; []; apply gkmin_unfold in VR; des)
    ; try (by exfalso; inv VR CB)
    ; try (by exfalso; inv VR WF)
    ; try (by exfalso; clarify; eapply f; eapply supp_name; eauto).
    - exfalso; inv VR; rrdes VR0; vauto.
    - apply PA.gkmin_unfold in VR0; des.
      + inv VR VR0; apply mybad_inj in Evm; des; subst.
        eapply (gk_mon GK); eauto using scomp_ext.
      + exfalso; inv VR VR0 WF WFb.
    - exfalso; inv VR VR0 CB.
  Qed.



  Lemma lsim_comp: forall vm sa sb g ty vt vs
      (GK: gknow (wlift lw) g)
      (SIMa: A.lsim _ (gka g sb) sa vt vm)
      (SIMb: B.lsim _ (gkb g sa) sb ty vm vs)
      r (CIH: esim_comp_CIH r),
    lsim_ (gres (pesim r \e/ r)) g (scomp sa sb) ty vt vs.
  Proof.
    clear; i; invc SIMa SIMb.
    econs; eauto using good_thunk_comp; i; apply gres_incl_id.

    destruct (LEM (gknow_ _ (gres (pesim r \e/ r)) g')) as [GK_|GK_]; last first.
    { left; apply pesim_gknow_; i; exfalso; eauto. }

    destruct (LEM (gknow _ g')) as [GK'|GK']; last first.
    { left; apply pesim_gknow; i; exfalso; eauto. }

    destruct (LEM (exists sa' sb', s' = scomp sa' sb')); last first.
    { left; eapply pesim_acquire; i.
      exfalso; rrdes HR; invc WHY; eauto.
    }
    des; subst.
    hexploit sdecomp_ext; eauto; i; des.
    destruct sa as [[[]]], sb as [[[]]], sa' as [[[]]], sb' as [[[]]].
    destruct s0, s3, s8, s11.
    simpl sdecomp in EXTa, EXTb; clean.

    right; invc STK; invc RK RF VR.
    - hexploit vclos_decomp; eauto; []; i; des.
      inv WF0 WFa.
      instantiate (1:=existT _ _ _).
      eapply CIH; try (by s; eauto); eauto.
      + eapply (CS ka badcontm); eauto using gka_mon.
        eapply A.call_app; eauto.
        * eapply scomp_cquerya; eauto.
        * econs.
        * eapply A.vclf_fix; try by s; eauto.
      + eapply (CS0 badcontm kb); eauto using gkb_mon.
        eapply B.call_app; eauto.
        * econs.
        * eapply scomp_cqueryb; eauto.
        * eapply B.vclf_fix; try by s; eauto.
    - inv WF0 WFa.
      eapply CIH; try (by s; eauto); eauto.
      + eapply (CS ka badcontm); eauto using gka_mon.
        eapply A.call_inst; eauto.
        * eapply scomp_cquerya; eauto.
        * econs.
        * eapply A.vclf_gen; try by s; eauto.
      + eapply (CS0 badcontm kb); eauto using gkb_mon.
        eapply B.call_inst; eauto.
        * econs.
        * eapply scomp_cqueryb; eauto.
        * eapply B.vclf_gen; try by s; eauto.
  Qed.



  
  Lemma gkcomp: forall r g sa sb ty vt vm vs
      (VRa: gka g sb sa vt vm)
      (VRb: gkb g sa sb ty vm vs)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (GK: gknow (wlift lw) g)
      (CIH: esim_comp_CIH r)
      (N: forall tn, ty <> FTname tn),
    g (scomp sa sb) ty vt vs.
  Proof.
    clear; i.
    apply PA.gkmin_unfold in VRa.
    apply PB.gkmin_unfold in VRb.
    des; try (by exfalso; eapply N; eauto).
    - (*| gka_enc gkb_enc *)
      invc VRa VRb; apply mybad_inj in Evm; des; subst.
      eapply (gk_mon GK); eauto using scomp_ext.
    - (*| lsim gkb_enc *)
      exfalso; inv VRb VRa WF WFb
      ; unfold mybad in *; s in H; clarify
      ; destruct (flext_of ty) in H; vauto.
    - (*| gka_enc lsim *)
      exfalso; inv VRa VRb WF WFa
      ; unfold mybad in *; s in H; clarify
      ; destruct (flext_of ty0) in H; vauto.
    - (*| lsim lsim *)
      apply (gk_lsim GK_).
      eapply lsim_comp; eauto.
  Qed.


  Lemma vclos_thunk_comp: forall g ga gb sa sb ty vt vm vs
      (VRa: A.vclos_thunk (A.wlift lwa) sa (ga sb sa) vt vm)
      (VRb: B.vclos_thunk (B.wlift lwb) sb (gb sa sb) ty vm vs)
      (COMP: forall (_: ga sb sa vt vm) (_: gb sa sb ty vm vs), g (scomp sa sb) ty vt vs : Prop), 
    vclos_thunk (wlift lw) (scomp sa sb) (g (scomp sa sb)) ty vt vs.
  Proof.
    clear; i; invc VRa VRb; try by inv CB CA0.
    - econs; last by eauto.
      + apply scomp_vquerya; eauto.
      + apply scomp_vqueryb; eauto.
    - econs; last by eauto.
      + apply scomp_vquerya; eauto.
      + apply scomp_vqueryb; eauto.
  Qed.


  Lemma vclos_comp: forall r g sa sb ty vt vm vs
      (RVa: A.vclos (A.wlift lwa) sa (gka g sb sa) vt vm)
      (RVb: B.vclos (B.wlift lwb) sb (gkb g sa sb) ty vm vs)
      (GK: gknow (wlift lw) g)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (CIH: esim_comp_CIH r),
    vclos (wlift lw) (scomp sa sb) (g (scomp sa sb)) ty vt vs.
  Proof.
    clear; i; generalize dependent vs; revert ty.
    eapply (A.vclos_induction_strong RVa); clear RVa.
    i; invc IH.

    - revert CA CB. pattern vb.
      eapply (B.vclos_induction RVb); clear RVb; i.
      invc CA CB.
      invc IH; vauto; try (by invc CA).
      + invc CA; econs.
        econs; eauto; last (by eapply VR; s; eauto); rr; eauto.
      + invc CA; econs.
        econs; eauto; last (by eapply VR; s; eauto); rr; eauto.
      + rrdes VR; vauto.
      + eapply vclos_comp_name; eauto.
        econs; econs; rr; eauto.
      + inv VR CA.

    - revert CA CB. pattern vb.
      eapply (B.vclos_induction RVb); clear RVb; i.
      invc CA CB.
      invc IH; vauto; try (by invc CA).
      + invc CA; econs.
        rrdes CB; econs; last first; rr; eauto.
      + invc CA; econs.
        econs; eauto; last (by eapply VR; s; eauto); rr; eauto.
      + invc CA; econs.
        econs; eauto; last (by eapply VR; s; eauto); rr; eauto.
      + rrdes VR; vauto.
      + eapply vclos_comp_name; eauto.
        econs; eapply A.vclgen_nat; last first; s; eauto.
      + inv VR CA.

    - des.
      revert CA CB. pattern vb.
      eapply (B.vclos_induction_strong RVb); clear RVb; i.
      invc CB.
      invc IH; vauto; try (by invc CA0); des.
      + invc CA0; econs; econs; eauto.
        * apply scomp_vquerya; eauto.
        * s; eauto.
      + invc CA0; econs; econs; last first; eauto.
        * eapply VR; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + invc CA0; econs; econs; last first; eauto.
        * eapply VR; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + rrdes VR; vauto.
      + eapply vclos_comp_name; eauto.
        econs; eapply A.vclgen_pair; eauto; s; eauto.
      + inv VR CA0.

    - des.
      revert CA CB. pattern vb.
      eapply (B.vclos_induction_strong RVb); clear RVb; i. 
      invc CB.
      invc IH; vauto; try (by invc CA0); des.
      + invc CA0; econs; econs; eauto.
        * apply scomp_vquerya; eauto.
        * s; eauto.
      + invc CA0; econs; econs; last first; eauto.
        * eapply VR; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + invc CA0; econs; econs; last first; eauto.
        * eapply VR; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + rrdes VR; vauto.
      + eapply vclos_comp_name; eauto.
        econs; eapply A.vclgen_inl; eauto; s; eauto.
      + inv VR CA0.

    - des.
      revert CA CB. pattern vb.
      eapply (B.vclos_induction_strong RVb); clear RVb; i. 
      invc CB.
      invc IH; vauto; try (by invc CA0); des.
      + invc CA0; econs; eapply B.vclgen_inr; eauto. 
        * apply scomp_vquerya; eauto.
        * s; eauto.
      + invc CA0; econs; econs; last first; eauto.
        * eapply VR; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + invc CA0; econs; econs; last first; eauto.
        * eapply VR; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + rrdes VR; vauto.
      + eapply vclos_comp_name; eauto.
        econs; eapply A.vclgen_inr; eauto; s; eauto.
      + inv VR CA0.

    - des; invc CA CB; eauto.
    - des; invc CA CB; eauto.

    - revert VR; pattern vb.
      eapply (B.vclos_induction RVb); clear RVb; i.
      dup VR; rrdes VR; subst.
      invc IH; vauto; try (by invc CA); des.
      + invc CA; econs; econs; last first; eauto.
        * s; eauto.
      + invc CA; econs; econs; last first; eauto.
        * s; eauto.
      + econs; econs; eapply rquery_comp; ss; eauto.
      + eapply vclos_comp_name; eauto.
        econs; eapply A.vclgen_ref; eauto; s; eauto.
      + inv VR CA.

    - revert CB VR; pattern vb.
      eapply (B.vclos_induction RVb); clear RVb; i.
      invc CB.
      invc IH; vauto; try (by invc CA); des.
      + invc CA; econs; econs; last first; eauto.
        * eapply VR0; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + invc CA; econs; econs; last first; eauto.
        * eapply VR0; s; eauto.
        * s; eauto.
        * apply scomp_vquerya; s; eauto.
      + rrdes VR0; vauto.
      + eapply vclos_comp_name; eauto.
        econs; eapply A.vclgen_name; eauto; s; eauto.
      + inv VR0 CA.

    - revert VR; pattern vb.
      eapply (B.vclos_induction_strong RVb); clear RVb; i.
      dup VR; invc VR CA CB.
      { invc IH; vauto; try (by invc CA); des.
        + invc CA; rrdes CB; econs; econs; last first; eauto.
          * rr; eauto.
          * apply scomp_vquerya; s; eauto.
        + invc CA; rrdes CB; econs; econs; last first; eauto.
          * rr; eauto.
        * apply scomp_vquerya; s; eauto.
        + rrdes VR; vauto.
        + eapply vclos_comp_name; eauto.
          econs; eapply A.vclgen_thunk; eauto; s; eauto.
        + econs; econs; eapply vclos_thunk_comp; eauto.
          i; eapply gkcomp; eauto; invc VR.
      }
      { invc IH; vauto; try (by invc CA); des.
        + invc CA; rrdes CB; econs; econs; last first; eauto.
          * rr; eauto.
          * apply scomp_vquerya; s; eauto.
        + invc CA; rrdes CB; econs; econs; last first; eauto.
          * rr; eauto.
          * apply scomp_vquerya; s; eauto.
        + rrdes VR; vauto.
        + eapply vclos_comp_name; eauto.
          econs; eapply A.vclgen_thunk; eauto; s; eauto.
        + econs; econs; eapply vclos_thunk_comp; eauto.
          i; eapply gkcomp; eauto; invc VR.
      }
  Qed.


  Lemma refst_minus_minusaD: forall sg sga lt lm
      (IN: GA.refst_minus (snd sga) (snd (minusa sga sg)) lt lm),
    exists ty ls, snd sg ty lt ls.
  Proof.
    i; rrdes IN.
    destruct (LEM (exists ty ls, snd sg ty lt ls)); [by des; eauto|].
    exfalso; apply IN0.
    destruct sga as [[]], sg as [[]]; rr; split; auto.
    i; rrdes FH; eauto.
  Qed.

  Lemma refst_minus_minusa: forall (sg: state ts_gw) (sga: state tm_gw) ty lt lm ls
      (IN: snd sg ty lt ls)
      (INa: snd sga lt lm),
    GA.refst_minus (snd sga) (snd (minusa sga sg)) lt lm.
  Proof.
    i; econs; rr; esplits; eauto.
    i; destruct sga as [[]], sg as [[]]; rrdes FH; eauto.
  Qed.

  Lemma refst_minus_minusbD: forall sg sgb ty lm ls
      (IN: GB.refst_minus (snd sgb) (snd (minusb sgb sg)) ty lm ls),
    exists lt, snd sg ty lt ls.
  Proof.
    i; rrdes IN.
    destruct (LEM (exists lt, snd sg ty lt ls)); [by des; eauto|].
    exfalso; apply IN0.
    destruct sgb as [[]], sg as [[]]; rr; split; auto.
    i; rrdes FH; eauto.
  Qed.

  Lemma refst_minus_minusb: forall (sg: state ts_gw) (sgb: state ms_gw) ty lt lm ls
      (IN: snd sg ty lt ls)
      (INb: snd sgb ty lm ls),
    GB.refst_minus (snd sgb) (snd (minusb sgb sg)) ty lm ls.
  Proof.
    i; econs; rr; esplits; eauto.
    i; destruct sgb as [[]], sg as [[]]; rrdes FH; eauto.
  Qed.


  Lemma refconfrel_comp: forall r g sga sla sgb slb crt crm crs
      (RCRa: tm_refconfrel' (GA.refst_minus (snd sga) (snd (minusa sga (fst (scomp (sga, sla) (sgb, slb))))))
                            (fun sga' => gka g (sgb, slb) (sga', sla)) sga crt crm)
      (RCRb: ms_refconfrel' (GB.refst_minus (snd sgb) (snd (minusb sgb (fst (scomp (sga, sla) (sgb, slb))))))
                            (fun sgb' => gkb g (sga, sla) (sgb', slb)) sgb crm crs)
      (GK: gknow (wlift lw) g)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (CIH: esim_comp_CIH r),
    ts_refconfrel (fun sg => g (sg, snd (scomp (sga, sla) (sgb, slb))))
                  (fst (scomp (sga, sla) (sgb, slb))) crt crs.
  Proof.
    i; rrdes RCRa RCRb; clarify; rr; esplits; eauto.
    - i; hexploit DOMa; eauto; []; i; des.
      eapply refst_minus_minusaD in X; []; des; eauto.
    - i; hexploit DOMb0; eauto; []; i; des.
      apply refst_minus_minusbD in X; []; des; eauto.
      
    - i; destruct sga as [[]], sgb as [[]]; rrdes REG.
      hexploit (RELV la lm).
      { rr; splits; auto; i; rrdes FH; eapply FH0; rr; eauto. } 
      hexploit (RELV0 ty lm lb).
      { rr; splits; auto; i; rrdes FH; eapply FH0; rr; eauto. }
      i; des.
      hexploit (pmread_uniq RDa0 RDb); []; i; inst; subst.
      esplits; eauto.
      eapply vclos_unlift with (gw:=ts_gw) (lw:=lw). 
      insts $ (snd (scomp (s, s0, r0, sla) (s1, s2, s3, slb))). 
      rewrite -prod_eta; eapply vclos_comp; eauto.
      + eapply PA.vclos_lift, A.vclos_mon; eauto.
      + eapply PB.vclos_lift, B.vclos_mon; eauto.
  Qed.




  Lemma minus_or_minusa: forall (sga: state tm_gw) sgb lt lm
      (IN: snd sga lt lm),
    snd (minusa sga (bullet sga sgb)) lt lm \/
    GA.refst_minus (snd sga) (snd (minusa sga (bullet sga sgb))) lt lm.
  Proof.
    i; destruct sga as [[]], sgb as [[]].
    destruct (LEM (exists ty ls, s4 ty lm ls)).
    - des; right; rr; split; auto.
      i; rrdes FH; eapply FH0; rr; esplits; eauto.
    - left; rr; split; auto.
      i; rrdes FH; apply f; clean.
      hexploit (GA.refst_bij FH IN); eauto; []; i; des.
      specialize (X eq_refl); des; subst; eauto.
  Qed.

  Lemma minus_or_minusb: forall (sgb: state ms_gw) sga ty lm ls
      (IN: snd sgb ty lm ls),
    snd (minusb sgb (bullet sga sgb)) ty lm ls \/
    GB.refst_minus (snd sgb) (snd (minusb sgb (bullet sga sgb))) ty lm ls.
  Proof.
    i; destruct sga as [[]], sgb as [[]].
    destruct (LEM (exists lt, s1 lt lm)).
    - des; right; rr; split; auto.
      i; rrdes FH; eapply FH0; rr; esplits; eauto.
    - left; rr; split; auto.
      i; rrdes FH; apply f; clean.
      hexploit (GB.refst_bij FH0 IN); eauto; []; i; des.
      specialize (X0 eq_refl); des; subst; eauto.
  Qed.

  Lemma confrel_comp: forall r g sa sb ct cm cs
      (CRa: A.confrel (A.wlift lwa) (gka g sb) sa ct cm)
      (CRb: B.confrel (B.wlift lwb) (gkb g sa) sb cm cs)
      (RL: forall em (CO: core em), exists mm, real _ (em # cm) mm)
      (GK: gknow (wlift lw) g)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (CIH: esim_comp_CIH r),
    confrel (wlift lw) g (scomp sa sb) ct cs.
  Proof.
    clear; i; rrdes CRa CRb WHX WHX0.
    rrdes WHX1 WHX CPb CPa0; subst; noemp.
    rename hay into clt, hby into clma, hay0 into clmb, hby0 into cls.
    rename hay1 into crt, hby1 into crma, hay2 into crmb, hby2 into crs.
    rename hax1 into cct, hbx2 into ccs.
    destruct sa, sb; clean.

    dup WHY1; eapply tm_refconfrel'_ext in WHY1
    ; [|by apply (minusa_ext s s1)]; rrdes WHY1; subst.
    rename hax into crt, hbx into crma, hay into crt', hby into crma'.
    dup WHY2; eapply ms_refconfrel'_ext in WHY2
    ; [|by apply (minusb_ext s1 s)]; rrdes WHY2; subst.
    rename hax into crmb, hbx into crs, hay into crmb', hby into crs'.

    hexploit (RL ((pmemp, Just (nil, IEnat 42)))); [by s; eauto|]; i; des.

    assert (crma' = crmb').
    { clear - WHY1 WHY2 HPB X.
      rrdes WHY1 WHY2; subst; f_equal; apply pm_extensionality; eauto; [].
      i; split; intro RD.
      - dup RD; apply DOMb in RD; des.
        hexploit refst_minus_minusaD; eauto; []; i; des.
        hexploit bullet_ref_decomp; [exact X0|]; i; des; clean.
        rrdes RD; hexploit (GA.refst_bij RD REGa); eauto; []; i; des.
        specialize (X1 eq_refl); des; subst.
        hexploit refst_minus_minusb; eauto; []; i; des.
        hexploit (RELV0 ty x ls); eauto; []; i; des.
        assert (va = y); last by subst; auto.
        clear - RDa RD0 HPB X; rrdes X; subst.
        destruct crmb, clmb, crma, clma; s in Ec; clarify; noemp.
        rewrite -!pmplus_assoc (pmplus_swap h2) (pmplus_swap h0) in PV, H1.
        eapply pmread_plus in RDa; eauto; [].
        rewrite -H1 in PV; eapply pmread_plus in RD0; eauto; [].
        apply (pmread_uniq RDa RD0).
      - dup RD; apply DOMa0 in RD; des.
        hexploit refst_minus_minusbD; eauto; []; i; des.
        hexploit bullet_ref_decomp; [exact X0|]; i; des; clean.
        rrdes RD; hexploit (GB.refst_bij RD REGb); eauto; []; i; des.
        specialize (X2 eq_refl); des; subst.
        hexploit refst_minus_minusa; eauto; []; i; des.
        hexploit (RELV lt x); eauto; []; i; des.
        assert (vb = y); last by subst; auto.
        clear - RDb RD0 HPB X; rrdes X; subst.
        destruct crmb, clmb, crma, clma; s in Ec; clarify; noemp.
        rewrite -!pmplus_assoc (pmplus_swap h2) (pmplus_swap h0) in PV, H1.
        eapply pmread_plus in RDb; eauto; [].
        rewrite -H1 in PV; eapply pmread_plus in RD0; eauto; [].
        eapply (pmread_uniq RDb RD0).
    }
    subst; rename crmb' into crm'.

    rr; exists (cct # crt'), (crt # clt), (ccs # crs'), (crs # cls).
    rewrite -!cplus_assoc (cplus_swap _ crt) (cplus_swap _ crs).
    splits; auto; last first.
    - dup WHY; apply A.lwconf_extra in WHY.
      dup WHY0; apply B.lwconf_extra in WHY0.
      des; invc WHY WHY0 WHY7 WHY8.
      econs; eauto.
      + rewrite -!cplus_assoc -!(cplus_swap _ crm') in HPB.
        rrdes WHX WHY1 WHX0 WHY2; subst; s in HPB; clarify.
        clean; s; f_equal; eapply pmplus_inj; eauto.
        rrdes X; s in Ec; subst; clarify; noemp.
        rewrite -pmplus_assoc pmplus_swap -H2 in PV; auto.
      + i; rrdes CO; subst; s; noemp.
        rrdes WHX0 WHX WHY1; subst.
        rrdes X; s in Ec; subst; clarify; noemp.
        rewrite -H2 -pmplus_assoc in PV; dup PV; split_pmvalid PV.
        econs; rr; esplits; eauto.
        * rrdes WHY3.
          apply pmvalid_plus_read; auto; []; i.
          apply DOM in RDa; des; apply RELV2 in RDa; des. s in Ea; s in Eb. clarify.
          clear - PV0 RDa0 RDb.
          rewrite pmplus_assoc in PV0; eapply (pmvalidD_plus_read PV0); eauto.
        * rewrite -H2 -pmplus_assoc in FIN; split_pmfinite FIN PV0.
          apply pmfinite_plus; auto.
          hexploit (GA.refst_fin (snd s)); []; i; des.
          exists (map snd l); i. apply DOM in RD; des.
          apply FIN2 in RD; apply in_map_snd; eauto. 
    - rr; esplits; eauto.
      + rr; splits; rewrite -?bullet_tgt -?bullet_src; eauto.
      + eapply refconfrel_comp; eauto.
  Qed.

  Lemma configure_true_comp: forall r g
      sa sb sa' sb' et em es cst csma csmb css cft cfs mt mm ms
      (CNFa: A.configure (A.wlift lwa) (gka g sb) sa' true et em
                         cst csma cft cemp mt mm)
      (CNFb: B.configure (B.wlift lwb) (gkb g sa) sb' true em es
                         csmb css cemp cfs mm ms)
      (EXTa: ext (A.wlift lwa) sa' sa)
      (EXTb: ext (B.wlift lwb) sb' sb)
      (GK: gknow (wlift lw) g)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (CIH: esim_comp_CIH r),
    configure (wlift lw) g (scomp sa' sb') true et es cst css cft cfs mt ms.
  Proof.
    i; invc CNFa CNFb; econs; eauto; des.
    assert (csma = csmb).
    { clear -MB MA0; noemp; rrdes MB MA0.
      destruct em, csma, csmb; ss; clarify.
      rewrite -H0 in H1.
      apply eq_sym, pmplus_inj in H2; subst; auto.
      destruct e, e0, e1; noemp; vauto; clarify.
    }
    subst; eapply confrel_comp; eauto.
    - eapply A.confrel_mon; eauto using gka_mon.
    - eapply B.confrel_mon; eauto using gkb_mon.
    - clear -MB CO0; noemp.
      i; rrdes MB; destruct em, csmb; s in Ec; clarify.
      rrdes CO0 CO; clarify.
      destruct e0; s; noemp; vauto.
      econs; rr; esplits; eauto.
    - eauto.
  Qed.




  Lemma esim_comp_err: forall g sa sb sb' ib ks ty em es mm ms mm' cfs csm css
      (GK: gknow (wlift lw) g)
      (SIMb: esim_ext (B.wlift lwb) (gkb_enc g sa) ib badcontm ks
                       (gkb g sa) sb sb' false ty em es)
      (XTRs: extra cfs)
      (CNFb: B.configure (B.wlift lwb) (gkb g sa) sb' false em es csm css cemp cfs mm ms)
      (EXT: ext _ sb' sb)
      (ST: rtc (step mid_lsi etau) mm mm')
      (ERR: error mid_lsi mm'),
    exists ms',
      rtc (step src_lsi etau) ms ms' /\ error _ ms'.
  Proof.
    clear; i.
    apply clos_rt_rt1n_iff in ST.
    revert SIMb; depgen em; depgen csm; revert es ms css.
    depgen sb'; generalize ib at 1.
    induction ST.
    - (*| no steps *)
      i; invc CNFb; noemp.
      exfalso; eapply ERR; eauto.
    - (*| some steps *)
      apply clos_rt_rt1n_iff in ST.
      i; punfold SIMb; invc SIMb.
      dup CNFb; apply ESIM in CNFb; eauto using gkb_gknow, extra_emp; []; des.
      clear ESIM; invc CNFb.
      + (*| err *)
        eauto.
      + (*| step *)
        specialize (EST _ _ H); des.
        * (*| progress *)
          rewrite clear_bote in ES; inv ES.
          hexploit IHST; [|eapply ext_trans| | |]; eauto; []; intro X; desH X.
          exists ms'; split; eauto.
          do 2 (eapply rt_trans; eauto); des; subst.
          { eapply rt_trans; [eapply rt_step|]; eauto. }
          { eapply rt_trans; eauto. }
        * (*| call *)
          exfalso.
          invc CNF ES SU RF; noemp; exfalso; invc VR
          ; [eapply bad_fun_not_error; eauto; eapply mybad_bad_fun; s; auto
            |eapply bad_gen_not_error; eauto; eapply mybad_bad_gen; s; auto].
      + (*| ret *)
        exfalso.
        eapply rt_trans in ST; [|by eapply rt_step; eauto].
        inv CNF; eauto using badcontm_ret_error.
  Qed.

  Lemma esim_comp_ret: forall
      r (ia: lwa) ib kt ks vt vm g sa sa' sa'' sa''' sb sb' sb''
      ty et em em' es mt mm mm' ms cst csm csm' css cft cfs
      (GK: gknow (wlift lw) g)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (SIMb: esim_ext (B.wlift lwb) (gkb_enc g sa') ib badcontm ks
                       (gkb g sa') sb sb'' false ty em es)
      (XTRs: extra cfs)
      (CNFb: B.configure (B.wlift lwb) (gkb g sa') sb'' false em es csm css cemp cfs mm ms)
      (ST: rtc (step mid_lsi etau) mm mm')
      (RETt: A.cquerya (A.wlift lwa) sa''' (ret vt kt) et)
      (RETm: A.cqueryb (A.wlift lwa) sa''' (ret vm badcontm) em')
      (RV: A.vclos (A.wlift lwa) sa''' (gka g sb' sa''') vt vm)
      (CNFa: A.configure (A.wlift lwa) (gka g sb') sa''' true
                         et em' cst csm' cft cemp mt mm')
      (EXTa: ext (A.wlift lwa) sa' sa)
      (EXTb: ext (B.wlift lwb) sb' sb)
      (EXTa': ext (A.wlift lwa) sa'' sa')
      (EXTb': ext (B.wlift lwb) sb'' sb')
      (EXTa'': ext (A.wlift lwa) sa''' sa'')
      (EXTPa: extpub (A.wlift lwa) sa''' sa)
      (CIH: esim_comp_CIH r),
    exists ms',
       rtc (step src_lsi etau) ms ms' /\
       esim_main (esim_progress (gres (pesim r \e/ r)) (existT _ ib ia) kt ks g (scomp sa sb) \8/
                  esim_call (gres (pesim r \e/ r)) kt ks g (scomp sa sb))
                 kt ks g (scomp sa sb) (scomp sa'' sb'') ty cft mt cfs ms'.
  Proof.
    clear; i.
    apply clos_rt_rt1n_iff in ST.
    revert SIMb; depgen em; depgen csm; revert es ms css.
    generalize ib at 1.
    assert (EXTb'': ext _ sb'' sb'') by auto; revert EXTb''.
    generalize sb'' at 1 3 4 as sb'''.
    induction ST.
    - (*| no steps *)
      i; punfold SIMb; invc SIMb.
      dup CNFb; apply ESIM in CNFb; eauto using gkb_gknow, extra_emp; []; des.
      invc CNFb0; noemp; clear ESIM; invc CNFb.
      + (*| err *)
        eauto.
      + (*| step right *)
        invc RETm CNFa; noemp; rrdes MB0; destruct csm'.
        destruct e; s in Ec; clarify; noemp.
        hexploit EST.
        * inst_pairs; econs 1; eauto; []; i; eauto using IRret.
        * i; des.
          { (*| progress *)
            inv ES; desH ES0; [|by vauto].
            apply PB.esim_ext_impl_esim in ES0; [|by i; eapply PB.gkmin_fold; eauto].
            hexploit esim_halteda; eauto using gkb_gknow, extra_emp; rr; eauto.
            i; des; exists mb'0; split; eauto.
            do 2 (eapply rt_trans; eauto); des.
            - eapply rt_trans; [eapply rt_step|]; eauto.
            - subst; repeat (eapply rt_trans; eauto).
          }
          { (*| call *)
            exfalso.
            invc CNF; noemp; destruct csa'; rrdes MA1; clarify.
            assert (e = Nothing). 
            { destruct sync.
              - des; destruct CO2; subst; destruct e; vauto.
              - des; invc RH.
            }
            subst; destruct ea'; destruct e; s in Ec; clarify.
            invc ES SU; [inv APPa|inv INSTa]. 
          }
      + (*| ret *)
        assert (vm = va /\ em' = ea').
        { invc CNF CNFa; des; noemp.
          hexploit (core_uniq _ _ _ CO0 CO2); eauto; []; i; des; subst.
          inv RETm RETa.
        }
        des; subst.
        esplits; eauto.
        rename s' into sb''''.
        eapply B.esim_ret with (s' := scomp sa''' sb'''').
        * apply rt_refl.
        * apply scomp_cquerya; eauto.
        * s; eauto.
        * eapply vclos_comp; eauto.
          { eapply A.vclos_mon; eauto using gka_mon. }
          { eapply B.vclos_mon; eauto using gkb_mon. }
        * eapply configure_true_comp; eauto using ext_trans.
        * eapply scomp_ext; eauto using ext_trans.
        * eapply scomp_extpub; eauto.
    - (*| some steps *)
      apply clos_rt_rt1n_iff in ST.
      i; punfold SIMb; invc SIMb.
      dup CNFb; apply ESIM in CNFb; eauto using gkb_gknow, extra_emp; []; des.
      clear ESIM; invc CNFb.
      + (*| err *)
        eauto.
      + (*| step *)
        specialize (EST _ _ H); des; rename s' into sb''''.
        * (*| progress *)
          inv ES; desH ES0; [|by vauto].
          hexploit IHST; [|eapply ext_trans| | |]; eauto; []; i; nbdes.
          exists ms'; split; eauto.
          do 2 (eapply rt_trans; eauto); des; subst.
          { eapply rt_trans; [eapply rt_step|]; eauto. }
          { eapply rt_trans; eauto. }
        * (*| call *)
          exfalso; inv CNF CNFa; invc ES SU RF VR; noemp
          ; [eapply bad_ret_app; eauto; eapply mybad_bad_fun; s; auto
            |eapply bad_ret_inst; eauto; eapply mybad_bad_gen; s; auto].
      + (*| ret *)
        exfalso.
        inv CNF RETa; rrdes MA; destruct csa'; s in Ec; destruct e; vauto.
        invc H; [|by exfalso; eapply STK; econs; eauto].
        invc STEP; rewrite GET in GETk; clarify.
        hexploit halted_steps; first last; first 1 last; first (eapply rtc_mon)
        ; eauto; rr; eauto; []; i; inst; subst.
        invc CNFa; eapply ret_halted; eauto; rr; eauto.
  Qed.

  Lemma ksim_comp: forall ia ib g sa sa' sb sb' tyf ty kit kis kt km ks
      (KSa: A.ksim (A.wlift lwa) ia kit badcontm (gka g sb')
                   sa sa' sa' kt km)
      (KSb: B.ksim (B.wlift lwb) ib badcontm kis (gkb g sa')
                   sb sb' sb' tyf ty km ks)
      (EXTa: ext _ sa' sa)
      (EXTb: ext _ sb' sb)
      r (CIH: esim_comp_CIH r),
    ksim_ (gres (pesim r \e/ r)) (existT _ ib ia) kit kis g
          (scomp sa sb) (scomp sa' sb') (scomp sa' sb') tyf ty kt ks.
  Proof.
    clear; i.
    invc KSa KSb; econs; i; apply gres_incl_id.

    destruct (LEM (gknow_ _ (gres (pesim r \e/ r)) g')) as [GK_|GK_]; last first.
    { left; apply pesim_gknow_; i; exfalso; eauto. }

    destruct (LEM (gknow _ g')) as [GK|GK]; last first.
    { left; apply pesim_gknow; i; exfalso; eauto. }

    destruct (LEM (exists sa'' sb'', s' = scomp sa'' sb'')); last first.
    { left; eapply pesim_acquire; i.
      exfalso; rrdes HR; invc WHY; eauto.
    }
    des; subst.
    hexploit sdecomp_ext; eauto; []; i; des.
    hexploit sdecomp_extpub; eauto; []; i; des.
    simpl sdecomp in EXTa, EXTb; clean.

    hexploit vclos_decomp; eauto; []; i; des.

    set (em := (pmemp, Just ([(inr 0, km); (inl (inr 1), vm)],
                             IEret 0 (inr 1))) : conf mid_lsi). 

    destruct sa' as [[[]]], sb' as [[[]]], sa'' as [[[]]], sb'' as [[[]]].
    destruct s0, s3, s8, s11.

    hexploit KR; first M; Mskip eauto.
    { eauto using gka_mon. }
    { eapply scomp_cquerya; eauto. }
    { instantiate (1 := em); unfold em; s; eauto. }
    intro SIMa.
    
    hexploit KR0; first M; Mskip eauto.
    { eauto using gkb_mon. }
    { instantiate (1 := em); unfold em; s; eauto. }
    { eapply scomp_cqueryb; eauto. }
    intro SIMb.

    right; eapply CIH; eauto using ext_trans.
    unfold em; s; eauto.
  Qed.


  Lemma call_comp: forall r g sa sb sa' sb' ia ib kt ks ty et em es
      (CAa: A.call (A.wlift lwa) sa' (gka_enc g sb' sa') (gka g sb' sa')
                   (A.ksim (A.wlift lwa) ia kt badcontm (gka g sb') sa sa' sa')
                   et em)
      (CAb: B.call (B.wlift lwb) sb' (gkb_enc g sa' sb') (gkb g sa' sb')
                   (B.ksim (wlift lwb) ib badcontm ks (gkb g sa') sb sb' sb')
                   ty em es)
      (EXTa: ext _ sa' sa)
      (EXTb: ext _ sb' sb)
      (GK: gknow (wlift lw) g)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (CIH: esim_comp_CIH r),
    call (wlift lw) (scomp sa' sb') (g (scomp sa' sb')) (g (scomp sa' sb'))
         (ksim_ (gres (pesim r \e/ r)) (existT _ ib ia) kt ks g (scomp sa sb) (scomp sa' sb') (scomp sa' sb'))
         ty et es.
    Proof.
      i; invc CAa CAb.
      - inv APPb APPa0; clean.
        eapply call_app.
        + eapply scomp_cquerya; eauto.
        + s; eauto.
        + eapply vclos_thunk_comp; eauto.
          i; eapply gkcomp; vauto.
        + eapply vclos_comp; eauto.
        + eapply ksim_comp; eauto.
      - exfalso; invc APPb INSTa; invc RF0 VR.
      - exfalso; invc INSTb APPa; invc RF0 VR RF VR
        ; inv Evm; destruct ty0; destruct flext_of0; vauto.
      - inv INSTb INSTa0; clean.
        eapply call_inst.
        + eapply scomp_cquerya; eauto.
        + s; eauto.
        + eapply vclos_thunk_comp; eauto.
          i; eapply gkcomp; vauto.
        + eauto.
        + eapply ksim_comp; eauto.
    Qed.


  Lemma configure_false_comp: forall
      g (GK: gknow (wlift lw) g)
      sa sb sa' sb' et em es cst csma csmb css cft cfs mt mm ms
      (CNFa: A.configure (A.wlift lwa) (gka g sb) sa' false et em
                         cst csma cft cemp mt mm)
      (CNFb: B.configure (B.wlift lwb) (gkb g sa) sb' false em es
                         csmb css cemp cfs mm ms),
    configure (wlift lw) g (scomp sa' sb') false et es cst css cft cfs mt ms.
  Proof.
    i; invc CNFa CNFb; econs; eauto; inst; des; eauto.
  Qed.


  Lemma configure_false_decomp: forall g sa sb et em es cst css cft cfs mt mm ms
      (GK: gknow _ g)
      (CNF: configure (wlift lw) g (scomp sa sb) false et es cst css cft cfs mt ms)
      (RL: real _ em mm),
    <<CNFa: A.configure (A.wlift lwa) (gka g sb)
              sa false et em cst cemp cft cemp mt mm>> /\
    <<CNFb: B.configure (B.wlift lwb) (gkb g sa)
              sb false em es cemp css cemp cfs mm ms>>.
  Proof.
    clear; i; invc CNF; des.
    esplits; eauto; econs; noemp; eauto.
  Qed.


  Lemma bla: forall La Lb (w: world La Lb)
      ms mbt mb' mbt0 es i i' kt ks g (s: state w) s' sync evt ty ea' eb'
      (ST: tc (step Lb etau) ms mbt)
      (ST': rtc (step Lb etau) mbt mb')
      (PR: esim_progress es i kt ks g s s' sync evt mb' mbt0 ty ea' eb'),
    exists mb'',
      <<ST: rtc (step Lb etau) ms mb''>> /\
      <<PR: esim_progress es i' kt ks g s s' sync evt mb'' mbt0 ty ea' eb'>>.
  Proof.
    i; invc PR; des.
    - exists mb'; esplits; eauto using tc_rtc, rt_trans.
    - subst; hexploit tc_last; [by clear LT; eapply tc_rtc_tc; eauto|].
      i; des; esplits; eauto.
  Qed.

  Lemma esim_comp_step: forall r g sa sb sa' sb' sa'' sb''
      ia ib kt ks ty em es mm ms cfs csm css cft mt
      (GK: gknow (wlift lw) g)
      (GK_: gknow_ (wlift lw) (gres (pesim r \e/ r)) g)
      (SIMb: esim_ext (wlift lwb) (gkb_enc g sa') ib badcontm ks
                       (gkb g sa') sb sb'' false ty em es)
      (CNFb: configure (wlift lwb) (gkb g sa') sb'' false em es csm css cemp cfs mm ms)
      (NH: ~ halted tgt_lsi mt)
      (EST: forall evt mt', step tgt_lsi evt mt mt' ->
            exists et' cst' em' csm' mm' mm'' sa''' sync,
              <<STP: rtc (step mid_lsi etau) mm mm'>> /\
              <<ES: A.esim_progress
                      (model_properties_untyped.esim_ext _ (gka_enc g sb')) ia kt badcontm
                      (gka g sb') sa sa''' sync evt mm' mm'' et' em' \/
                    model_properties_untyped.esim_call_ext kt badcontm 
                      (gka_enc g sb') (gka g sb') sa sa''' sync evt mm' mm'' et' em'>> /\
              <<CNF: A.configure (A.wlift lwa) (gka g sb') sa''' sync et' em'
                       cst' csm' cft cemp mt' mm''>> /\
              <<EXT: ext (A.wlift lwa) sa''' sa''>>)
      (EXTa: ext _ sa' sa)
      (EXTb: ext _ sb' sb)
      (EXTa': ext _ sa'' sa')
      (EXTb': ext _ sb'' sb')
      (XTRt: extra cft)
      (XTRs: extra cfs)
      (CIH: esim_comp_CIH r),
    exists ms',
       rtc (step src_lsi etau) ms ms' /\
       esim_main (esim_progress (gres (pesim r \e/ r)) (existT _ ib ia) kt ks g (scomp sa sb) \8/
                  esim_call (gres (pesim r \e/ r)) kt ks g (scomp sa sb))
                 kt ks g (scomp sa sb) (scomp sa'' sb'') ty cft mt cfs ms'.
  Proof.
    clear; i.

    destruct (LEM (exists ms', rtc (step _ etau) ms ms' /\ error _ ms')).
    { des; eauto. }
    esplits; eauto; []; eapply esim_step; [by eauto|]; i.
    specialize (EST _ _ ST); nbdes.

    apply clos_rt_rt1n_iff in STP.
    revert SIMb; depgen em; depgen csm; revert es ms css f.
    generalize ib.
    assert (EXTb'': ext _ sb'' sb'') by auto; revert EXTb''. 
    generalize sb'' at 1 3 4 as sb'''.
    induction STP.

    - (*| no steps *)
      i; dup SIMb; punfold SIMb; invc SIMb.
      dup CNFb; apply ESIM in CNFb; eauto using gkb_gknow, extra_emp; [].
      inv CNFb0; noemp; clear ESIM; invc CNFb.
      + (*| err (rhs) *)
        
        exfalso; apply f; eauto.
      + (*| step (rhs) *)
        rewrite clear_bote in EST.
        des; [invc ES; des|].
        * (*| physical progress (lhs) *)
          specialize (EST _ _ STEP); nbdes.
          rename s' into sb''''.
          inv CNF CNF0; des; subst; noemp.
          { (*| progress (rhs) *)
            hexploit (mid_real_inj MA1 MB0); []; i; inst; subst. 
            invc ES; esplits; first 1 last.
            - left; econs; last by eauto.
              { des; subst; [by left; eauto|right].
                insts $ (existT _ i'0 i').
                s; splits; eauto using lexprod_tc1.
              }
              eapply PA.esim_ext_impl_esim in ES0; last first.
              { i; eapply PA.gkmin_fold; eauto. }
              eapply PB.esim_ext_impl_esim in ES1; last first.
              { i; eapply PB.gkmin_fold; eauto. }
              insts $ eb' et' (scomp sa''' sb'''').
              nbdes; subst; noemp.     
              apply gres_incl_id; right.
              eapply CIH; eauto using ext_trans.
            - des; subst; noemp; eapply configure_false_comp; eauto.
            - eauto using scomp_ext, ext_trans.
            - eauto.
          }
          { (*| call (rhs) *)
            exfalso.
            invc ES SU.
            - invc RF VR APPa CNF0.
              rrdes MA2 MB0; destruct csa'; destruct e; vauto; [].
              s in Ec; clarify; noemp.
              eapply PA.esim_ext_impl_esim in ES0; last first.
              { i; eapply PA.gkmin_fold; eauto. }
              eapply esim_haltedb in ES0; eauto using gka_gknow, extra_emp.
              des.
              { invc CNF; eapply ERR; eauto. }
              { invc RET0; rrdes RET; destruct csb'0; destruct e; s in Ec; clarify. } 
            - invc RF VR INSTa CNF0.
              rrdes MA2 MB0; destruct csa'; destruct e; vauto; [].
              s in Ec; clarify; noemp.
              eapply PA.esim_ext_impl_esim in ES0; last first.
              { i; eapply PA.gkmin_fold; eauto. }
              eapply esim_haltedb in ES0; eauto using gka_gknow, extra_emp.
              des.
              { invc CNF; eapply ERR; eauto. }
              { invc RET0; rrdes RET; destruct csb'0; destruct e; s in Ec; clarify. }
          }
        * (*| logical progress (lhs) *)
          subst; inv CNF CNFb0; des; subst; noemp.
          hexploit (mid_real_inj MA1 MB0); []; i; inst; subst.
          esplits; first 1 last.
          { left; econs; first (insts $ (existT _ ib0 i'))
            ; [right; split; auto| |by eauto].
            - s; splits; eauto using lexprod_tc2.
            - eapply PA.esim_ext_impl_esim in ES0; last first.
              { i; eapply PA.gkmin_fold; eauto. }
              eapply PB.esim_ext_impl_esim in SIMb0; last first.
              { i; eapply PB.gkmin_fold; eauto. }
              insts $ es et' (scomp sa''' sb''').
              apply gres_incl_id; right.
              eapply CIH; eauto using ext_trans.
          }
          { eapply configure_false_comp; eauto. }
          { eauto using scomp_ext. }
          { eauto. }
        * (*| call (lhs) *)
          subst; invc ES; specialize (EST _ _ ST0); des; rename s' into sb''''.
          { (*| progress (rhs) *)
            invc ES.
            eapply PB.esim_ext_impl_esim in ES0; last first.
            { i; eapply PB.gkmin_fold; eauto. }
            esplits.
            - eauto.
            - left; econs; auto.
              + instantiate (1:= existT _ _ _).
                des; [left|right]; s; eauto using lexprod_tc1.
              + eapply gres_incl_id; left; pfold; econs; i.
                eapply configure_false_decomp with (mm := mm'') in CNF1
                ; eauto; last first.
                { invc CNF0; des; subst; noemp; eauto. }
                nbdes.
                hexploit esim_halteda; first (by eauto)
                ; first (by invc CNFb; econs; eauto); eauto using gkb_gknow, extra_emp.
                { clear - CNF SU.
                  invc SU CNF; [invc APPb|invc INSTb]; rrdes MB
                  ; destruct csm'; destruct e0; vauto; s in Ec; clarify; noemp
                  ; inv RF CB VR; destruct ty; destruct flext_of0; vauto; inv Evm. 
                }
                i; des; eapply esim_err; eauto.
            - econs; eauto.
              + invc CNF; rewrite cplus_assoc in MA0; noemp; eauto.
              + invc CNF0; des; subst; noemp; eauto.
            - eauto.
          }
          { (*| call (rhs) *)
            invc ES.
            assert (em' = ea').
            { invc CNF CNF0; des; noemp.
              hexploit (core_uniq _ _ _ CO3 CO1); eauto; []; i; des; subst; auto.
            }
            subst; esplits; first by eauto.
            - right; econs; eauto.
              instantiate (5:= existT _ _ _).
              eapply call_comp
              ; [eapply A.call_mon; first by eauto
                |eapply B.call_mon; first by eauto| | | | |]
              ; eauto using ext_trans.
              + i; eapply gka_enc_mon; last first; eauto using ext_trans.
              + i; eapply gka_mon; eauto using ext_trans.
              + i; eapply A.ksim_mon; eauto using gka_mon.
              + i; eapply gkb_enc_mon; last first; eauto using ext_trans.
              + i; eapply gkb_mon; eauto using ext_trans.
              + i; eapply B.ksim_mon; eauto using gkb_mon.
            - eapply configure_true_comp; eauto using ext_trans.
            - eauto using scomp_ext, ext_trans.
          }
      + (*| ret (rhs) *)
        inv RETa CNF0.
        rrnbdes MA0; destruct csa'; destruct e; vauto; [].
        s in Ec; noemp; clarify; des; first (invc ES; des).
        * (*| physical progress (lhs) *)
          exfalso.
          invc STEP; [|by exfalso; eapply STK; econs; eauto].
          invc STEP0; rewrite GET in GETk; clarify. 
          eapply PA.esim_ext_impl_esim in ES0; last first. 
          { i; eapply PA.gkmin_fold; eauto. }
          hexploit esim_haltedb; eauto using gka_gknow, extra_emp; rr; eauto; [].
          i; des.
          { invc CNF; eapply ERR; eauto. }
          { invc RET0; rrdes RET; destruct csb'0; destruct e; s in Ec; clarify. }
        * (*| logical progress (lhs) *)
          subst.
          eapply PA.esim_ext_impl_esim in ES0; last first.
          { i; eapply PA.gkmin_fold; eauto. }
          eapply PB.esim_ext_impl_esim in SIMb0; last first.
          { i; eapply PB.gkmin_fold; eauto. }
          assert (em = em').
          { invc CNF; des; subst; noemp.
            eapply mid_real_inj; eauto.
          }
          subst; esplits; first 1 last.
          { left; econs; last by eauto.
            - insts $ (existT _ ib0 i').
              right; splits; s; eauto using lexprod_tc2.
            - insts $ es et' (scomp sa''' sb''').
              apply gres_incl_id; right.
              eapply CIH; eauto using ext_trans.
          }
          { eapply configure_false_comp; eauto. }
          { eapply scomp_ext; eauto. }
          { eauto. }
        * (*| call (lhs) *)
          exfalso; invc ES CNF0 CNF; noemp.
          hexploit bad_ret_call; eauto.
    - (*| some steps *)
      apply clos_rt_rt1n_iff in STP.
      i; punfold SIMb; invc SIMb.
      dup CNFb; apply ESIM in CNFb; eauto using gkb_gknow, extra_emp; [].
      clear ESIM; invc CNFb.
      + (*| err *)
        
        exfalso; apply f; eauto.
      + (*| step *)
        specialize (EST _ _ H); desH EST; rename s' into sb''''.
        * (*| progress *)
          rewrite clear_bote in ES0; inv ES0.
          desH ST0; last first.
          { hexploit IHSTP; [|eapply ext_trans|M| | |]; (Mskip eauto).
            - i; nbdes; apply f; esplits; last first; eauto.
              do 2 (eapply rt_trans; eauto); desH ST0; subst; eauto.
            - i; nbdes.
              esplits; [ | |by eauto|by eauto]; last first.
              + desH ES2; last by eauto.
                left; eapply esim_progress_mon
                ; try r; s; eauto using tc_rtc, lexprod_tc1.
              + subst; eauto using rt_trans.
          }
          { hexploit IHSTP; [|eapply ext_trans|M| | |]; (Mskip eauto).
            - i; nbdes; apply f; esplits; last first; eauto.
              do 2 (eapply rt_trans; eauto).
              eapply rt_trans; [eapply rt_step|]; eauto.
            - hexploit rtc_tc_tc; [|by eapply t_step; exact STEP|]; eauto; []; i.
              nbdes; desH ES2.
              + hexploit bla; eauto; i; nbdes; eauto 20.
              + esplits; last first; eauto using rt_trans, tc_rtc.
          }
        * (*| call *)
          exfalso; invc CNF0 ES0 SU RF; noemp; invc VR.
          { eapply mess_fun; eauto; eapply mybad_bad_fun; s; eauto. }
          { eapply mess_gen; eauto; eapply mybad_bad_gen; s; eauto. }
      + (*| ret *)
        exfalso; eapply mess_ret; eauto.
    Grab Existential Variables.
      exact wfo_zero.
  Qed.


  Lemma pesim_false_comp: forall g em sa sb sa' sb' sa'' sb''
      ia kt et (SIMa: A.esim (A.wlift lwa) ia kt badcontm (gka g sb') sa sa'' false et em)
      ib ks ty es (SIMb: B.esim (B.wlift lwb) ib badcontm ks (gkb g sa') sb sb'' false ty em es)
      mm (RL: real _ em mm)
      (EXTa: ext _ sa' sa)
      (EXTb: ext _ sb' sb)
      (EXTa: ext _ sa'' sa')
      (EXTb: ext _ sb'' sb')
      r (CIH: esim_comp_CIH r),
    pesim r (existT _ ib ia) kt ks g (scomp sa sb) (scomp sa'' sb'') false ty et es.
  Proof.
    clear; i.
    apply pesim_gknow; i.

    eapply PA.esim_impl_esim_ext in SIMa; last first.
    { i; apply PA.gkmin_unfold in G; des; eauto. }
    eapply PB.esim_impl_esim_ext in SIMb; last first.
    { i; apply PB.gkmin_unfold in G; des; eauto.
      destruct fty; destruct flext_of0; vauto.
    }

    assert (SIMa':=SIMa); clear SIMa
    ; rename SIMa' into SIMa. 
    assert (SIMb':=SIMb); clear SIMb
    ; rename SIMb' into SIMb.

    punfold SIMa; invc SIMa.
    rewrite PA.clear_bote in ESIM.
    pfold; econs; i.
    eapply configure_false_decomp in CNF; eauto; []; des.

    eapply A.configure_false in CNFa.
    eapply B.configure_false in CNFb.
    dup CNFa; apply ESIM in CNFa; eauto using gka_gknow, extra_emp; []; des.
    invc CNFa.
    - hexploit esim_comp_err; eauto using ext_trans; []; i; des; eauto.
    - hexploit esim_comp_step; eauto.
      i; des; eauto using esim_main_taus.
    - hexploit esim_comp_ret; eauto.
      i; des; eauto using esim_main_taus.
  Qed.

  

  Lemma pesim_true_comp: forall g em sa sb sa' sb'
      ia kt et (SIMa: A.esim (A.wlift lwa) ia kt badcontm (gka g sb') sa sa' true et em)
      ib ks ty es (SIMb: B.esim (B.wlift lwb) ib badcontm ks (gkb g sa') sb sb' true ty em es)
      (CO: core em)
      (EXTa: ext _ sa' sa)
      (EXTb: ext _ sb' sb)
      r (CIH: esim_comp_CIH r),
    pesim r (existT _ ib ia) kt ks g (scomp sa sb) (scomp sa' sb') true ty et es.
  Proof.
    clear; i.
    eapply pesim_gknow; i; eapply pesim_gknow_; i.
    destruct (LEM (core et /\ core es)); last by (pfold; econs; i; invc CNF).
    des; eapply pesim_acquire; i.
    eapply confrel_decomp in HR; eauto; []; des.
    hexploit RL; [by exact CO|]; i; des.
    eapply PA.pesim_release in SIMa; eauto.
    eapply PB.pesim_release in SIMb; eauto; [].
    eapply pesim_false_comp; eauto.
  Qed.


  Lemma esim_comp:
    esim_comp_CIH (esim (wlift lw)).
  Proof.
    clear; pcofix CIH; i; destruct sync; des; subst.
    - eapply pesim_true_comp; eauto.
    - eapply pesim_false_comp; eauto.
  Qed.

  
  

  Lemma SIM:
    forall (f: label) ty (IN: lookup f oenv = Some ty), exists vt vs,
      <<Et: vload tgt_lsi mt ancht subt f vt>> /\
      <<Es: vload src_lsi ms anchs subs f vs>> /\
      <<SIM: tlsim ienv subt subs si vt vs ty>>.
  Proof.
    i.
    hexploit SIMa; [by eapply in_map_fst; eauto using list_assoc_someD|]; i; des.
    hexploit SIMb; eauto; []; i; des.
    rrdes Em Em0; rewrite Em0 in Em; clarify.
    esplits; eauto.
    ii; des; rrdes RLS; invc RLS.

    assert (RVa: A.vclos_thunk (A.wlift lwa) (sga, sla)
                               (gka g (sgb, slb) (sga, sla)) vt vm).
    { apply SIM.
      - destruct sga as [[]], sgb as [[]].
        hexploit sdecomp_ext; eauto; i; des; eauto.
        unfold si in EXTa.
        destruct sia as [[[]]], sib as [[[]]]; eauto.
      - eapply gka_gknow; eauto.
      - des; rr; esplits; eauto.
      - i; apply subm_lookupD in INsubb; des; subst.
        rewrite INsuba in INsubb0; clarify.
        hexploit RS; eauto; i.
        eapply vclos_thunk_decomp; eauto.
    }

    assert (RVb: B.vclos_thunk (B.wlift lwb) (sgb, slb)
                               (gkb g (sga, sla) (sgb, slb)) ty vm vs).
    { apply SIM0.
      - destruct sga as [[]], sgb as [[]].
        hexploit sdecomp_ext; eauto; i; des; eauto.
        unfold si in EXTb.
        destruct sia as [[[]]], sib as [[[]]]; eauto.
      - eapply gkb_gknow; eauto.
      - des; rr; esplits; eauto.
      - i; apply subm_lookupD in INsuba; des; subst.
        rewrite INsubb in INsuba1; clarify.
        hexploit RS; eauto; i.
        eapply vclos_thunk_decomp; eauto.
    }

    eapply vclos_thunk_comp; eauto.
    i; eapply gkcomp; eauto using esim_comp.
    - eapply gknow_ant; eauto.
      i; apply pesim_incl_gres; eapply gres_mon; eauto.
      i; ss; des; eauto; apply paco10_mult; eauto.
    - i; inv RVb.
  Qed.

End vcomp_tms.


Theorem vcomp_tms: forall ienv oenv mt mm ms
    (MSIMa: A.msim tm_gw (map fst ienv) (map fst oenv) mt mm)
    (MSIMb: B.msim ms_gw ienv oenv mm ms),
  msim ts_gw ienv oenv mt ms.
Proof.
  ii.
  rrdes MSIMa; rename lw0 into lwa.
  specialize (MSIMb _ CI); rrdes MSIMb; rename lw0 into lwb.

  exists (lw lwa lwb); i.
  rename ancha into ancht.
  rename suba into subt.
  rename ciga into cigt.
  rename cila into cilt.
  rename LMa into LMt.
  rename SUBa into SUBt.
  rename anchb into anchs.
  rename subb into subs.
  rename cigb into cigs.
  rename cilb into cils.
  rename LMb into LMs.
  rename SUBb into SUBs.

  specialize (MSIMa _ _ _ _ LMt _ _ _ _ (LMm mm ienv subt subs) SUBt (SUBm ienv subt subs)); [i; des].
  rename si0 into sia.
  rename STA0 into STAa.
  rename INIg0 into INIga.
  rename INIl0 into INIla.
  rename REF0 into REFa.
  rename SIM0 into SIMa.

  specialize (MSIMb _ _ _ _ (LMm mm ienv subt subs) _ _ _ _ LMs (SUBm ienv subt subs) SUBs); [i; des].
  rename si0 into sib.
  rename STA0 into STAb.
  rename INIg0 into INIgb.
  rename INIl0 into INIlb.
  rename REF0 into REFb.
  rename SIM0 into SIMb.
  rename NAM into NAMb.

  exists (si lwa lwb sia sib); splits.
  - eauto.
  - eapply STA; eauto.
  - eapply INIg; eauto.
  - eapply INIl; eauto.
  - eapply REF; eauto.
  - eapply SIM; eauto.
Qed.
